#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2024  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import os
import sys

if "__file__" in locals():
    sys.path.insert(0, os.path.realpath(os.path.join( os.path.dirname(os.path.realpath(__file__)),'..')))

import uuid as _uuid

import datetime
import getpass
import traceback
import platform
import hashlib
import re
import time
import json
import ujson

from peewee import Proxy, TextField,BooleanField,AutoField,CharField, DateTimeField, IntegerField, BigIntegerField, BigAutoField
from peewee import fn,SQL,PrimaryKeyField,Model,TimestampField, ForeignKeyField, CompositeKey,Function,FloatField,IntegrityError

from waptutils import __version__

from playhouse.postgres_ext import BinaryJSONField, JSONField
from playhouse.postgres_ext import ArrayField

from playhouse.pool import PooledPostgresqlExtDatabase

#from playhouse.shortcuts import dict_to_model, model_to_dict
from playhouse.signals import Model as SignaledModel, pre_save

from waptutils import Version
from waptutils import ensure_unicode, ensure_list, datetime2isodate, isodate2datetime, jsondump

from waptcrypto import SSLCABundle, EWaptCryptoException
from waptpackage import PackageVersion, make_packages_filter_string

from waptserver.utils import setloglevel

from waptserver.app import app

from optparse import OptionParser

import waptserver.config

# You must be sure your database is an instance of PostgresqlExtDatabase in order to use the JSONField.

import logging
logger = logging.getLogger('waptdb')

wapt_db = Proxy()

DB_STRUCTURE_VERSION = '2.5.2.1'

def load_db_config(server_config=None):
    """Initialise db proxy with parameters from inifile

    Args:
        serverconfig (dict): dict of server parameters as returned by waptserver.config.load_config(ainifilename)

    Returns
        configured db : db which has been put in wapt_db proxy
    """
    global wapt_db
    if server_config is None:
        server_config = waptserver.config.load_config()

    logger.info('Initializing a DB connection pool for db host %s:%s db_name:%s. Size:%s' %
                (server_config['db_host'], server_config['db_port'], server_config['db_name'], server_config['db_max_connections']))
    pgdb = PooledPostgresqlExtDatabase(
        database=server_config['db_name'],
        host=server_config['db_host'],
        port=server_config['db_port'],
        user=server_config['db_user'],
        password=server_config['db_password'],
        max_connections=server_config['db_max_connections'],
        stale_timeout=server_config['db_stale_timeout'],
        timeout=server_config['db_connect_timeout'],
        autoconnect=False)
    wapt_db.initialize(pgdb)
    return pgdb


def wapt_db_connect():
    if wapt_db and wapt_db.is_closed():
        logger.info('Connecting to waptdb explicitely. Count: %s' % len(wapt_db._in_use))
        wapt_db.connect()
        return True
    return False


def wapt_db_close():
    if wapt_db and not wapt_db.is_closed():
        if wapt_db.in_transaction():
            try:
                logger.critical('Transaction is active, but DB must be closed. Rollbacking')
                wapt_db.rollback()
            except Exception as e:
                logger.critical('Unable to rollback:\n%s' % repr(e))
                logger.critical(traceback.format_exc())
        logger.debug('Disconnecting from waptdb explicitely. Count: %s' % len(wapt_db._in_use))
        wapt_db.close()
        return True
    return False


class WaptDB:
    def __init__(self):
        self.must_close = True

    def __enter__(self):
        self.must_close = wapt_db_connect()

    def __exit__(self, type, value, tb):
        if self.must_close:
            if not value:
                #logger.debug(u'DB exit %i' % self.transaction_depth)
                wapt_db_close()
            else:
                try:
                    logger.debug('Error at DB exit %s, rollbacking\n%s' % (value, ensure_unicode(traceback.format_tb(tb))))
                    if wapt_db.in_transaction():
                        wapt_db.rollback()
                    wapt_db_close()
                except Exception as e:
                    logger.critical('Unable to rollback or close DB:\n%s' % repr(e))
                    logger.critical(traceback.format_exc())


class WaptBaseModel(SignaledModel):
    """A base model that will use our Postgresql database"""
    class Meta(object):
        database = wapt_db

    # audit data
    created_on = DateTimeField(null=True)
    created_by = CharField(null=True)
    updated_on = DateTimeField(null=True)
    updated_by = CharField(null=True)

    def __unicode__(self):
        return '%s' % (self.__data__,)

    def __str__(self):
        return '%s' % (self.__data__,)


@pre_save(sender=WaptBaseModel)
def waptbasemodel_pre_save(model_class, instance, created):
    if created:
        instance.created_on = datetime.datetime.utcnow()
    instance.updated_on = datetime.datetime.utcnow()


class ServerAttribs(SignaledModel):
    """key/value registry"""
    class Meta(object):
        database = wapt_db

    key = CharField(primary_key=True, null=False, index=True)
    value = BinaryJSONField(null=True)

    @classmethod
    def dump(cls):
        for key, value in cls.select(cls.key, cls.value).tuples():
            print(('%s: %s' % (key, repr(value))))

    @classmethod
    def get_value(cls, key):
        v = cls.select(cls.value).where(cls.key == key).dicts().first()
        if v:
            return v['value']
        else:
            return None

    @classmethod
    def set_value(cls, key, value):
        with cls._meta.database.atomic():
            try:
                cls.create(key=key, value=value)
            except IntegrityError:
                wapt_db.rollback()
                cls.update(value=value).where(cls.key == key).execute()

    def __unicode__(self):
        return '%s' % (self.__data__,)

    def __str__(self):
        return self.__unicode__().encode('utf8')


class Hosts(WaptBaseModel):
    """Inventory informations of a host
    """
    # from bios
    uuid = CharField(primary_key=True, index=True)

    # inventory type data (updated on register)
    computer_fqdn = CharField(null=True, index=True)
    description = CharField(null=True, index=True)
    computer_name = CharField(null=True)
    computer_type = CharField(null=True)  # tower, laptop,etc..
    computer_architecture = CharField(null=True)  # tower, laptop,etc..
    manufacturer = CharField(null=True)
    productname = CharField(null=True)
    serialnr = CharField(null=True)

    host_certificate = TextField(null=True, help_text='Host public X509 certificate bundle')
    registration_auth_user = CharField(null=True)

    #authorized_certificates = ArrayField(CharField, null=True, help_text='authorized packages signers certificates sha1 fingerprint')
    authorized_certificates_sha256 = ArrayField(CharField, null=True, help_text='authorized packages signers certificates sha256 fingerprint')

    os_name = CharField(null=True)
    os_version = CharField(null=True)
    os_architecture = CharField(null=True)
    platform = CharField(null=True)

    # frequently updated data from host update_status
    connected_users = ArrayField(CharField, null=True)
    connected_ips = ArrayField(CharField, null=True)
    mac_addresses = ArrayField(CharField, null=True)
    gateways = ArrayField(CharField, null=True)
    networks = ArrayField(CharField, null=True)
    dnsdomain = CharField(null=True)

    computer_ad_site = CharField(null=True, index=True)
    computer_ad_ou = CharField(null=True, index=True)
    computer_ad_groups = ArrayField(CharField, null=True)

    # for websockets
    server_uuid = CharField(null=True)

    # for repo sync
    repositories = CharField(null=True)

    # OK, TO-UPGRADE, ERROR, RUNNING
    host_status = CharField(null=True)
    last_seen_on = CharField(null=True)
    last_logged_on_user = CharField(null=True)

    # OK, WARNING, ERROR
    audit_status = CharField(null=True)

    #
    wapt_version = CharField(null=True)

    # raw json data
    wapt_status = BinaryJSONField(null=True)

    # running, pending, errors, finished , upgradable, errors,
    last_update_status = BinaryJSONField(null=True, index=False)
    host_info = BinaryJSONField(null=True)
    host_capabilities = BinaryJSONField(null=True)
    host_metrics = BinaryJSONField(null=True)
    host_networking = BinaryJSONField(null=True)
    host_packages_filter = TextField(null=True,index=True)

    # variable structures... so keep them as json
    dmi = BinaryJSONField(null=True, index=False)
    wmi = BinaryJSONField(null=True)

    wuauserv_status = BinaryJSONField(null=True, index=False)
    waptwua_status = BinaryJSONField(null=True, index=False)
    waptwua_rules = BinaryJSONField(null=True, index=False)

    #
    status_hashes = BinaryJSONField(null=True, index=False)

    def __repr__(self):
        return '<Host fqdn=%s / uuid=%s>' % (self.computer_fqdn, self.uuid)

    @classmethod
    def fieldbyname(cls, fieldname):
        return cls._meta.fields[fieldname]


class HostPackagesStatus(WaptBaseModel):
    """Stores the status of packages installed on a host
    """
    id = BigAutoField(primary_key=True)
    host = ForeignKeyField(Hosts, on_delete='CASCADE', on_update='CASCADE')
    install_id = IntegerField(null=True)
    package_uuid = CharField(null=True)
    name = CharField(null=True)
    package = CharField(null=True, index=True)
    version = CharField(null=True)
    architecture = CharField(null=True)
    locale = CharField(null=True)
    maturity = CharField(null=True)
    section = CharField(null=True)
    priority = CharField(null=True)
    signer = CharField(null=True)
    signer_fingerprint = CharField(null=True)
    signature_date = CharField(null=True)
    description = TextField(null=True)
    impacted_process = ArrayField(CharField, null=True)
    install_status = CharField(null=True)
    install_date = CharField(null=True)
    install_output = TextField(null=True)
    install_params = CharField(null=True)
    uninstall_key = ArrayField(CharField, null=True)
    explicit_by = CharField(null=True)
    repo_url = CharField(max_length=600, null=True)
    depends = ArrayField(CharField, null=True)
    conflicts = ArrayField(CharField, null=True)
    last_audit_status = CharField(null=True)
    last_audit_on = CharField(null=True)
    last_audit_output = TextField(null=True)
    next_audit_on = CharField(null=True)

    def __repr__(self):
        return '<HostPackageStatus uuid=%s packages=%s (%s) install_status=%s>' % (self.id, self.package, self.version, self.install_status)


class WaptUsers(WaptBaseModel):
    """Users
    """
    id = AutoField(primary_key=True)
    name = CharField(null=True, index=False, help_text='Username to authenticate with password')
    auth_method = CharField(null=True, index=False, help_text='Authorized authentication method')
    description = CharField(null=True, index=False)
    user_certificate = TextField(null=True, help_text='User public X509 certificates chain. Root certificate must be approved on server')
    user_fingerprint = CharField(null=True, index=True, help_text='Sha256 calculated from user_certificate x509')
    user_fingerprint_sha1 = CharField(null=True, index=True, help_text='Sha1 calculated from user_certificate x509')
    last_login_timestamp = TimestampField(null=True)

    def __repr__(self):
        return '<WaptUsers id=%s name=%s>' % (self.id, self.name)


PACKAGE_SECTIONS = ['base','group','unit','profile','wsus','selfservice']


class WaptUserAcls(WaptBaseModel):
    """User acls for a list of perimeters.
    Acls is a JSON dict.
    Keys are acl names.
    Value is either boolean (no specific restriction) or list of perimter where the right applies :
      - server uuids if the ACL is related to database or server management
      - fingerprints of CA certificate of packages/actions if the ACL is related to host perimeters
      - name patterns (can include wildcards) if the ACL is related to a package
    """
    id = AutoField(primary_key=True)
    username = CharField(null=False, index=False, help_text='Username pattern, * for all')
    user_fingerprint = CharField(null=True, index=True, help_text='sha256 fingerprint of user certificate')  # matches perimeters
    user_fingerprint_sha1 = CharField(null=True, index=True, help_text='sha1 fingerprint of user certificate')  # sha1 here to be able to match nginx auth
    acls = JSONField(null=True) # dict {'aclname': true/false or list}
    expiration_date = CharField(null=True, help_text='End of ACL date')
    signer = CharField(null=True, help_text='ACL signer')
    signer_fingerprint = CharField(null=True, help_text='SHA256 signer fingerprint')
    signature_date = CharField(null=True, help_text='ACL date')
    signed_attributes = BinaryJSONField(null=True)
    signer_certificate = TextField(null=True, help_text='ACL signer certificate PEM')
    signature = TextField(null=True, help_text='ACL signature (serialize as as json dict with sorted keys)')


    def _signed_attributes(self):
        return ['username', 'user_fingerprint', 'user_fingerprint_sha1', 'acls',
                'expiration_date',
                'signer', 'signature_date', 'signer_fingerprint', 'signed_attributes', 'signer_certificate']

    def __repr__(self):
        return '<WaptUserAcls id=%s for username=%s -> acls=%s>' % (self.id,self.username,self.acls)

    def to_dict(self):
        _dict = {k: getattr(self, k) for k in self._signed_attributes()}
        return _dict

    def sign(self, key, cert):
        self.signer = cert.cn
        self.signer_fingerprint = cert.fingerprint
        self.signature_date = datetime2isodate(datetime.datetime.utcnow())
        self.signature = key.sign_content(self.to_dict(), pre_py3=False)
        self.save()

    def verify(self, cabundle):
        """Check that the ACL is properly signed and signed by a trusted user
        """
        # get signer certificate
        signer_cert = cabundle.certificate_for_fingerprint(self.signer_fingerprint)
        if not signer_cert:
            raise EWaptCryptoException('ACL signer is not known')
        user_dict = self.to_dict()
        # verify signature
        signer_cert.verify_content(user_dict, self.signature)

        # verify signer is trusted
        cert_chain = cabundle.check_certificates_chain(signer_cert, check_is_trusted=True)
        logger.debug('ACL is trusted by %s' % cert_chain[-1])
        return cert_chain


class WaptRoles(WaptBaseModel):
    """Roles.
    Acls is a JSON dict.
    Keys are acl names.
    Value is either boolean (no specific restriction) or list of perimter where the right applies :
      - server uuids if the ACL is related to database or server management
      - fingerprints of CA certificate of packages/actions if the ACL is related to host perimeters
      - name patterns (can include wildcards) if the ACL is related to a package

    """
    id = AutoField(primary_key=True)
    name = CharField(null=False, index=True, help_text='Role name')
    description = CharField(null=False, help_text='Role description')
    acls = JSONField(null=True, help_text='Dict {"aclname": bool or list}')
    """
        view = can "View" hosts and packages for (package/action trusted CA fingerprints)
        admin = is Admin for (server uuids)
        register_host = can "register" hosts to servers (server uuids)
        unregister_host = can "unregister" hosts from servers (server uuids)
        edit_repo_rules = can edit repo rules (server uuids)
        trigger_host_action = can trigger remote action (package/action trusted CA fingerprints)
        edit_host_package = can edit "host" package (package/action trusted CA fingerprints)
        edit_base_package = can edit "base" package (package name patterns)
        edit_group_package = can edit "group" package (package name patterns)
        edit_unit_package = can edit "unit" package (package name patterns)
        edit_wsus_package = can edit "wsus" package (package name patterns)
        edit_profile_package = can edit "profile" package (package name patterns)
        edit_selfservice_package = can edit "selfservice" package (package name patterns)
        assign_package can assign or unassign package (package name patterns)
        edit_report = can edit reports (server uuids)
        run_report = can run reports (server uuids)
    """
    signer = CharField(null=True, help_text='Role signer')
    signature_date = CharField(null=True, help_text='Role date')
    signer_fingerprint = CharField(null=True, help_text='Signer fingerprint')
    signature = CharField(null=True, help_text='Serialzed Role signature (serialize as as json dict with sorted keys)')

    def _signed_attributes(self):
        return ['name', 'description','acls',
                'signer', 'signature_date', 'signer_fingerprint']

    def __repr__(self):
        return '<WaptRoles id=%s name=%s acls=%s>' % (self.id,self.name,self.acls)

    def to_dict(self):
        _dict = {k: getattr(self, k) for k in self._signed_attributes()}
        return _dict

    def sign(self, key, cert):
        self.signer = cert.cn
        self.signer_fingerprint = cert.fingerprint
        self.signature_date = datetime2isodate(datetime.datetime.utcnow())
        self.signature = key.sign_content(self.to_dict(), pre_py3=False)
        self.save()

    def verify(self, cabundle):
        """Check that the ACL is properly signed and signed by a trusted user
        """
        # get signer certificate
        signer_cert = cabundle.certificate_for_fingerprint(self.signer_fingerprint)
        if not signer_cert:
            raise EWaptCryptoException('ACL signer is not known')
        user_dict = self.to_dict()
        # verify signature
        signer_cert.verify_content(user_dict, self.signature)

        # verify signer is trusted
        cert_chain = cabundle.check_certificates_chain(signer_cert, check_is_trusted=True)
        logger.debug('Role is trusted by %s' % cert_chain[-1])
        return cert_chain


class Packages(WaptBaseModel):
    """Stores the content of packages of repositories
    """
    id = BigAutoField(primary_key=True)
    package_uuid = CharField(null=True)
    package = CharField(null=False, index=True)
    version = CharField(null=False)
    description = CharField(max_length=1200, null=True)
    architecture = ArrayField(CharField, null=True) # List
    locale = ArrayField(CharField, null=True) # List
    maturity = CharField(null=True)
    section = CharField(null=True)
    priority = CharField(null=True)
    signer = CharField(null=True)
    signer_fingerprint = CharField(null=True)
    signature_date = CharField(null=True)
    description = TextField(null=True)
    depends = ArrayField(CharField, null=True)
    conflicts = ArrayField(CharField, null=True)
    audit_schedule = CharField(null=True)
    valid_from = CharField(null=True)
    valid_until = CharField(null=True)
    forced_install_on = CharField(null=True)
    installed_size = BigIntegerField(null=True)
    target_os = ArrayField(CharField, null=True) #list
    min_os_version = CharField(null=True)
    max_os_version = CharField(null=True)
    min_wapt_version = CharField(null=True)
    impacted_process = ArrayField(CharField, null=True)
    keywords = ArrayField(CharField, null=True)
    name = CharField(null=True)
    categories = ArrayField(CharField, null=True)
    licence = ArrayField(CharField, null=True) #list
    editor = CharField(null=True)
    homepage = CharField(null=True)
    filename = CharField(null=True, index=True)
    icon_sha256sum = CharField(null=True)
    is_available = BooleanField(null=True,help_text='True if package is available on this server')
    changelog = TextField(null=True)
    maintainer = TextField(null=True)
    md5sum = TextField(null=True)
    sources = TextField(null=True)
    size = BigIntegerField(null=True)
    filename = TextField(null=True)
    description_localized=BinaryJSONField(null=True)

    @classmethod
    def _as_attribute(cls, k, v):
        if k in ['depends', 'conflicts', 'impacted_process', 'keywords', 'architecture', 'categories', 'locale', 'target_os', 'licence']:
            return ensure_list(v or None)
        else:
            return v or None

    @classmethod
    def update_from_control(cls, entry, is_available=None):
        """Create or update a single package entry in database given a PackageEntry

        """
        if entry.package_uuid:
            key = {'package_uuid': entry.package_uuid,
               'package': entry.package, 'version': entry.version}
        else:
            key = {'package': entry.package, 'version': entry.version}

        (rec, _isnew) = Packages.get_or_create(**key)
        for (a, v) in entry.as_dict().items():
            if a in cls._meta.columns and not a in key:
                new_value = cls._as_attribute(a, v)
                if new_value != getattr(rec, a):
                    setattr(rec, a, cls._as_attribute(a, v))
            elif a.startswith('description_'):
                locale = a[12:]
                if rec.description_localized is None or v != rec.description_localized.get(locale):
                    if rec.description_localized is None:
                        rec.description_localized = { locale:v }
                    else:
                        if (v is None or v == ''):
                            if locale in rec.description_localized:
                                del rec.description_localized[locale]
                        else:
                            rec.description_localized[locale] = v

        if not rec.package_uuid:
            rec.package_uuid = entry.make_fallback_uuid()
        if is_available is not None:
            rec.is_available = is_available

        if rec.is_dirty():
            rec.save()

        return (rec, _isnew)

    @classmethod
    def update_from_repo(cls, repo):
        """Update Packages table with all the Packages entries from repo WaptRepo

        Args:
            repo (WaptRepo):
        Returns:
            list of PackagEntry added to the table
        """
        result = []
        all_available = []
        package_uuids = {}
        with cls._meta.database.atomic() as trans:
            try:
                for pe in repo.packages():
                    uuid = pe.package_uuid or pe.make_fallback_uuid()
                    if not uuid in package_uuids:
                        package_uuids[uuid] = 1
                        try:
                            (rec, _isnew) = cls.update_from_control(pe,is_available=True)
                            if _isnew:
                                result.append(pe)
                            all_available.append(pe.package_uuid)
                        except Exception as e:
                            logger.critical('%s %s' % (pe,e))
                Packages.update(is_available=False).where(~ Packages.package_uuid.in_(all_available)).execute()
                # remove records for not available packages which are no longer present on hosts.
                to_delete = wapt_db.execute_sql('delete from packages p where p.is_available=False and not exists(select s.package_uuid from hostpackagesstatus s where s.package_uuid=p.package_uuid)')
                logger.debug('%s DB Packages records deleted because not available in repo and not referenced by hosts anymore',to_delete.rowcount)
                trans.commit()
            except:
                trans.rollback()
                raise
        return result


class HostSoftwares(WaptBaseModel):
    """Content of Host's softwares uninstall registry
    """
    id = BigAutoField(primary_key=True)
    host = ForeignKeyField(Hosts, on_delete='CASCADE', on_update='CASCADE')
    software_id = CharField(max_length=600, null=True)
    name = CharField(max_length=2000, null=True, index=True)
    version = CharField(max_length=1000, null=True)
    publisher = CharField(max_length=2000, null=True)
    key = CharField(max_length=600, null=True)
    system_component = CharField(null=True)
    uninstall_string = CharField(max_length=2000, null=True)
    install_date = CharField(null=True)
    install_location = CharField(max_length=2000, null=True)

    def __repr__(self):
        return '<HostSoftwares uuid=%s name=%s (%s) key=%s>' % (self.uuid, self.name, self.version, self.key)


class HostGroups(WaptBaseModel):
    """Mirror of the depends attribut of Hosts's packages

    Updated when a host package is uploaded to the server
    Helps to speed up filtering Hosts list on groups.
    """
    id = BigAutoField(primary_key=True)
    host = ForeignKeyField(Hosts, on_delete='CASCADE', on_update='CASCADE')
    group_name = CharField(null=False, index=True)

    def __repr__(self):
        return '<HostGroups id=%s group_name=%s>' % (self.id, self.group_name)

    @classmethod
    def sync_from_host_package(cls, entry):
        """Update HostGroups table from Host Package depends.
        Add / Remove host <-> group link based on entry.depends csv attribute

        Args:
            entry (PackageEntry): Host package entry

        Returns
            tuple: (added depends, removed depends)
        """
        with cls._meta.database.atomic() as trans:
            try:
                host_id = entry.package

                # insert /delete depends as groups
                if entry.depends:
                    depends = [s.strip() for s in ensure_list(entry.depends)]
                else:
                    depends = []
                old_groups = [h['group_name'] for h in HostGroups.select(HostGroups.group_name).where(HostGroups.host == host_id).dicts()]
                to_delete = [g for g in old_groups if not g in depends]
                to_add = [g for g in depends if not g in old_groups]
                if to_delete:
                    HostGroups.delete().where((HostGroups.host == host_id) & (HostGroups.group_name.in_(to_delete))).execute()
                if to_add:
                    HostGroups.insert_many([dict(host=host_id, group_name=group) for group in to_add]).execute()  # pylint: disable=no-value-for-parameter
                return (to_add, to_delete)
            except IntegrityError:
                trans.rollback()
                return (0, 0)

class HostWebsocket(Model):
    """Stores the timed audit data reported by hosts audit

    """
    class Meta(object):
        database = wapt_db

    id = BigAutoField(primary_key=True)
    host = ForeignKeyField(Hosts, index=True, unique=True, on_delete='CASCADE', on_update='CASCADE')
    server_uuid = CharField(null=True)
    session_id = CharField(null=True,index=True)
    timestamp = TimestampField(null=True)


class HostAuditData(Model):
    """Stores the timed audit data reported by hosts audit

    """
    class Meta(object):
        database = wapt_db

    id = BigAutoField(primary_key=True)
    host = ForeignKeyField(Hosts, on_delete='CASCADE', on_update='CASCADE')
    value_id = BigIntegerField(index = True,null=True)
    value_date = TextField(index = True)
    value_section = TextField(null=True)
    value_key = TextField(index = True)
    value = BinaryJSONField(null=True)
    expiration_date = TextField(index = True,null=True)
    created_on = DateTimeField(null=True)
    created_by = CharField(null=True)

    def __repr__(self):
        return '<HostAuditData id=%s host=%s date=%s %s.%s=%s>' % (self.id, self.host, self.value_date, self.value_section, self.value_key, self.value)

    def __unicode__(self):
        return '%s' % (self.__data__,)

    def __str__(self):
        return self.__unicode__().encode('utf8')

class ReportingQueries(WaptBaseModel):
    """Reporting queries"""
    id = AutoField(primary_key=True)
    name = CharField(null=False, index=True)
    query = TextField(null=True)
    settings = BinaryJSONField(null=True, index=False)
    snapshot_scheduling = CharField(null=True)  # period in seconds
    snapshot_name = CharField(null=True)
    last_snapshot_date = TimestampField(null=True)
    snapshot_ttl = IntegerField(null=True)     # period in seconds

    as_host_uuid = CharField(null=True)
    as_package_uuid = CharField(null=True)
    as_kb_update_id = CharField(null=True)
    as_software_key = CharField(null=True)

    def __repr__(self):
        return '<ReportingQueries id=%s name=%s>' % (self.id, self.name)

    def snapshot(self):
        if self.snapshot_name:
            with wapt_db.atomic() as trans:
                report_date = datetime.datetime.utcnow()

                query = self.query.replace('\n',' ')
                if query.endswith(';'):
                    query = query[:-1]

                insquery = ReportingSnapshots.insert_from(
                    query=SQL('(with reportquery as (%s) SELECT %%s, %%s, %%s, row_number() over(),row_to_json(reportquery) from reportquery)' % query,[self.id,self.snapshot_name,report_date]),
                    fields=['report_id','report_name','report_date','row_nr','row_data']
                )
                result = insquery.execute()

                self.last_snapshot_date = report_date
                self.save()

                if not self.snapshot_ttl is None:
                    ReportingSnapshots.delete().where(
                        ReportingSnapshots.report_id == self.id,
                        ReportingSnapshots.report_date <= report_date - datetime.timedelta(seconds=self.snapshot_ttl)
                    ).execute()

                trans.commit()
                return result

class ReportingSnapshots(Model):
    """Stores the Reporting results snapshots
    Each result row is described with a json object
    """

    class Meta(object):
        database = wapt_db

    id = BigAutoField(primary_key=True)
    report_id = ForeignKeyField(ReportingQueries, on_delete='CASCADE', on_update='CASCADE')
    report_name = CharField()
    report_date = DateTimeField(null=True)
    row_nr = BigIntegerField(null=True)
    row_data = BinaryJSONField(null=True)

    def __repr__(self):
        return '<ReportingSnapsphots id=%s name=%s date=%s row_nr=%s>' % (self.id, self.report_name, self.report_date, self.row_nr)


class Normalization(WaptBaseModel):
    """Normalization table"""
    original_name = TextField()
    key = TextField()
    normalized_name = TextField(null=True,  index=True)
    banned = BooleanField(null=True)
    windows_update = BooleanField(null=True)

    class Meta(object):
        primary_key = CompositeKey('original_name', 'key')

    def __repr__(self):
        return '<Normalization uuid=%s name=%s>' % (self.uuid, self.name)


class WsusUpdates(WaptBaseModel):
    update_id = CharField(primary_key=True, index=True)
    title = CharField()
    update_type = CharField()
    kbids = ArrayField(CharField, index=True, null=True)
    severity = CharField(null=True)
    changetime = CharField(null=True)
    product = CharField(null=True, index=True)
    classification = CharField(null=True)
    download_urls = ArrayField(CharField, null=True)
    min_download_size = BigIntegerField(null=True)
    max_download_size = BigIntegerField(null=True)
    superseded_update_ids = ArrayField(CharField, null=True)
    security_bulletin_ids = ArrayField(CharField, null=True)
    cve_ids = ArrayField(CharField, null=True)
    is_mandatory = BooleanField(null=True)
    reboot_behaviour = CharField(null=True)
    can_request_user_input = CharField(null=True)
    requires_network_connectivity = BooleanField(null=True)
    is_beta = BooleanField(null=True)
    is_uninstallable = BooleanField(null=True)
    installation_impact = CharField(null=True)
    uninstallation_impact = CharField(null=True)
    support_url = CharField(null=True)
    release_notes = TextField(null=True)
    uninstallation_notes = TextField(null=True)
    languages = ArrayField(CharField, null=True)
    downloaded_on = CharField(null=True)


class HostWsus(WaptBaseModel):
    """List of Windows Update discovered by waptwua for a Host with install status
    """
    id = BigAutoField(primary_key=True)
    host = ForeignKeyField(Hosts, on_delete='CASCADE', on_update='CASCADE')
    update_id = ForeignKeyField(WsusUpdates, on_delete='CASCADE', on_update='CASCADE')
    status = CharField(null=True)
    allowed = BooleanField(null=True)
    installed = BooleanField(null=True)
    present = BooleanField(null=True)
    hidden = BooleanField(null=True)
    delayed = BooleanField(null=True)
    downloaded = BooleanField(null=True)
    install_date = CharField(null=True)
    history = BinaryJSONField(null=True, index=False)


class SignedModel(WaptBaseModel):
    uuid = CharField(primary_key=True, null=False, default=str(_uuid.uuid4()))

    def save(self, *args, **argvs):
        if 'uuid' in self._dirty:
            argvs['force_insert'] = True
        return super(SignedModel, self).save(*args, **argvs)


class WsusDownloadTasks(WaptBaseModel):
    id = AutoField(primary_key=True)
    task_id = CharField(null=True, index=True)
    kind = CharField(null=True, index=True)
    url = CharField(null=True)
    target_size = BigIntegerField(null=True)
    local_filename = CharField(null=True)
    file_date = DateTimeField(null=True)
    file_size = BigIntegerField(null=True)
    status = CharField(null=True)
    started_on = DateTimeField(null=True, index=True)
    finished_on = DateTimeField(null=True, index=True)
    msg = CharField(null=True)
    forced = BooleanField(null=True)
    skipped = BooleanField(null=True)
    error = CharField(max_length=1200, null=True)


class SiteRules(WaptBaseModel):
    id = AutoField(primary_key=True)
    sequence = IntegerField(null=True)
    name = CharField(null=True)
    repositories = JSONField(null=True)
    condition = CharField(null=True)
    value = CharField(null=True)
    repo_url = CharField(null=True)
    signature_date = CharField(null=True)
    signed_attributes = JSONField(null=True)
    signer = CharField(null=True)
    signature = TextField(null=True)
    signer_certificate = TextField(null=True)
    negation = BooleanField(null=True)
    no_fallback = BooleanField(null=True)
    has_proxy = BooleanField(null=True, default=False)
    http_proxy = TextField(null=True)
    signer_fingerprint = CharField(null=True)


class SyncStatus(WaptBaseModel):
    id = AutoField(primary_key=True)
    version = IntegerField(null=True)
    changelog = JSONField(null=True)
    progress = IntegerField(null=True)


class HostSyncStatus(WaptBaseModel):
    id = AutoField(primary_key=True)
    uuid = CharField(null=False)
    status = CharField(null=True)
    current_download = TextField(null=True)
    last_sync = DateTimeField(null=True)
    last_audit = DateTimeField(null=True)
    current_speed = CharField(null=True)
    speed_average = CharField(null=True)
    errors = JSONField(null=True)
    progress = FloatField(null=True)
    sync_id = IntegerField(null=True)
    version = IntegerField(null=True)


class LogsAPI(WaptBaseModel):
    """
    Logs of API requests
    """
    id = BigAutoField(primary_key=True)
    endpoint = TextField(null=True)
    args = JSONField(null=True)
    data = TextField(null=True)
    forbidden = BooleanField(null=True)
    authenticated = BooleanField(null=True)
    user = TextField(null=True)
    user_acls = TextField(null=True)
    host = TextField(null=True)
    remote_addr = TextField(null=True)
    header_user_agent = TextField(null=True)
    method = TextField(null=True)
    scheme = TextField(null=True)
    path = TextField(null=True)
    full_path = TextField(null=True)

class Plugins(WaptBaseModel):
    """ Plugin Parameters """
    name = TextField(unique=True)
    arguments = TextField(null=True)
    show_result = BooleanField(null=True, default=True)
    executable = TextField(null=True)
    timeout = IntegerField(null=True)
    retry = IntegerField(null=True)
    filepath = TextField(null=True)
    acls = TextField(null=True)


class WaptSignedModel(WaptBaseModel):
    """ Parameters of new host to deploy """
    id = AutoField(primary_key=True)
    signature_date = TextField(null=True)
    signer = TextField(null=True)
    signer_fingerprint = TextField(null=True)
    signer_certificate = TextField(null=True)
    signed_attributes = BinaryJSONField(null=True)
    signature = TextField(null=True)


class HostOSDeploy(WaptSignedModel):
    """ Parameters of new host to deploy """
    mac_addresses = TextField(null=True)
    status = TextField(null=True)
    waiting_to_deploy = BooleanField(null=True)
    uuid = TextField(null=True)
    hostname = TextField(null=True)
    deploy_config_name = TextField(null=True)
    product_key = TextField(null=True)
    driver_bundle = TextField(null=True)
    disk_format = TextField(null=True)
    djoin = TextField(null=True)
    manufacturer = TextField(null=True)
    product = TextField(null=True)
    serial_number = TextField(null=True)
    asset = TextField(null=True)
    ei_cfg = TextField(null=True)
    pid_txt = TextField(null=True)
    diskpart_list_disk = TextField(null=True)
    diskpart_list_vol = TextField(null=True)
    ip_address = TextField(null=True)
    use_ip_from_winpe = BooleanField(null=True)

class DeployConfig(WaptSignedModel):
    name = TextField(null=True)
    os_type = TextField(null=True)
    xml_conf = TextField(null=True)
    iso_config_name = TextField(null=True)
    script_postinstall = TextField(null=True)
    install_wapt = BooleanField(null=True)
    ipxe_script = TextField(null=True)

class IsoConfig(WaptSignedModel):
    id = PrimaryKeyField(primary_key=True)
    name = TextField(null=True)
    architecture = TextField(null=True)
    hash = TextField(null=True)

class MessagesTemplate(WaptBaseModel):
    name = TextField(null=True)
    html_content = TextField(null=True)
    user = TextField(null=True)

class HostConfigs(WaptBaseModel):
    config = JSONField(null=True)
    priorities = JSONField(null=True)
    host = ForeignKeyField(Hosts, on_delete='CASCADE', on_update='CASCADE')

def get_member_variables(aclass, default_exclude=['created_on', 'updated_by', 'created_by', 'updated_on'], add_exclude=[]):
    return [attr for attr in vars(aclass) if not(attr.startswith('_')) and (attr not in ['DoesNotExist']+default_exclude+add_exclude)]


def dictgetpath(adict, pathstr):
    """Iterates a list of path pathstr of the form 'key.subkey.sskey' and returns
    the first occurence in adict which returns not None.

    A path component can contain a wildcard '*' to match an array.

    """
    if not isinstance(pathstr, (list, tuple)):
        pathstr = [pathstr]
    for path in pathstr:
        result = adict
        for k in path.split('.'):
            if isinstance(result, dict):
                # assume this level is an object and returns the specified key
                result = result.get(k)
            elif isinstance(result, list) and k.isdigit() and int(k) < len(result):
                # assume this level is a list and return the n'th item
                result = result[k]
            elif k == '*' and isinstance(result, list):
                # assume this level is an array, and iterates all items
                continue
            elif isinstance(k, str) and isinstance(result, list):
                # iterate through a list returning only a key
                result = [item.get(k) for item in result if item.get(k)]
            else:
                # key not found, we have to retry with next path
                result = None
                break
        if result:
            break
    return result


def set_host_field(host, fieldname, data):
    # awfull hack for data containing null char, not accepted by postgresql.
    if fieldname in ('host_info', 'wmi', 'dmi'):
        jsonrepr = ujson.dumps(data)
        if '\\u0000' in jsonrepr:
            logger.warning('Workaround \\u0000 not handled by postgresql json for host %s field %s' % (getattr(host, 'uuid', '???'), fieldname))
            data = ujson.loads(jsonrepr.replace('\\u0000', ' '))

    setattr(host, fieldname, data)
    return host


def package_version_from_prequest(prequest):
    """Split package and version from  "package(=version)"

    Returns:
        tuple (str,str): (package,version)
    """
    package_version_re = re.compile('([^()]+)\s*(\(\s*([<=>]*)\s*(\S+)\s*\))?.*')
    match = package_version_re.match(prequest)
    if match and len(match.groups()) == 4:
        result = match.groups()
        return (result[0], result[3])
    else:
        return None

def update_installed_packages(uuid, data, applied_status_hashes):
    """Stores packages json data into separate HostPackagesStatus

    Merge

    Args:
        uuid (str) : unique ID of host
        data (dict): data from host
        applied_status_hashes (dict): add the supplied hash into 'installed_packages' key

    Returns:
        None

    """
    def _get_uninstallkeylist(uninstall_key_str):
        """Decode uninstallkey list from db field
        For historical reasons, this field is encoded as str(pythonlist)
        or sometimes simple repr of a str

        Returns:
            list
        """
        if uninstall_key_str:
            if uninstall_key_str.startswith("['") or uninstall_key_str.startswith("[u'"):
                # python encoded repr of a list
                try:
                    # transform to a json like array.
                    guids = json.loads(uninstall_key_str.replace("[u'", "['").replace(", u'", ',"').replace("'", '"'))
                except:
                    logger.warning('Bad uninstallkey list format : %s' % uninstall_key_str)
                    guids = uninstall_key_str
            elif uninstall_key_str[0] in ["'", '"']:
                # simple python string, removes quotes
                guids = uninstall_key_str[1:-1]
            else:
                try:
                    # normal json encoded list
                    guids = ujson.loads(uninstall_key_str)
                except:
                    guids = uninstall_key_str

            if isinstance(guids, str):
                guids = [guids]
            return guids
        else:
            return []

    def encode_value(value,key):
        if key == 'uninstall_key':
            value = _get_uninstallkeylist(value)
        elif key in ('depends','conflicts','impacted_process'):
            value =  ensure_list(value)
        elif isinstance(value, str):
            value = value.replace('\x00', ' ')
        return value

    installed_packages = data.get('installed_packages', data.get('packages', None))

    # new delta format: the list of all install_id is transmitted with the delta
    # old format: all packages install status record are transmitted each time
    all_ids = data.get('installed_packages_ids',[])

    # we will merge pending packages install/remove actions from  last_update_status json data
    last_update_status = data.get('last_update_status', None)
    if last_update_status is None:
        last_update_status = Hosts.select(Hosts.last_update_status).where(Hosts.uuid == uuid).tuples().first()[0]

    if not last_update_status:
        last_update_status = {}

    pending = last_update_status.get('pending', {})
    missing = [package_version_from_prequest(pr) for pr in ((pending.get('install', []) or []) + (pending.get('additional', []) or []))]
    upgrades = [package_version_from_prequest(pr) for pr in pending.get('upgrade', [])] or []
    removes = [package_version_from_prequest(pr)[0] for pr in pending.get('remove', [])] or []
    errors = [package_version_from_prequest(pr) for pr in last_update_status.get('errors', [])] or []

    fake_install_id=-100000

    def prepare_package_rec(package):
        """change package dict from inventory into record suitable for database update/insert
        """
        nonlocal fake_install_id
        if 'id' in package:
            package['install_id'] = package['id']
            del package['id']
        else:
            # Old behaviour. Id was not sent by client. We use a fake negative
            fake_install_id -= 1
            package['install_id'] = fake_install_id

        package['host_id'] = uuid
        # csv str on the client, Array on the server
        package['updated_on'] = datetime.datetime.utcnow()
        if not package.get('package_uuid'):
            package['package_uuid'] = 'fb-%s' % (hashlib.sha256('-'.join([
                str(package.get('package') or ''),
                str(package.get('version') or ''),
                str(package.get('architecture') or ''),
                str(package.get('locale') or ''),
                str(package.get('maturity') or '')]).encode('utf8')).hexdigest(),)

        # merge update status from Hosts.last_update_status
        pv = (package['package'], package['version'])
        if pv in errors:
            package['install_status'] = 'ERROR'
        elif package['package'] in [p[0] for p in upgrades]:
            p = [p for p in upgrades if p[0] == package['package']][0]
            if PackageVersion(p[1]) > PackageVersion(package['version']):
                package['install_status'] = 'NEED-UPGRADE'
            elif PackageVersion(p[1]) < PackageVersion(package['version']):
                package['install_status'] = 'NEED-DOWNGRADE'
        elif package['package'] in removes:
            package['install_status'] = 'NEED-REMOVE'



    if installed_packages is not None:
        # delta is sent as a list of list with first row as columns header.
        if len(installed_packages)>0 and isinstance(installed_packages[0],list):
            # list of lists
            installed_packages = rows_to_dicts(installed_packages)

        # get all field names except id

        valid_fields =  ['host_id', 'install_id', 'package_uuid', 'name', 'package',
                'version', 'architecture', 'locale', 'maturity', 'section', 'priority',
                'signer', 'signer_fingerprint', 'signature_date', 'description', 'impacted_process',
                'install_status', 'install_date', 'install_output', 'install_params', 'uninstall_key',
                'explicit_by', 'repo_url', 'depends', 'conflicts', 'created_on', 'created_by', 'updated_on', 'updated_by']

        sql_insert = """\
              insert into hostpackagesstatus(%s)
              values (%s)
              on conflict (host_id,install_id) do update
                 set %s
                 where %s
              """ % ( ','.join(valid_fields),
                      ','.join(['%s'] * len(valid_fields)),
                      ','.join(['%s=excluded.%s' % (f,f) for f in valid_fields]),
                      ' or '.join(['hostpackagesstatus.%s is distinct from excluded.%s' % (f,f) for f in valid_fields]),
                      )

        updates_count = 0
        for package_row in installed_packages:
            prepare_package_rec(package_row)
            all_ids.append(package_row['install_id'])

            # build a value list
            rec = []
            for k in valid_fields:
                v = package_row.get(k,None)
                rec.append(encode_value(v,k))

            insert_query = wapt_db.execute_sql(sql_insert, rec)
            updates_count += insert_query.rowcount

        logger.info('packages status updated: %s' % updates_count)
        applied_status_hashes['installed_packages'] = data.get('status_hashes', {}).get('installed_packages')

        # old packages status format
        if isinstance(applied_status_hashes['installed_packages'],str):
            del_query = HostPackagesStatus.delete().where((HostPackagesStatus.host==uuid) & ~HostPackagesStatus.install_id.in_(all_ids)).execute()
            logger.info('packages status removed: %s' % del_query)

    if 'installed_packages_ids' in data.get('status_hashes', {}):
        all_ids = data.get('installed_packages_ids',[])
        del_query = HostPackagesStatus.delete().where((HostPackagesStatus.host==uuid) & ~HostPackagesStatus.install_id.in_(all_ids)).execute()
        logger.info('packages status removed: %s' % del_query)
        applied_status_hashes['installed_packages_ids'] = data.get('status_hashes', {}).get('installed_packages_ids')

    if installed_packages is None:
        # merge last_update_status in current records
        # errors
        for pv in errors:
            HostPackagesStatus.update(install_status='ERROR').where(
                (HostPackagesStatus.host == uuid) & (HostPackagesStatus.package == pv[0].encode('utf8')) & (HostPackagesStatus.version == pv[1].encode('utf8'))).execute()
        # removes
        for pv in removes:
            HostPackagesStatus.update(install_status='NEED-REMOVE').where(
                (HostPackagesStatus.host == uuid) & (HostPackagesStatus.package == pv.encode('utf8'))).execute()
        # upgrades
        for pv in upgrades:
            HostPackagesStatus.update(install_status='NEED-UPGRADE').where(
                (HostPackagesStatus.host == uuid) & (HostPackagesStatus.package == pv[0].encode('utf8')) & (HostPackagesStatus.version != pv[1].encode('utf8'))).execute()

    # add missing
    # remove previous temporary records -1000000 .. 0, but keep fake install_id from old clients -inf .. -100000
    tmp_install_id=0
    HostPackagesStatus.delete().where((HostPackagesStatus.host==uuid) & HostPackagesStatus.install_id.between(-100000,0) ).execute()
    # add fictive records
    for pv in missing+upgrades:
        (ps, _created) = HostPackagesStatus.get_or_create(
            host=uuid,
            package_uuid='fb-%s' % (hashlib.sha256(pv[0].encode('utf8')).hexdigest()),
        )
        ps.package = pv[0]
        ps.version = pv[1]
        if _created:
            tmp_install_id -= 1
            ps.install_id = tmp_install_id
            ps.created_on = isodate2datetime(last_update_status['date'])
        ps.install_status = 'NEED-INSTALL'
        ps.save()


def create_default_users(server_config={}):
    # be sure to have at least admin
    if not server_config:
        server_config = {}
    admin_user = server_config.get('wapt_user','admin')
    with wapt_db.atomic():
        (admin, created) = WaptUsers.get_or_create(name=admin_user)
        if created or admin.user_fingerprint is None:
            admin.user_fingerprint = 'admin'
            admin.save()

        (admin_acls, created) = WaptUserAcls.get_or_create(username=admin_user)
        if created or (not admin_acls.acls) or (not admin_acls.acls.get('admin')):
            admin_acls.acls = {'admin':True}
            admin_acls.save()

def update_installed_softwares(uuid, data, applied_status_hashes):
    """Stores softwares json data into separate HostSoftwares table

    Args:
        uuid (str) : unique ID of host
        data (dict): data from host

    Returns:
        None
    """
    installed_softwares = data.get('installed_softwares', data.get('softwares', None))
    if installed_softwares is not None:

        def encode_value(value,k):
            if isinstance(value, str) and k in ('key','name','version','uninstallstring','publisher')  and '\x00' in value:
                value = value.replace('\x00', ' ')
            return value

        header = [f.column.name for f in HostSoftwares._meta.fields.values()][1:] # we exclude the 'id' DB primary key

        '''
        sql_insert = """\
              insert into hostwsus(%s)
              values (%s)
              on conflict (host_id,software_id) do update
                 set %s
              """ % (','.join(header),','.join(['%s'] * len(header)),
                      ','.join(['%s=excluded.%s' % (f,f) for f in header]))
        '''

        sql_insert = """\
              insert into hostsoftwares(host_id,software_id,name,version,publisher,key,system_component,uninstall_string,install_date,install_location,created_on,created_by,updated_on,updated_by)
              values (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)
              on conflict (host_id,software_id) do update
                 set name=excluded.name, version=excluded.version,publisher=excluded.publisher,
                     key=excluded.key,uninstall_string=excluded.uninstall_string,install_date=excluded.install_date,
                     install_location=excluded.install_location,updated_on=excluded.updated_on,updated_by=excluded.updated_by
                 where
                     hostsoftwares.name is distinct from excluded.name or
                     hostsoftwares.version is distinct from excluded.version or
                     hostsoftwares.publisher is distinct from excluded.publisher or
                     hostsoftwares.key is distinct from excluded.key or
                     hostsoftwares.uninstall_string is distinct from excluded.uninstall_string or
                     hostsoftwares.install_date is distinct from excluded.install_date or
                     hostsoftwares.install_location is distinct from excluded.install_location or
                     hostsoftwares.updated_on is distinct from excluded.updated_on or
                     hostsoftwares.updated_by is distinct from excluded.updated_by
        """
        all_ids = []
        for software in installed_softwares:
            software['host_id'] = uuid
            software['updated_on'] = datetime.datetime.utcnow()
            if not 'software_id' in software:
                software['software_id'] = software['name'] + software['version'] + {False:'##win32',True:'',None:''}.get(software.get('win64'))
            all_ids.append(software['software_id'])

            # build a values list from dict
            # filter out all unknown fields from json data for the SQL insert
            rec = []
            for k in header:
                v = software.get(k,None)
                rec.append(encode_value(v,k))

            wapt_db.execute_sql(sql_insert, rec)

        # remove disapeared
        HostSoftwares.delete().where((HostSoftwares.host == uuid) & ~HostSoftwares.software_id.in_(all_ids)).execute()

        applied_status_hashes['installed_softwares'] = data.get('status_hashes', {}).get('installed_softwares')


def update_host_audit_data(uuid, data, applied_status_hashes):
    """

    """
    audit_data = data.get('audit_data',None)
    if audit_data is not None:
        audit_data[0].extend(['host_id','created_on'])
        # for backward compatibility. But now, value_type should be only json
        if 'value_type' in audit_data[0]:
            value_type_index = audit_data[0].index('value_type')
        else:
            value_type_index = -1


        fields = [fn if fn != 'id' else 'value_id' for fn in audit_data[0] if fn != 'value_type']
        value_index = fields.index('value')

        for row in audit_data[1:]:
            if value_type_index >=0:
                del row[value_type_index]
            row.extend([uuid,datetime.datetime.utcnow()])
            if isinstance(row[value_index],str) and chr(0) in row[value_index]:
                row[value_index] = row[value_index].replace('\x00', ' ')

        # bulk insert
        HostAuditData.insert_many(audit_data[1:],fields ).on_conflict('IGNORE').execute()
        # remove obsolete
        HostAuditData.delete().where((HostAuditData.host == uuid) & (HostAuditData.expiration_date != '') & (HostAuditData.expiration_date <= datetime2isodate(datetime.datetime.utcnow())) ).execute()
        applied_status_hashes['audit_data'] = data.get('status_hashes', {}).get('audit_data')

        last_audit_data_server_date = HostAuditData.select(fn.MAX(HostAuditData.value_date)).where(HostAuditData.host_id==uuid).tuples().first()
        if last_audit_data_server_date:
            return last_audit_data_server_date[0]
        else:
            return None

def update_packages_audit_status(uuid, data, applied_status_hashes):
    """takes a list of records (list) out of data dict. Key is packages_audit_status.
    First record is header (list of fieldnames).
    Next records are list of values.
    """
    audit_data = data.get('packages_audit_status',None)
    if audit_data is not None:
        fields = audit_data[0]  # first field is id and is not used.
        valid_fields =  ('last_audit_on','next_audit_on','last_audit_output','last_audit_status')

        update_query = '''update hostpackagesstatus set %s,updated_on=%%s where host_id=%%s and package_uuid=%%s''' % (
                ','.join("%s=%%s" % f for f in fields if f in valid_fields),
                )

        for row in audit_data[1:]:
            rec = []
            package_uuid = None
            # build a list of values to bind to query
            for i in range(len(fields)):
                if fields[i] == 'package_uuid':
                    package_uuid = row[i]
                elif fields[i] == 'last_audit_output' and isinstance(row[i],str) and chr(0) in row[i]:
                    rec.append(row[i].replace('\x00', ' '))
                elif fields[i] in valid_fields:
                    rec.append(row[i])
            #HostPackagesStatus.update(rec).where(HostPackagesStatus.host_id==uuid & HostPackagesStatus.package_uuid==rec['pac']).execute()
            if package_uuid:
                wapt_db.execute_sql(update_query, rec+[datetime.datetime.utcnow(), uuid, package_uuid])

        applied_status_hashes['packages_audit_status'] = data.get('status_hashes', {}).get('packages_audit_status')

        last_packages_audit_status_server_date = HostPackagesStatus.select(fn.MAX(HostPackagesStatus.last_audit_on)).where(HostPackagesStatus.host_id==uuid).tuples().first()
        if last_packages_audit_status_server_date:
            return last_packages_audit_status_server_date[0]
        else:
            return None

def update_waptwua(uuid, data, applied_status_hashes):
    """Stores discovered windows update into WindowsUpdates table and
    links between host and updates into HostWsus

    Args:
        uuid (str) : unique ID of host
        data (dict): data from host

    Returns:
        None
    """
    def encode_value(value,key):
        if isinstance(value, str):
            value = value.replace('\x00', ' ')
        elif key in ('history',):  # bson fields
            value = jsondump(value)
        return value

    timings = {}

    if 'waptwua_updates' in data:
        with Timeit('Timing update_waptwua.waptwua_updates',timings):

            header = [f.column.name for f in WsusUpdates._meta.fields.values()]

            sql_insert = """\
                  insert into wsusupdates(%s)
                  values (%s)
                  on conflict (update_id) do update
                     set %s
                  """ % (','.join(header),','.join(['%s'] * len(header)),
                          ','.join(['%s=excluded.%s' % (f,f if f != 'updated_on' else 'created_on')
                                        for f in header if f != 'created_on'])) # put date into updated_on only if updated
            # TODO : append new URL to the list of download_urls
            # set download_urls = ARRAY( SELECT DISTINCT url FROM unnest(excluded.download_urls) as a(url) )
            # and reset   update['downloaded_on'] = None
            # or better : add a hash of languages to the primary key...

            for wu in data['waptwua_updates']:
                wu['created_on'] = datetime.datetime.utcnow()

                # build a value list
                rec = []
                for k in header:
                    v = wu.get(k,None)
                    rec.append(encode_value(v,k))

                wapt_db.execute_sql(sql_insert, rec)

            """
                # if win update has already been registered, we don't update it, but simply ignore the insert
                # we should append missing URLS... to do in a dedicated PG sql proc will be better.
                for update in windows_updates:
                    download_urls = WsusUpdates.select(WsusUpdates.update_id, WsusUpdates.download_urls).where(WsusUpdates.update_id == update['update_id']).first()
                    if not download_urls:
                        WsusUpdates.insert(update).on_conflict('IGNORE').execute()
                    # elif download_urls.download_urls != update['download_urls']:
                    #    new_urls = list(set(download_urls.download_urls + update['download_urls']))
                    #    WsusUpdates.update(download_urls=new_urls).where(WsusUpdates.update_id == update['update_id']).execute()
                    else:
                        # append new urls when different locales for example
                        if download_urls.download_urls != update['download_urls']:
                            new_urls = list(set(download_urls.download_urls + update['download_urls']))
                            update['download_urls'] = new_urls
                            update['downloaded_on'] = None
                        WsusUpdates.update(**update).where(WsusUpdates.update_id == update['update_id']).execute()
            """

            applied_status_hashes['waptwua_updates'] = data.get('status_hashes', {}).get('waptwua_updates')

    if 'waptwua_updates_localstatus' in data:
        with Timeit('Timing update_waptwua.host_wsus',timings):
            waptwua_updates_localstatus = data['waptwua_updates_localstatus']

            # get all field names except id
            header = [f.column.name for f in HostWsus._meta.fields.values()][1:]

            sql_insert = """\
                  insert into hostwsus(%s)
                  values (%s)
                  on conflict (host_id,update_id) do update
                     set %s
                  """ % (','.join(header),','.join(['%s'] * len(header)),
                          ','.join(['%s=excluded.%s' % (f,f) for f in header]))

            # update existing
            all_update_ids = []
            for new_rec in waptwua_updates_localstatus:
                all_update_ids.append(new_rec['update_id'])
                new_rec['host_id'] = uuid
                new_rec['created_on'] = datetime.datetime.utcnow()
                if not 'install_date' in new_rec:
                    # retrieve sucessful install date from history
                    if 'history' in new_rec:
                        success_install = [h.get('date') for h in new_rec['history'] if h['result_code']=='Succeeded' and h['operation'] == 1]
                        if success_install:
                            new_rec['install_date'] = success_install[-1]
                    else:
                        new_rec['install_date'] = None

                # build a value list
                rec = []
                for k in header:
                    v = new_rec.get(k,None)
                    rec.append(encode_value(v,k))

                wapt_db.execute_sql(sql_insert, rec)

            HostWsus.delete().where((HostWsus.host == uuid) & (~ HostWsus.update_id.in_(all_update_ids))).execute()

            applied_status_hashes['waptwua_updates_localstatus'] = data.get('status_hashes', {}).get('waptwua_updates_localstatus')
    logger.info(repr(timings))

def update_known_ssl_certificates(data, server_conf):
    """Append supplied certificates_pems to the local ssl known certificates directory
    (<waptrepodir>/ssl/<sha256 of cert>.crt)

    Args:
        server_conf (dict)
        certificates (list of X509 pem encoded certificates)

    Returns:
        None
    """
    if server_conf is None:
        return False

    certificates = data.get('authorized_certificates', None)
    fingerprints = data.get('wapt_status', {}).get('authorized_certificates_sha256', None)
    if not certificates:
        return False

    ssl_dir = server_conf['known_certificates_folder']
    if not os.path.isdir(ssl_dir):
        os.makedirs(ssl_dir)

    # optimization... if all fingerprints are already in ssl dir, skip whole proc
    if fingerprints is not None:
        skip = True
        for fg in fingerprints:
            cert_fn = os.path.join(ssl_dir, '%s.crt' % fg)
            if not os.path.isfile(cert_fn):
                skip = False
                break

        if skip:
            return True

    for pem in certificates:
        certs = SSLCABundle()
        certs.add_certificates_from_pem(pem)
        for cert in certs.certificates():
            cert_fn = os.path.join(ssl_dir, '%s.crt' % cert.fingerprint)
            if not os.path.isfile(cert_fn):
                with open(cert_fn, 'wb') as f:
                    f.write(cert.as_pem())

    return True

def get_keys(data_table,keyname):
    """Return list of all key in data
    first row of data are the field names

    Returns:
        list
    """

    key_idx = data_table[0].index(keyname)
    return [row[key_idx] for row in data_table[1:]]

def row_to_dict(fields,row):
    """

    Args:
        fields (list of str)
        row (list): of cells

    Returns:
        dict
    """
    i = 0
    result = {}
    for fieldname in fields:
        result[fieldname] = row[i]
        i+=1
    return result

def rows_to_dicts(rows):
    header = rows[0]
    result = []
    for row in rows[1:]:
        result.append(row_to_dict(header,row))
    return result


class Timeit:
    def __init__(self,title='',store=None):
        self.title = title
        self.store = store
    def __enter__(self):
        self.t1 = time.time()
    def __exit__(self,exc_type, exc_value, traceback):
        timing = time.time()-self.t1
        logger.debug('%s timing: %s s' % (self.title,timing))
        if self.store is not None:
            self.store[self.title] = timing


def get_server_uuid():
    """Returns this server UUID as configured in configuration file waptserver.ini
    """
    return app.conf.get('server_uuid', None)

def update_host_config_overview(uuid, data):
    if not data.get('configurations', None):
        return False

    conf = data['configurations']['config']
    priorities = data['configurations']['priorities']

    host_conf = HostConfigs.select().where(HostConfigs.host == uuid).first()
    if not host_conf:
        HostConfigs.insert(host=uuid,config=conf, priorities=priorities).execute()
    elif host_conf.config != conf or host_conf.priorities != priorities:
        host_conf.config = conf
        host_conf.priorities = priorities
        host_conf.save()
    return True

def update_host_data(data, server_conf=None):
    """Helper function to insert or update host data in db

    Args :
        data (dict) : data to push in DB with at least 'uuid' key
                      if uuid key already exists, update the data
                      else insert
                      only keys in data are pushed to DB.
                      Other data (fields) are left untouched

        server_conf(dict) : server configuraation dict (to get wapt_folder)

    Returns:
        dict : with uuid,computer_fqdn,host_info from db after update
    """
    migrate_map_13_14 = {
        'packages': None,
        'installed_packages': None,
        'softwares': None,
        'installed_softwares': None,

        'update_status': 'last_update_status',
        'host': 'host_info',
        'wapt': 'wapt_status',
    }

    uuid = data['uuid']
    with wapt_db.atomic() as trans:
        try:
            timings = {}
            supplied_hashes = data.get('status_hashes', {})
            applied_status_hashes = {}
            existing = Hosts.select(Hosts.uuid, Hosts.computer_fqdn, Hosts.status_hashes).where(Hosts.uuid == uuid).first()
            with Timeit('Store Hosts',timings):

                def update_keys():
                    for k in data.keys():
                        # supplied status_hashes is a subset of known hashes
                        if (not existing) or (k != 'status_hashes'):
                            # manage field renaming between 1.3 and >= 1.4
                            target_key = migrate_map_13_14.get(k, k)
                            if target_key and hasattr(updhost, target_key):
                                set_host_field(updhost, target_key, data[k])
                                if k in supplied_hashes:
                                    # logger.info('set data: %s , size: %s' % (target_key,len(jsondump(data[k]))))
                                    applied_status_hashes[k] = supplied_hashes[k]


                if not existing:
                    logger.debug('Inserting new host %s with fields %s' % (uuid, list(data.keys())))
                    # wapt update_status packages softwares host
                    updhost = Hosts()

                    update_keys()

                    updhost.status_hashes = applied_status_hashes
                    updhost.server_uuid = get_server_uuid()
                    updhost.save(force_insert=True)
                else:
                    logger.debug('Updating %s for fields %s' % (uuid, list(data.keys())))

                    updhost = Hosts.get(uuid=uuid)
                    if not updhost.status_hashes:
                        updhost.status_hashes = {}

                    update_keys()

                # update sub keys of host_info
                host_info = updhost.host_info
                for k in data.keys():
                    if k.startswith('host_info.'):
                        (_,key) = k.split('.',1)
                        host_info[key] = data[k]
                        applied_status_hashes[k] = supplied_hashes[k]

                # merge already known data with new set
                updhost.status_hashes.update(applied_status_hashes)
                updhost.server_uuid = get_server_uuid()

            # separate tables
            # we are tolerant on errors here as we don't know exactly if client send good encoded data
            # but we still want to get host in table
            with Timeit('Store installed_softwares',timings):
                with wapt_db.atomic() as translocal:
                    try:
                        update_installed_softwares(uuid, data, applied_status_hashes)
                    except Exception as e:
                        # be tolerant
                        translocal.rollback()
                        logger.critical('Unable to update installed_softwares for %s: %s' % (uuid, ensure_unicode(e)))
                        logger.debug(traceback.format_exc())

            with Timeit('Store installed_packages',timings):
                with wapt_db.atomic() as translocal:
                    try:
                        update_installed_packages(uuid, data, applied_status_hashes)
                    except Exception as e:
                        # be tolerant
                        translocal.rollback()
                        logger.critical('Unable to update installed_packages for %s: %s' % (uuid, ensure_unicode(e)))
                        logger.debug(traceback.format_exc())

            with Timeit('Store packages_audit_status',timings):
                with wapt_db.atomic() as translocal:
                    try:
                        last_packages_audit_status_server_date = update_packages_audit_status(uuid, data, applied_status_hashes)
                    except Exception as e:
                        # be tolerant
                        last_packages_audit_status_server_date = None
                        translocal.rollback()
                        logger.critical('Unable to update packages_audit_status for %s: %s' % (uuid, ensure_unicode(e)))
                        logger.debug(traceback.format_exc())

            with Timeit('Store waptwua',timings):
                with wapt_db.atomic() as translocal:
                    try:
                        # update waptwua state tables
                        update_waptwua(uuid, data, applied_status_hashes)
                    except Exception as e:
                        # be tolerant
                        translocal.rollback()
                        logger.critical('Unable to update wuauserv_status or waptwua_status for %s: %s' % (uuid, ensure_unicode(e)))
                        logger.debug(traceback.format_exc())

            with Timeit('Store audit_data',timings):
                with wapt_db.atomic() as translocal:
                    try:
                        last_audit_data_server_date = update_host_audit_data(uuid, data, applied_status_hashes)
                    except Exception as e:
                        last_audit_data_server_date = None
                        # be tolerant
                        translocal.rollback()
                        logger.critical('Unable to update audit_data for %s: %s' % (uuid, ensure_unicode(e)))
                        logger.debug(traceback.format_exc())

            with Timeit('Store known_ssl_certificates',timings):
                # extract X509 certificates PEM and store them in repo wapt/ssl
                if update_known_ssl_certificates(data, server_conf):
                    if 'authorized_certificates' in supplied_hashes:
                        updhost.status_hashes['authorized_certificates'] = supplied_hashes['authorized_certificates']

            with Timeit('Update host config', timings):
                if update_host_config_overview(uuid, data):
                    if 'configurations' in supplied_hashes:
                        updhost.status_hashes['configurations'] = supplied_hashes['configurations']


            with Timeit('Update update_file_tree',timings):
                if server_conf['remote_repo_diff']:
                    try:
                        from waptserver.repositories import update_file_tree_of_files
                        update_file_tree_of_files(username='TASK_UPDATE_HOST_DATA')
                    except Exception as e:
                        logger.critical("Something went wrong with diff repo : can't launch: %s" % e)
                        logger.debug(traceback.format_exc())

            with Timeit('Store status_hashes',timings):
                # merge new known hashes for properly applied data, for next round
                updhost.status_hashes.update(applied_status_hashes)
                updhost.save()
                logger.info('Applied data for host %s: %s' % (uuid, list(applied_status_hashes.keys())))

            # returns actual registered fqdn and status hashes so that host can omit to send some data next time if they have not changed
            result_query = Hosts.select(Hosts.uuid, Hosts.computer_fqdn, Hosts.status_hashes)
            logger.info('update_host_data %s: %s' % (uuid, repr(timings)))
            result = result_query.where(Hosts.uuid == uuid).dicts().first()
            if last_audit_data_server_date:
                result['last_audit_data_server_date'] = last_audit_data_server_date
            if last_packages_audit_status_server_date:
                result['last_packages_audit_status_server_date'] = last_packages_audit_status_server_date
            return result

        except Exception as e:
            trans.rollback()
            logger.critical('Error updating data for %s : %s' % (uuid, ensure_unicode(e)))
            logger.debug(traceback.format_exc())
            raise e


@pre_save(sender=Hosts)
def wapthosts_json(model_class, instance, created):
    """Stores in plain table fields data from json"""
    # extract data from json into plain table fields
    if (created or Hosts.host_info in instance.dirty_fields) and instance.host_info:
        def extract_ou(host_info):
            dn = host_info.get('computer_ad_dn', None)
            if dn:
                parts = dn.split(',', 1)
                if len(parts) >= 2:
                    return parts[1]
                else:
                    return ''
            else:
                return None

        extractmap = [
            ['computer_fqdn', 'computer_fqdn'],
            ['computer_name', 'computer_name'],
            ['description', 'description'],
            ['manufacturer', 'system_manufacturer'],
            ['productname', 'system_productname'],
            ['os_name', 'os_name'],
            ['platform', 'platform'],
            ['repositories', 'repositories'],
            ['os_version', ('os_version', 'os_name')],
            ['connected_ips', 'connected_ips'],
            ['connected_users', ('connected_users', 'current_user')],
            ['last_logged_on_user', 'last_logged_on_user'],
            ['mac_addresses', 'mac'],
            ['dnsdomain', ('dnsdomain', 'dns_domain')],
            ['gateways', ['gateways', 'default_gateways']],
            ['computer_ad_site', 'computer_ad_site'],
            ['computer_ad_ou', extract_ou],
            ['computer_ad_groups', 'computer_ad_groups'],
            ['networking', 'host_networking'],
        ]

        for field, attribute in extractmap:
            if callable(attribute):
                setattr(instance, field, attribute(instance.host_info))
            else:
                setattr(instance, field, dictgetpath(instance.host_info, attribute))

        instance.os_architecture = 'x64' if instance.host_info.get('win64', '?') else 'x86'

    if (created or Hosts.dmi in instance.dirty_fields) and instance.dmi:
        extractmap = [
            ['serialnr', ('System_Information.Serial_Number','Chassis_Information.Serial_Number','System_Information.IdentifyingNumber')],
            ['computer_type', 'Chassis_Information.Type'],
        ]
        for field, attribute in extractmap:
            if callable(attribute):
                setattr(instance, field, attribute(instance.dmi))
            else:
                setattr(instance, field, dictgetpath(instance.dmi, attribute))

    if (created or Hosts.host_metrics in instance.dirty_fields) and instance.host_metrics:
        extractmap = [
            ['connected_users', ('logged_in_users', 'connected_users')],
            ['last_logged_on_user', 'last_logged_on_user'],
        ]
        for field, attribute in extractmap:
            if callable(attribute):
                setattr(instance, field, attribute(instance.host_metrics))
            else:
                setattr(instance, field, dictgetpath(instance.host_metrics, attribute))

    # extract list for fast query.
    if (created or Hosts.host_capabilities in instance.dirty_fields) and instance.host_capabilities:
        instance.authorized_certificates_sha256 = dictgetpath(instance.host_capabilities, 'packages_trusted_ca_fingerprints')
        instance.wapt_version = dictgetpath(instance.host_capabilities, 'wapt_version')
        instance.host_packages_filter = make_packages_filter_string(instance.host_capabilities)

    if instance.host_networking and Hosts.host_networking in instance.dirty_fields:
        instance.connected_ips = instance.host_networking.get('connected_ips',instance.connected_ips)
        instance.mac_addresses = instance.host_networking.get('mac',instance.mac_addresses)
        instance.gateways = instance.host_networking.get('gateways',instance.gateways)
        instance.dnsdomain = instance.host_networking.get('dnsdomain',instance.dnsdomain)

    # update host update status based on update_status json data or packages collection
    if not instance.host_status or created or Hosts.last_update_status in instance.dirty_fields:
        instance.host_status = None
        if instance.last_update_status:
            if instance.last_update_status.get('errors', []):
                instance.host_status = 'ERROR'
            elif instance.last_update_status.get('upgrades', []):
                instance.host_status = 'TO-UPGRADE'
        if not instance.host_status:
            instance.host_status = 'OK'

    if instance.last_update_status and Hosts.last_update_status in instance.dirty_fields:
        instance.audit_status = instance.last_update_status.get('audit_status',instance.audit_status)


class ColumnDef(object):
    """Holds definitin of column for updatable remote GUI table
    """

    def __init__(self, field, in_update=None, in_where=None, in_key=None, calc_field_name=None):
        self.field = field
        if in_update is not None:
            self.in_update = in_update
        else:
            self.in_update = not isinstance(field, ForeignKeyField) and not isinstance(field, Function)

        self.in_where = None
        if in_where is not None:
            self.in_where = in_where

        if in_where is None:
            self.in_where = not isinstance(field, ForeignKeyField) and not isinstance(field, Function)

        self.in_key = field and field.primary_key
        self.visible = False
        self.default_width = None
        self.calc_field_name = calc_field_name
        if self.calc_field_name is None and self.field is not None:
            self.calc_field_name = getattr(self.field, '_alias', self.field.name)

    def as_metadata(self):
        result = dict()
        if isinstance(self.field, Function):
            result = {'name': getattr(self.field, '_alias', self.field.name), 'org_name': self.field.name, 'type': self.field._node_type}
        else:
            if self.field:
                result = {'name': getattr(self.field, '_alias', self.field.name),
                          'field_name': self.field.name,
                          'type': self.field.field_type,
                          'table_name': self.field.model._meta.table_name,
                          }
                attlist = ('primary_key', 'description', 'help_text', 'choice',
                           'default', 'sequence', 'max_length')
                for att in attlist:
                    if hasattr(self.field, att):
                        value = getattr(self.field, att)
                        if value is not None and not isinstance(value, Function):
                            if callable(value):
                                result[att] = value()
                            else:
                                result[att] = value

            elif self.calc_field_name:
                result = {'name': self.calc_field_name,
                          'field_name': self.calc_field_name,
                          'type': 'calc',
                          'table_name': None,
                          }

        for att in ('visible', 'default_width'):
            value = getattr(self, att)
            if value is not None:
                if callable(value):
                    result[att] = value()
                else:
                    result[att] = value

        result['required'] = self.field and not self.field.is_null()
        return result

    def to_client(self, data):
        """Return a serialization of data, suitable for client application"""
        if isinstance(data, list):
            return json.dumps(data)
        else:
            return data

    def from_client(self, data):
        """Return db suitable value from a serialization from client application"""
        if isinstance(self.field, ArrayField):
            return json.loads(data)
        else:
            return data


class TableProvider(object):
    """Updatable dataset provider based on a list of column defs and
    a where clause.

    >>>
    """

    def __init__(self, query=None, model=None, columns=None, where=None):
        self.query = query
        self.model = model
        self._columns = columns
        self.where = where
        self._columns_idx = None

    @property
    def columns(self):
        if not self._columns:
            self._columns = []
            if self.model is not None:
                for field in self.model._meta.sorted_fields:
                    column = ColumnDef(field)
                    self._columns.append(column)
            elif self.query is not None:
                cursor = self.query.execute()
                cursor._initialize_columns()
                col = 0
                for field in cursor.fields:
                    column = ColumnDef(field, calc_field_name=cursor.columns[col])
                    col += 1
                    self._columns.append(column)
        return self._columns

    def get_data(self, start=0, count=None):
        """Build query, retrieve rows"""

        query = self.query
        if query is None and self.model is not None:
            query = self.model.select(* [f.field for f in self.columns])
            if self.where:
                query = query.where(self.where)

        rows = []
        for row in query.dicts():
            rows.append([column.to_client(row[column.calc_field_name]) for column in self.columns])

        return dict(
            metadata=[c.as_metadata() for c in self.columns],
            rows=rows
        )

    def column_by_name(self, name):
        """Return ColumnDef for field name"""
        if self._columns_idx is None:
            self._columns_idx = dict([(c.field.column_name, c) for c in self.columns])
        return self._columns_idx.get(name, None)

    def _where_from_values(self, old_values={}):
        """Return a where clause from old and new dict for update and delete"""
        result = None
        for (column_name, column_value) in old_values.items():
            column = self.column_by_name(column_name)
            if column:
                if column.in_key or column.in_where:
                    if result is None:
                        result = column.field == column_value
                    else:
                        result = result & column.field == column_value
        return result

    def _record_values_from_values(self, new_values={}):
        """Return a dict for the insert/update into database from supplied dict
        filtering out non updateable values.
        """
        result = {}
        for (column_name, column_value) in new_values.items():
            column = self.column_by_name(column_name)
            if column:
                if column.in_update:
                    result[column_name] = column.from_client(column_value)
        return result

    def _values_from_record_values(self, values):
        """Return a dict
        """
        result = {}
        for (column_name, column_value) in values.items():
            column = self.column_by_name(column_name)
            if column:
                result[column_name] = column.to_client(column_value)
        return result

    def _update_set_from_values(self, new_values={}):
        """Return a dict for the insert/update into database from supplkied dict
        filtering out non updateable values.
        """
        result = {}
        for (column_name, column_value) in new_values.items():
            column = self.column_by_name(column_name)
            if column:
                if column.in_update:
                    result[column.field] = column.from_client(column_value)
        return result

    def apply_updates(self, delta):
        """Build update queries from delta

        Args:
            delta (list): list of (update_type,old_data,new_data)
                update_type (str) = ('insert','update','delete')
                old_data (dict)  = empty dict for insert, list of old values for update / delete.
                           must include in_key, in_where, and updated fields with in_update flag
                new_data (dict) = dict for updated / inserted data, empty for delete

        """
        result = []
        with self.model._meta.database.atomic():
            for (update_type, old, new) in delta:
                # translates old / new value to
                if update_type == 'insert':
                    query = self.model.insert(self._update_set_from_values(new))
                elif update_type == 'update':
                    old_db_values = self._record_values_from_values(old)
                    query = self.model.update(self._update_set_from_values(new)).where(self._where_from_values(old_db_values))
                elif update_type == 'delete':
                    old_db_values = self._record_values_from_values(old)
                    query = self.model.delete().where(self._where_from_values(old_db_values))
                query.execute()
        return result


def get_db_version():
    with WaptDB():
        try:
            return Version(ServerAttribs.get(key='db_version').value, 4)
        except:
            wapt_db.rollback()
            return None

def get_db_need_upgrade():
    if Version(get_db_version()) < Version(DB_STRUCTURE_VERSION) :
        return True
    else:
        return False


def set_db_version(version):
    with WaptDB():
        (v, created) = ServerAttribs.get_or_create(key='db_version')
        v.value = version
        v.save()
        return Version(ServerAttribs.get(key='db_version').value, 4)


def init_db(drop=False,server_config=None):
    with WaptDB():
        try:
            wapt_db.execute_sql('CREATE EXTENSION hstore;')
        except:
            wapt_db.rollback()

        try:
            wapt_db.execute_sql('CREATE EXTENSION jsquery;')
        except:
            wapt_db.rollback()

        list_tables = [Packages,HostSoftwares,HostGroups,HostWebsocket,HostAuditData,ReportingQueries,ReportingSnapshots,
                       Normalization,WsusUpdates,HostWsus,WsusDownloadTasks,SiteRules,
                       SyncStatus,HostSyncStatus,LogsAPI,HostOSDeploy,DeployConfig,IsoConfig,
                       MessagesTemplate,HostConfigs,WaptBaseModel,ServerAttribs,Hosts,HostPackagesStatus,WaptUsers,
                       WaptUserAcls,WaptRoles]
        if drop:
            for table in reversed(list_tables):
                table.drop_table(fail_silently=True)

        try:
            wapt_db.create_tables(list_tables, safe=True)
        except Exception as e:
            wapt_db.rollback()
            print('Unable to create tables, will try to upgrade step by step instead... : %s' % (repr(e),))

        if get_db_version() is None:
            # new database install, we setup the db_version key
            set_db_version(DB_STRUCTURE_VERSION)
        elif get_db_version() != DB_STRUCTURE_VERSION:
            with wapt_db.atomic():
                upgrade_db_structure(server_config=server_config)

        create_default_users(server_config=server_config)

        return get_db_version()

def add_column_safe(opes,migrator,column):
    """Add a column add operation to migrator if column does not exist yet in actual table"""
    model = column.model
    columns = [c.name for c in wapt_db.get_columns(model._meta.table_name)]
    if not column.column_name in columns:
        opes.append(migrator.add_column(model._meta.table_name,column.column_name,column))

def drop_column_safe(opes,migrator,table_name,column_name):
    """Drio a column add operation to migrator if column exists in table"""
    columns = [c.name for c in wapt_db.get_columns(table_name)]
    if column_name in columns:
        opes.append(migrator.drop_column(table_name,column_name))

def upgrade_db_structure(server_config=None):
    """Upgrade the tables version by version"""
    from playhouse.migrate import PostgresqlMigrator, migrate

    migrator = PostgresqlMigrator(wapt_db)
    packages_table_recreated = False

    logger.info('Current DB: %s version: %s' % (wapt_db.connect_params, get_db_version()))

    if get_db_version() < "1.8.2.0":
        logger.error('Your current version is %s, Please upgrade to 1.8.2.7393 first' % get_db_version())
        raise Exception('Incompatible Version Upgrade')


    next_version = '1.8.2.1'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            HostSyncStatus.create_table(fail_silently=True)

            set_db_version(next_version)

    next_version = '1.9.0.4'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            opes = []
            columns = [c.name for c in wapt_db.get_columns('waptusers')]
            if 'user_fingerprint_sha1' in columns:
                opes.append(migrator.drop_column(WaptUsers._meta.name, 'user_fingerprint_sha1'))
            if 'user_fingerprint_sha256' in columns:
                opes.append(migrator.rename_column(WaptUsers._meta.name,'user_fingerprint_sha256','user_fingerprint'))
            migrate(*opes)

            WaptUserAcls.drop_table(safe=True)
            WaptUserAcls.create_table(safe=True)
            set_db_version(next_version)

    next_version = '1.9.0.12'
    if get_db_version() <= next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            opes = []

            add_column_safe(opes,migrator,SyncStatus.progress)
            add_column_safe(opes,migrator,HostPackagesStatus.install_id)
            add_column_safe(opes,migrator,HostSoftwares.software_id)

            add_column_safe(opes,migrator,Packages.min_os_version)
            add_column_safe(opes,migrator,Packages.max_os_version)
            add_column_safe(opes,migrator,Packages.is_available)
            add_column_safe(opes, migrator, Packages.icon_sha256sum)

            add_column_safe(opes, migrator, SiteRules.has_proxy)
            add_column_safe(opes, migrator, SiteRules.http_proxy)

            migrate(*opes)

            # be sure to have not null unique data in this new field, even if it's not the proper id
            # next update from client will fix it anyway.

            HostPackagesStatus.update(install_id = HostPackagesStatus.id).where(HostPackagesStatus.install_id.is_null()).execute()
            HostSoftwares.update(software_id = HostSoftwares.id).where(HostSoftwares.software_id.is_null()).execute()

            #opes.append(migrator.add_unique('hostwsus','host_id','update_id'))
            #opes.append(migrator.add_unique('hostpackagesstatus','host_id','install_id'))


            set_db_version(next_version)

    next_version = '2.0.1'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            LogsAPI.drop_table(safe=True)
            LogsAPI.create_table(safe=True)
            set_db_version(next_version)

    next_version = '2.0.2'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            opes = []

            add_column_safe(opes, migrator, LogsAPI.data)
            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.0.3'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            HostAuditData.create_table(safe=True)

            HostWebsocket.create_table(safe=True)

            opes = []
            add_column_safe(opes, migrator, WaptUsers.last_login_timestamp)
            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.0.4'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            # will be recreated with unique constraint
            if 'waptusers_name' in [i.name for i in wapt_db.get_indexes('waptusers')]:
                wapt_db.execute_sql('drop index waptusers_name')
            if 'waptuseracls_username' in [i.name for i in wapt_db.get_indexes('waptuseracls')]:
                wapt_db.execute_sql('drop index waptuseracls_username')

            set_db_version(next_version)

    next_version = '2.0.5'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            # size was too small
            if 'signature' in [i.name for i in wapt_db.get_columns('waptuseracls')]:
                opes.append(migrator.drop_column(WaptUserAcls._meta.name, 'signature'))
            migrate(*opes)

            opes = []
            add_column_safe(opes, migrator, WaptUserAcls.signature)
            # mandatory with current sign_claim check
            add_column_safe(opes, migrator, WaptUserAcls.signer_certificate)
            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.0.8'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            columns = [c.name for c in wapt_db.get_columns('packages')]
            opes = []
            add_column_safe(opes, migrator, Packages.icon_sha256sum)
            if 'icon_md5sum' in columns:
                opes.append(migrator.drop_column(Packages._meta.name, 'icon_md5sum'))
            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.1.0'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            opes = []

            logger.info("Fix column type for created_by and updated_by on all tables")
            for table_name in wapt_db.get_tables():
                columns = [c.name for c in wapt_db.get_columns(table_name)]
                if 'created_by' in columns:
                    opes.append(migrator.alter_column_type(table_name, 'created_by',WaptBaseModel.created_by))
                if 'updated_by' in columns:
                    opes.append(migrator.alter_column_type(table_name, 'updated_by',WaptBaseModel.updated_by))


            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.1.0.1'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            opes = []

            logger.info("Switch column type (CharField -> TextField) for normalization table")
            opes.append(migrator.alter_column_type(Normalization._meta.name, 'original_name',Normalization.original_name))
            opes.append(migrator.alter_column_type(Normalization._meta.name, 'key',Normalization.key))
            opes.append(migrator.alter_column_type(Normalization._meta.name, 'normalized_name',Normalization.normalized_name))

            migrate(*opes)
            set_db_version(next_version)


    next_version = '2.2.3.0'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            MessagesTemplate.create_table(safe=True)

            set_db_version(next_version)

    next_version = '2.2.3.7'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            HostConfigs.create_table(safe=True)
            opes = []

            migrate(*opes)

            set_db_version(next_version)


    next_version = '2.2.4.3'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []

            logger.info("Switch column type (CharField -> TextField) for reporting table")
            opes.append(migrator.alter_column_type(ReportingQueries._meta.name, 'query',ReportingQueries.query))

            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.2.4.6'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []

            drop_column_safe(opes, migrator, 'hosts','reachable')
            drop_column_safe(opes, migrator, 'hosts','listening_protocol')
            drop_column_safe(opes, migrator, 'hosts','listening_address')
            drop_column_safe(opes, migrator, 'hosts','listening_port')
            drop_column_safe(opes, migrator, 'hosts','listening_timestamp')

            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.2.4.10'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            add_column_safe(opes, migrator, ReportingQueries.as_host_uuid)
            add_column_safe(opes, migrator, ReportingQueries.as_package_uuid)
            add_column_safe(opes, migrator, ReportingQueries.as_kb_update_id)
            add_column_safe(opes, migrator, ReportingQueries.as_software_key)

            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.2.5.1'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            HostOSDeploy.drop_table(fail_silently=True)
            DeployConfig.drop_table(fail_silently=True)
            IsoConfig.drop_table(fail_silently=True)

            HostOSDeploy.create_table()
            DeployConfig.create_table()
            IsoConfig.create_table()

            set_db_version(next_version)

    next_version = '2.2.5.2'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            add_column_safe(opes, migrator, DeployConfig.ipxe_script)

            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.2.6.0'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []

            logger.info("Switch column type (TextField -> BinaryJsonField) for audit data")

            if  list(wapt_db.execute_sql("""select column_name, data_type from information_schema.columns where table_schema = 'public' AND table_name='hostauditdata' AND column_name='value' AND data_type='text' ;""")):
                wapt_db.execute_sql("""\
create or replace function is_valid_jsonb(p_json text) returns boolean
as $$
begin
  return (p_json::jsonb is not null);
exception
  when others then
     return false;
end; $$ language plpgsql immutable;
                """)
                wapt_db.execute_sql("""update hostauditdata set value = to_jsonb(value::text)::text where not is_valid_jsonb(value) and not is_valid_jsonb(replace(value,'''','"'));""")
                wapt_db.execute_sql("""update hostauditdata set value = replace(value,'''','"') where not is_valid_jsonb(value);""")
                wapt_db.execute_sql("""alter table hostauditdata alter column value type jsonb using value::jsonb;""")

            drop_column_safe(opes, migrator, 'hostauditdata','value_type')
            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.2.6.1'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            wapt_db.execute_sql("""alter table hostauditdata alter column value_id type bigint;""")
            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.2.6.2'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            add_column_safe(opes, migrator, SiteRules.signer_fingerprint)

            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.2.6.3'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            wapt_db.execute_sql("""drop table if exists hostsglpisyncstatus ;""")
            set_db_version(next_version)

    next_version = '2.3.0.0'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            # HostPackagesStatus, Packages, HostSoftwares, HostGroups, HostWebsocket, HostAuditData, ReportingSnapshots, HostWsus, LogsAPI
            wapt_db.execute_sql("ALTER TABLE hostsoftwares ALTER COLUMN id TYPE bigint USING id::bigint;")
            wapt_db.execute_sql("ALTER TABLE hostpackagesstatus ALTER COLUMN id TYPE bigint USING id::bigint;")
            wapt_db.execute_sql("ALTER TABLE hostwsus ALTER COLUMN id TYPE bigint USING id::bigint;")
            wapt_db.execute_sql("ALTER TABLE hostauditdata ALTER COLUMN id TYPE bigint USING id::bigint;")
            wapt_db.execute_sql("ALTER TABLE packages ALTER COLUMN id TYPE bigint USING id::bigint;")
            wapt_db.execute_sql("ALTER TABLE hostgroups ALTER COLUMN id TYPE bigint USING id::bigint;")
            wapt_db.execute_sql("ALTER TABLE hostwebsocket ALTER COLUMN id TYPE bigint USING id::bigint;")
            wapt_db.execute_sql("ALTER TABLE reportingsnapshots ALTER COLUMN id TYPE bigint USING id::bigint;")
            wapt_db.execute_sql("ALTER TABLE logsapi ALTER COLUMN id TYPE bigint USING id::bigint;")

            set_db_version(next_version)

    next_version = '2.3.0.1'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            add_column_safe(opes, migrator, WsusUpdates.cve_ids)

            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.3.0.3'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            add_column_safe(opes, migrator, HostOSDeploy.asset)

            migrate(*opes)
            set_db_version(next_version)

    next_version = '2.3.0.4'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            logger.info("Drop/recreate Packages table because some setups have duplicated package_uuid")
            Packages.drop_table(safe=True)
            Packages.create_table(safe=True)
            packages_table_recreated = True

            set_db_version(next_version)

    next_version = '2.3.0.5'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []

            add_column_safe(opes, migrator, HostOSDeploy.ei_cfg)
            add_column_safe(opes, migrator, HostOSDeploy.pid_txt)

            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.3.0.6'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []

            add_column_safe(opes, migrator, HostOSDeploy.diskpart_list_disk)
            add_column_safe(opes, migrator, HostOSDeploy.diskpart_list_vol)
            add_column_safe(opes, migrator, HostOSDeploy.ip_address)

            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.3.0.7'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []

            add_column_safe(opes, migrator, Hosts.host_networking)

            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.3.0.8'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            add_column_safe(opes, migrator, Packages.changelog)
            add_column_safe(opes, migrator, Packages.maintainer)
            add_column_safe(opes, migrator, Packages.md5sum)
            add_column_safe(opes, migrator, Packages.sources)
            add_column_safe(opes, migrator, Packages.size)
            add_column_safe(opes, migrator, Packages.filename)
            add_column_safe(opes, migrator, Packages.description_localized)

            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.3.0.9'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            add_column_safe(opes, migrator, WaptUserAcls.signed_attributes)
            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.4.0.0'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []
            add_column_safe(opes, migrator, HostOSDeploy.signer_fingerprint)
            add_column_safe(opes, migrator, DeployConfig.signer_fingerprint)
            add_column_safe(opes, migrator, IsoConfig.signer_fingerprint)
            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.4.2.0'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            if not packages_table_recreated:
                wapt_db.execute_sql("ALTER TABLE packages ALTER COLUMN target_os TYPE varchar[] USING string_to_array(target_os,',');")
                wapt_db.execute_sql("ALTER TABLE packages ALTER COLUMN licence TYPE varchar[] USING string_to_array(licence,',');")
                wapt_db.execute_sql("ALTER TABLE packages ALTER COLUMN locale TYPE varchar[] USING string_to_array(locale,',');")
                wapt_db.execute_sql("ALTER TABLE packages ALTER COLUMN architecture TYPE varchar[] USING string_to_array(architecture,',');")

    next_version = '2.4.2.1'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            ReportingSnapshots.drop_table()
            ReportingSnapshots.create_table()
            opes = []
            drop_column_safe(opes, migrator, 'reportingqueries','snapshot_period')
            add_column_safe(opes, migrator, ReportingQueries.snapshot_scheduling)
            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.4.2.2'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            opes = []
            add_column_safe(opes, migrator, WaptUsers.user_fingerprint_sha1)
            add_column_safe(opes, migrator, WaptUserAcls.user_fingerprint_sha1)
            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.5.0.0'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            opes = []
            add_column_safe(opes, migrator, Hosts.host_packages_filter)
            migrate(*opes)

            for host in Hosts.select(Hosts.uuid, Hosts.host_capabilities, Hosts.host_packages_filter).where(Hosts.host_packages_filter.is_null() & ~Hosts.host_capabilities.is_null()):
                host.host_packages_filter = make_packages_filter_string(host.host_capabilities)
                host.save()

            set_db_version(next_version)

    next_version = '2.5.2.0'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))

            opes = []

            add_column_safe(opes, migrator, HostOSDeploy.use_ip_from_winpe)

            migrate(*opes)

            set_db_version(next_version)

    next_version = '2.5.2.1'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            # HostPackagesStatus, Packages, HostSoftwares, HostGroups, HostWebsocket, HostAuditData, ReportingSnapshots, HostWsus, LogsAPI
            wapt_db.execute_sql("ALTER SEQUENCE hostsoftwares_id_seq AS bigint;")
            wapt_db.execute_sql("ALTER SEQUENCE hostpackagesstatus_id_seq AS bigint;")
            wapt_db.execute_sql("ALTER SEQUENCE hostwsus_id_seq AS bigint;")
            wapt_db.execute_sql("ALTER SEQUENCE hostauditdata_id_seq AS bigint;")
            wapt_db.execute_sql("ALTER SEQUENCE packages_id_seq AS bigint;")
            wapt_db.execute_sql("ALTER SEQUENCE hostgroups_id_seq AS bigint;")
            wapt_db.execute_sql("ALTER SEQUENCE hostwebsocket_id_seq AS bigint;")
            wapt_db.execute_sql("ALTER SEQUENCE reportingsnapshots_id_seq AS bigint;")
            wapt_db.execute_sql("ALTER SEQUENCE logsapi_id_seq AS bigint;")

            set_db_version(next_version)

    next_version = '2.5.2.2'
    if get_db_version() < next_version:
        with wapt_db.atomic():
            logger.info("Migrating from %s to %s" % (get_db_version(), next_version))
            opes = []
            add_column_safe(opes, migrator, HostAuditData.created_by)
            migrate(*opes)
            set_db_version(next_version)

    set_db_version(DB_STRUCTURE_VERSION)


    create_default_users(server_config=server_config)


    # be sure to have triggers and functions
    with wapt_db.atomic():
        req = "SELECT EXISTS(SELECT * FROM pg_proc WHERE proname = 'update_host_sync_status')"
        if not list(wapt_db.execute_sql(req))[0][0]:
            function_update_host_sync_status = """
                CREATE OR REPLACE FUNCTION update_host_sync_status()
                    RETURNS trigger
                    LANGUAGE 'plpgsql'
                    VOLATILE
                    COST 100
                AS $BODY$
                BEGIN
                    IF (TG_OP = 'INSERT') THEN
                        IF ((NEW.wapt_status->>'is_remote_repo')::boolean=True) THEN
                            INSERT INTO hostsyncstatus (uuid) VALUES (NEW.uuid);
                        END IF;
                        RETURN NEW;
                    ELSEIF (TG_OP = 'UPDATE') THEN
                        IF ((NEW.wapt_status->>'is_remote_repo')::boolean=True) THEN
                            IF NOT EXISTS (SELECT 1 FROM hostsyncstatus WHERE uuid = NEW.uuid) THEN
                                INSERT INTO hostsyncstatus (uuid) VALUES (NEW.uuid);
                            END IF;
                        ELSE
                            DELETE FROM hostsyncstatus WHERE hostsyncstatus.uuid = OLD.uuid;
                        END IF;
                        RETURN NEW;
                    ELSE
                        DELETE FROM hostsyncstatus WHERE hostsyncstatus.uuid = OLD.uuid;
                        RETURN OLD;
                    END IF;
                END
                $BODY$;
                """
            wapt_db.execute_sql(function_update_host_sync_status)

        req = "SELECT EXISTS(SELECT tgname FROM pg_trigger WHERE not tgisinternal AND tgname = 'UPDATE_HOST_SYNC_STATUS')"
        if not list(wapt_db.execute_sql(req))[0][0]:
            trigger_update_host_sync_status = """
                CREATE TRIGGER "UPDATE_HOST_SYNC_STATUS"
                BEFORE INSERT OR DELETE OR UPDATE
                ON hosts
                FOR EACH ROW
                EXECUTE PROCEDURE update_host_sync_status();
                """

            wapt_db.execute_sql(trigger_update_host_sync_status)

        if not 'hosts_computer_ad_ou' in [i.name for i in wapt_db.get_indexes('hosts')]:
            wapt_db.execute_sql('create index hosts_computer_ad_ou on hosts(computer_ad_ou)')

        if not 'packages_package_uuid' in [i.name for i in wapt_db.get_indexes('packages')]:
            wapt_db.execute_sql('create unique index packages_package_uuid on packages(package_uuid)')

        if not 'hostsoftwares_host_key' in [i.name for i in wapt_db.get_indexes('hostsoftwares')]:
            wapt_db.execute_sql('create index hostsoftwares_host_key on hostsoftwares(host_id,key)')

        if not 'hostsoftwares_host_software_id' in [i.name for i in wapt_db.get_indexes('hostsoftwares')]:
            wapt_db.execute_sql('create unique index hostsoftwares_host_software_id on hostsoftwares(host_id,software_id)')

        if not 'hostpackagesstatus_host_install_id' in [i.name for i in wapt_db.get_indexes('hostpackagesstatus')]:
            wapt_db.execute_sql('create unique index hostpackagesstatus_host_install_id on hostpackagesstatus(host_id,install_id)')

        if not 'hostwsus_host_update_id' in [i.name for i in wapt_db.get_indexes('hostwsus')]:
            wapt_db.execute_sql('create unique index hostwsus_host_update_id on hostwsus(host_id,update_id)')

        if not 'hostauditdata_hostid' in [i.name for i in wapt_db.get_indexes('hostauditdata')]:
            wapt_db.execute_sql('create index hostauditdata_hostid on hostauditdata(host_id,value_key,value_date)')
        if not 'hostauditdata_unique' in [i.name for i in wapt_db.get_indexes('hostauditdata')]:
            wapt_db.execute_sql('create unique index hostauditdata_unique on hostauditdata(host_id,value_id)')

        if not 'waptusers_name' in [i.name for i in wapt_db.get_indexes('waptusers')]:
            wapt_db.execute_sql('create unique index waptusers_name on waptusers(name)')

        if not 'waptuseracls_username' in [i.name for i in wapt_db.get_indexes('waptuseracls')]:
            wapt_db.execute_sql('create unique index waptuseracls_username on waptuseracls(username)')

        if not 'reportingsnapshots_report_name' in [i.name for i in wapt_db.get_indexes('reportingsnapshots')]:
            wapt_db.execute_sql('create index reportingsnapshots_report_name on reportingsnapshots(report_name,report_date,row_nr)')

        if not 'reportingsnapshots_report_id' in [i.name for i in wapt_db.get_indexes('reportingsnapshots')]:
            wapt_db.execute_sql('create index reportingsnapshots_report_name on reportingsnapshots(report_id,report_date)')

        wapt_db.execute_sql('alter table hostwebsocket set unlogged')
        wapt_db.execute_sql('alter table hostpackagesstatus set logged')
        wapt_db.execute_sql('alter table hostsoftwares set logged')
        wapt_db.execute_sql('alter table reportingsnapshots set logged')

if __name__ == '__main__':
    if platform.system() != 'Windows' and getpass.getuser() != 'wapt':
        print("""you should run this program as wapt:
                     sudo -u wapt python /opt/wapt/waptserver/model.py  <action>
                 actions : init_db
                           upgrade2postgres""")
        sys.exit(1)

    usage = """\
    %prog [-c configfile] [action]

    WAPT Server database reset / init / upgrade.

    action is either :
       init_db, upgrade_db: initialize or upgrade an existing DB without dropping data
       reset_db: initiliaze or recreate an empty database dropping the data.
    """
    parser = OptionParser(usage=usage, version=__version__)
    parser.add_option(
        '-c',
        '--config',
        dest='configfile',
        default=waptserver.config.DEFAULT_CONFIG_FILE,
        help='Config file full path (default: %default)')
    parser.add_option('-l', '--loglevel', dest='loglevel', default=None, type='choice',
                      choices=['debug',   'warning', 'info', 'error', 'critical'],
                      metavar='LOGLEVEL', help='Loglevel (default: warning)')
    parser.add_option('-d', '--devel', dest='devel', default=False, action='store_true',
                      help='Enable debug mode (for development only)')

    parser.add_option('-o', '--overwrite-version', dest='overwrite_version', default=None,
                      help='Overwrite db structure version before upgrade-db to force structure upgrade replay.')

    (options, args) = parser.parse_args()
    conf = waptserver.config.load_config(options.configfile)
    load_db_config(conf)

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s')

    if options.loglevel is not None:
        setloglevel(logger, options.loglevel)
    else:
        setloglevel(logger, conf['loglevel'])

    if len(args) == 1:
        action = args[0]
        if action in ('init_db','upgrade-db','upgrade_db'):
            print('initializing missing wapt tables without dropping data.')
            if options.overwrite_version is not None:
                print('Current DB structure before upgrade is %s' % get_db_version())
                print('Forcing db structure version to %s' % options.overwrite_version)
                set_db_version(options.overwrite_version)
            init_db(False,server_config=conf)
            print('Current DB version is now: %s' % get_db_version())
            sys.exit(0)
        elif action == 'reset_db':
            print('Drop existing tables and recreate wapt tables.')
            init_db(True,server_config=conf)
            sys.exit(0)
        elif action == 'reportsnapshot':
            with wapt_db:
                reports = ReportingQueries.select()
                for r in reports:
                    with wapt_db.atomic() as translocal:
                        try:
                            r.snapshot()
                        except Exception as e:
                            translocal.rollback()
                            logger('Error snapshoting %s: %s' % (r.report_name,e))


    else:
        parser.print_usage()
