#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2024  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##


import json
import os
import datetime
import time
import ujson

from waptserver.app import app, append_ws_status

__all__ = [
    'get_all_rules',
    'get_all_agentrepos',
    'get_sync_version',
    'get_sync_changelog',
    'save_rules',
    'get_createupdatefilesync',
    'update_file_tree_of_files',
    'repositories'
]


from flask import request, session, Response, send_from_directory, Blueprint

#from waptserver.utils import *

#from waptserver.model import *
from peewee import JOIN,fn

from waptserver.model import ServerAttribs,SiteRules,Hosts,HostSyncStatus,SyncStatus,WaptDB

from waptserver.common import make_response, make_response_from_exception
from waptserver.decorators import requires_auth, require_wapt_db, requires_enterprise
from waptserver.repositories_tasks import update_file_tree_of_files_in_huey

try:
    # i18n
    from flask_babel import Babel
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)
    Babel = None

_ = gettext

repositories = Blueprint('repositories', __name__)

@repositories.route('/api/v3/get_all_rules')
@requires_auth(['admin', 'view'])
@require_wapt_db
@requires_enterprise
def get_all_rules():
    """
    Get all rules (rules for the agent to select a repo_url) from the DB
    """
    start_time = time.time()
    try:
        rules = list(SiteRules.select().order_by(SiteRules.sequence.asc()).dicts())
        for row in rules:
            row['created_on'] = row['created_on'].isoformat()
            row['updated_on'] = row['updated_on'].isoformat()
        return make_response(result=rules, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@repositories.route('/api/v3/get_all_agentrepos')
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def get_all_agentrepos():
    """
    Get all agent who are secondary repository
    """
    start_time = time.time()
    try:
        pattern = request.args.get('pattern', '')
        where_clause = Hosts.wapt_status.contains({'is_remote_repo': True}) & (Hosts.computer_name.contains(pattern) | Hosts.computer_fqdn.contains(pattern) | Hosts.uuid.contains(pattern))
        list_agent_repo = []
        for host in Hosts.select(Hosts.uuid, Hosts.computer_fqdn, HostSyncStatus).join(HostSyncStatus, JOIN.LEFT_OUTER, (Hosts.uuid == HostSyncStatus.uuid)).where(where_clause).order_by(HostSyncStatus.id.asc()).dicts():
            list_agent_repo.append(host)
            append_ws_status(host)
        return make_response(result=list_agent_repo, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@repositories.route('/api/v3/get_sync_version')
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def get_sync_version():
    """
    Get sync.json version if there is no sync.json return {version : null}
    """
    start_time = time.time()
    try:
        if os.path.isfile(os.path.join(os.path.abspath(os.path.join(app.conf['wapt_folder'], os.pardir)), 'sync.json')):
            try:
                version = int(SyncStatus.select(fn.MAX(SyncStatus.version)).scalar())
                id = int(SyncStatus.select(fn.MIN(SyncStatus.id)).scalar())
            except:
                version = -1
                id = 0
        else:
            version = -1
            id = 0
        try:
            progression = ServerAttribs.get(key='repo_sync_progression').value
        except:
            progression = None
        try:
            sync_in_progress = ServerAttribs.get(key='repo_sync_in_progress').value
        except:
            sync_in_progress = False
        data = {'version': version, 'id': id, 'progression' : progression, 'sync_in_progress' : sync_in_progress}
        return make_response(result=data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@repositories.route('/api/v3/get_sync_changelog')
@requires_auth(['admin', 'view'])
@requires_enterprise
@require_wapt_db
def get_sync_changelog():
    """
    """
    start_time = time.time()
    query = SyncStatus.select().order_by(SyncStatus.id.desc())
    limit = int(request.args.get('limit', '0'))
    if limit:
        query = query.limit(limit)
    try:
        data = list(query.dicts())
        return make_response(result=data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@repositories.route('/api/v3/save_rules', methods=['HEAD', 'POST'])
@requires_auth(['admin', 'edit_repo_rules'])
@requires_enterprise
@require_wapt_db
def save_rules():
    """
    Save signed rules to packages
    """
    start_time = time.time()
    try:
        data = ujson.loads(request.data)

        # Add rules to rules.json
        rules_fname = os.path.join(os.path.abspath(os.path.join(app.conf['wapt_folder'], os.pardir)), 'rules.json')

        with open(rules_fname, 'w') as outfile:
            json.dump(data, outfile, indent=4, sort_keys=True)

        # Add rules to DB
        SiteRules.delete().execute()
        for row in data:
            row['created_on'] = datetime.datetime.strptime(row['created_on'], "%Y-%m-%dT%H:%M:%S")
            row['updated_on'] = datetime.datetime.strptime(row['updated_on'], "%Y-%m-%dT%H:%M:%S")
        result = SiteRules.insert_many(data).execute()
        return make_response(result=result, success=True, request_time=time.time()-start_time)
    except Exception as e:
        return make_response_from_exception(e)


def recreate_rules_from_db():
    with WaptDB():
        try:
            rules = list(SiteRules.select().dicts())
            for row in rules:
                row['created_on'] = datetime.datetime.strftime(row['created_on'], "%Y-%m-%dT%H:%M:%S")
                row['updated_on'] = datetime.datetime.strftime(row['updated_on'], "%Y-%m-%dT%H:%M:%S")
        except:
            rules = []
    return rules


def recreate_rules_from_json():
    rules_fname = os.path.join(os.path.abspath(os.path.join(app.conf['wapt_folder'], os.pardir)), 'rules.json')
    with open(rules_fname, 'r') as rule_json:
        rules = json.load(rule_json)
    with WaptDB():
        for row in rules:
            row['created_on'] = datetime.datetime.strptime(row['created_on'], "%Y-%m-%dT%H:%M:%S")
            row['updated_on'] = datetime.datetime.strptime(row['updated_on'], "%Y-%m-%dT%H:%M:%S")
        SiteRules.insert_many(rules).execute()


@repositories.route('/api/v3/createupdatefilesync')
@requires_auth(['admin', 'edit_repo_rules'])
@requires_enterprise
def get_createupdatefilesync():
    """
    Create or update the file sync.json
    """
    start_time = time.time()
    res = update_file_tree_of_files(username=session.get('user'))
    result = {}
    result['res'] = res()
    result['time'] = time.time()-start_time
    result['success'] = True
    result = json.dumps(result)
    try:
        return Response(result, mimetype='application/json')
    except Exception as e:
        return make_response_from_exception(e)


def update_file_tree_of_files(username='SERVER'):
    return update_file_tree_of_files_in_huey(username=username)
