#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2024  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import time
import datetime


from flask import request, Blueprint, session
from waptserver.utils import logger

from waptserver.model import WaptDB,Hosts,HostSyncStatus,wapt_db

from waptserver.common import make_response, make_response_from_exception
from waptserver.decorators import requires_auth, allow_local, requires_enterprise

from waptserver.server_socketio import socketio, WS_SID
from waptserver.app import get_ws_connection

repositories_socketio = Blueprint('repositories_socketio', __name__)

try:
    # i18n
    from flask_babel import Babel
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)
    Babel = None

_ = gettext


@repositories_socketio.route('/api/v3/launch_sync_on_remotes_repos', methods=['GET', 'HEAD', 'POST'])
@requires_enterprise
@requires_auth(['admin','view'])
def launch_sync_on_remotes_repos():
    """Launch sync on remotes repositories"""
    start_time = time.time()
    try:
        if not socketio:
            raise Exception('socketio unavailable')

        # in case a POST is issued with a selection of uuids to launch sync

        if request.json is not None:
            uuids = request.json.get('uuids', None) or None
            force = request.json.get('force', None) or None
            audit = request.json.get('audit', None) or None
        else:
            uuids = None
            force = None
            audit = None

        s = "; "

        if force:
            msg = _('Synchronization (force)')
        elif audit:
            msg = _('Audit of files')
        else:
            msg = _('Synchronization')

        message = _('%s launched for host(s) : %s' % (msg, s.join(target_for_sync(uuids, force, audit))))

        return make_response(msg=message, success=True, request_time=time.time()-start_time)
    except Exception as e:
        return make_response_from_exception(e)


@repositories_socketio.route('/api/v3/broadcast_sync_on_remotes_repo', methods=['GET'])
@requires_enterprise
@allow_local
def broadcast_sync_on_remotes_repo():
    start_time = time.time()
    try:
        if not socketio:
            raise Exception('socketio unavailable')

        s = "; "
        message = _('Synchronization launched for host(s) : %s' % s.join(target_for_sync()))
        return make_response(msg=message, success=True, request_time=time.time()-start_time)
    except Exception as e:
        return make_response_from_exception(e)


def target_for_sync(uuids=None, force=False, audit=False):
    with WaptDB():
        logger.debug('Emit websockets sync')
        where_clause_is_remote = Hosts.wapt_status.contains({'is_remote_repo': True})
        if uuids:
            where_clause = Hosts.uuid.in_(uuids)
            hosts = list(Hosts.select(Hosts.uuid, Hosts.computer_fqdn).where(where_clause, where_clause_is_remote).dicts())
        else:
            hosts = list(Hosts.select(Hosts.uuid, Hosts.computer_fqdn).where(where_clause_is_remote).dicts())

        listfqdn = []
        for host in hosts:
            ws = get_ws_connection(host['uuid'])
            if ws[WS_SID]:
                listfqdn.append(host['computer_fqdn'])
                if force:
                    socketio.emit('sync_remote_repo_force', room=ws[WS_SID])
                elif audit:
                    socketio.emit('audit_remote_repo', room=ws[WS_SID])
                else:
                    socketio.emit('sync_remote_repo', room=ws[WS_SID])
        return listfqdn


@socketio.on('synchronization_started')
def on_sync_remote_repo_started(sync_version):
    data = {'status': 'SYNC_STARTED', 'progress': 0.0, 'sync_id': sync_version['id'], 'current_download': None, 'errors': None, 'current_speed': None}
    update_host_sync_status(data, session.get('uuid', None), request.sid)
    logger.info('Synchronization started for remote repo client sid %s (uuid:%s)' % (request.sid, session.get('uuid', None)))
    return True


@socketio.on('synchronization_error')
def on_sync_remote_repo_error(errors):
    data = {'status': 'SYNC_ERROR', 'version': errors['version']['version'], 'progress': None, 'sync_id': errors['version']['id'], 'current_download': None, 'errors': errors['errors'], 'current_speed': None}
    update_host_sync_status(data, session.get('uuid', None), request.sid)
    logger.info('Synchronization error for remote repo client sid %s (uuid:%s)' % (request.sid, session.get('uuid', None)))
    return True


@socketio.on('synchronization_finished')
def on_sync_remote_repo_finished(sync_version):
    data = {'status': 'SYNC_FINISHED', 'version': sync_version['version'], 'progress': None, 'sync_id': sync_version['id'], 'current_download': None, 'errors': None, 'current_speed': None, 'last_sync': datetime.datetime.utcnow()}
    update_host_sync_status(data, session.get('uuid', None), request.sid)
    logger.info('Synchronization finished for remote repo client sid %s (uuid:%s)' % (request.sid, session.get('uuid', None)))
    return True


@socketio.on('sync_progress')
def on_sync_progress(data):
    update_host_sync_status(data, session.get('uuid', None), request.sid)
    logger.info('Synchronization progress for remote repo client sid %s (uuid:%s) : %s/100' % (request.sid, session.get('uuid', None), data['progress']))
    return True


@socketio.on('synchronization_not_in_time_range')
def on_sync_remote_repo_not_in_time_range():
    data = {'status': 'NOT_IN_TIME_RANGE'}
    update_host_sync_status(data, session.get('uuid', None), request.sid)
    logger.info('Synchronization not in time range for remote repo client sid %s (uuid:%s)' % (request.sid, session.get('uuid', None)))
    return False


@socketio.on('synchronization_not_a_local_remote_repo')
def on_sync_remote_repo_not_a_local_remote_repo():
    data = {'status': 'NOT_A_REMOTE_REPO', 'errors' : {'Remote Repo' : 'Not a remote repo'}}
    update_host_sync_status(data, session.get('uuid', None), request.sid)
    logger.info('Synchronization impossible not a local remote repo for remote repo client sid %s (uuid:%s)' % (request.sid, session.get('uuid', None)))
    return False


@socketio.on('synchronization_audit_started')
def on_audit_remote_repo_started():
    data = {'status': 'AUDIT_STARTED', 'progress': 0.0}
    update_host_sync_status(data, session.get('uuid', None), request.sid)
    logger.info('Audit started for remote repo client sid %s (uuid:%s)' % (request.sid, session.get('uuid', None)))
    return True


@socketio.on('synchronization_audit_finished')
def on_audit_remote_repo_finished():
    data = {'status': 'SYNC_FINISHED', 'progress': None, 'current_download': None, 'errors': None, 'current_speed': None, 'last_audit': datetime.datetime.utcnow()}
    update_host_sync_status(data, session.get('uuid', None), request.sid)
    logger.info('Audit started for remote repo client sid %s (uuid:%s)' % (request.sid, session.get('uuid', None)))
    return True


def update_host_sync_status(data, uuid, sid):
    with WaptDB():
        with wapt_db.atomic():
            HostSyncStatus.update(**data).where(HostSyncStatus.uuid == uuid).execute()
