#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2024  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import os
import sys

if "__file__" in locals():
    sys.path.insert(0, os.path.realpath(os.path.join( os.path.dirname(os.path.realpath(__file__)),'..')))

try:
    wapt_root_dir = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

import warnings
warnings.simplefilter('ignore', category=UserWarning)

if __name__ == '__main__':
    # as soon as possible, we must monkey patch the library...
    # monkeypatching for eventlet greenthreads

    from eventlet import monkey_patch
    import psycogreen.eventlet
    # os=False for windows see https://mail.python.org/pipermail/python-bugs-list/2012-November/186579.html
    if hasattr(sys, 'getwindowsversion'):
        # interactive debug mode on PyScripter hang if tread is patched.
        if 'rpyc' in sys.modules:
            monkey_patch(os=False, thread=False)
        else:
            monkey_patch(os=False)
            psycogreen.eventlet.patch_psycopg()
    else:
        monkey_patch()
        psycogreen.eventlet.patch_psycopg()

import platform

oldsysargv = sys.argv
import waptlicences
sys.argv = oldsysargv

# we can call setproctitle only after importing waptlicences
# else we have a SEGFAULT duing waptlicences import.
if not sys.platform == 'win32':
    import setproctitle
    setproctitle.setproctitle('waptserver')

from waptutils import __version__

import time
import json
import ujson
import logging
import logging.handlers
import fnmatch
import zlib
import socket
import shutil
import tempfile
import traceback
import datetime
import glob
import base64
import ldap3.utils
import wakeonlan
import codecs
import hashlib

from optparse import OptionParser

from werkzeug.utils import secure_filename

from flask import request, Response, send_from_directory, session, redirect, render_template, abort

from peewee import SQL,fn,JOIN,Select


from playhouse.postgres_ext import ArrayField,JSONField, BinaryJSONField

from waptserver.model import Hosts, HostSoftwares, HostPackagesStatus, HostGroups, HostWsus, WsusUpdates, Packages
from waptserver.model import WaptUsers, WaptUserAcls, LogsAPI, HostAuditData, Normalization, ReportingQueries, MessagesTemplate, HostConfigs
from waptserver.model import ServerAttribs
from waptserver.model import get_db_version, init_db, wapt_db, update_host_data,get_db_need_upgrade
from waptserver.model import upgrade_db_structure
from waptserver.model import WaptDB, load_db_config, wapt_db_connect
from waptserver.config import get_http_proxies, WAPTLOGGERS

from waptpackage import PackageEntry, WaptLocalRepo, EWaptBadSignature, EWaptMissingCertificate
from waptcrypto import SSLCertificate, SSLVerifyException, SSLCertificateSigningRequest, InvalidSignature, SSLPrivateKey
from waptcrypto import SSLCABundle, SSLCRL

from waptutils import datetime2isodate, ensure_list, Version, setloglevel, sha256_for_data, sha256_for_file

from waptserver.utils import EWaptAuthenticationFailure, EWaptForbiddden, EWaptHostUnreachable
from waptserver.utils import EWaptMissingParameter, EWaptTimeoutWaitingForResult, EWaptUnknownHost
from waptserver.utils import get_disk_space, get_forwarded_url, get_all_json_configs
from waptserver.utils import get_dns_domain, get_wapt_exe_version

from waptserver.common import get_secured_token_generator, get_server_uuid, get_secondary_repo
from waptserver.common import make_response, make_response_from_exception, licences_list, LicenceStatus, PostDataStream
from waptserver.common import update_all_crls, get_revision_hash, get_wapt_full_version

from waptserver.app import app, WS_SID, WS_TIMESTAMP, append_ws_status, connected_uuids, host_is_reachable, get_ws_connections
from waptserver.auth import check_auth, change_admin_password, get_user_acls, has_any_right, set_user_password_htpasswd, delete_user_password_htpasswd,list_user_password_htpasswd
from waptserver.decorators import requires_auth, get_request_auth_http_status, authenticate, gzipped, require_wapt_db, add_logs

from waptserver.wsus_tasks import download_wsusscan, download_wsusscan_crontab, download_windows_update_task, download_missing_cabs
from waptserver.repositories_tasks import update_file_tree_of_files_in_huey
app.enterprise_tasks = [download_wsusscan, download_wsusscan_crontab, download_windows_update_task, download_missing_cabs,
    update_file_tree_of_files_in_huey]

import waptserver.config

import cryptography.hazmat.backends.openssl.backend

# socketio is loaded conditionally if iwe are running in app mode, not uwsgi mode
socketio = None
git_hash = ''
wapt_edition = ''

# imported later in case we are running a socketio process
update_ws_db = None

import auth_module_ad

logger = logging.getLogger('waptserver')

# Ensure that any created files have sane permissions.
# uWSGI implicitely sets umask(0).
try:
    os.umask(0o022)
except Exception:
    pass

ALLOWED_EXTENSIONS = set(['.wapt'])

# i18n
try:
    from flask_babel import Babel
    from flask_babel import gettext

    babel = Babel(app)

    @babel.localeselector
    def get_locale():
        browser_lang = request.accept_languages.best_match(['en', 'fr'])
        user_lang = session.get('lang', browser_lang)
        return user_lang

except ImportError:
    gettext = (lambda s: s)
    Babel = None
_ = gettext

from waptserver import wsus, enterprise, repositories, repositories_socketio, wads

app.register_blueprint(wsus.wsus)
app.register_blueprint(enterprise.enterprise)
app.register_blueprint(repositories.repositories)
app.register_blueprint(repositories_socketio.repositories_socketio)
app.register_blueprint(wads.wads)


def get_wapt_edition():
    global licences_list
    return 'enterprise' if licences_list.is_enterprise else 'community'

@app.teardown_request
def _db_close(error):
    """Closes the database again at the end of the request."""
    if wapt_db and not wapt_db.is_closed():
        logger.warning('waptdb was not closed after request. Used DB connections Count: %s' % len(wapt_db._in_use))
        # perhaps we should not cloe DB as request are runnong in background with same db connection ?
        # wapt_db_close()


@app.route('/lang/<language>')
def lang(language=None):
    session['lang'] = language
    return redirect('/')


@app.route('/')
def index():
    if app.conf['login_on_homepage'] :
        auth_result = check_auth(session=session, request=request, methods=app.conf['login_auth_methods'])
        if not auth_result:
            time.sleep(3)
            return authenticate()
        if not has_any_right(get_user_acls(username=auth_result['user']),['admin', 'register_host', 'view']):
            return make_response(success=False,msg='forbidden',status=403)

    global full_version
    waptagent = os.path.join(app.conf['wapt_folder'], 'waptagent.exe')
    waptsetup = os.path.join(app.conf['wapt_folder'], 'waptagent', 'waptsetup.exe')
    waptdeploy = os.path.join(app.conf['wapt_folder'], 'waptagent', 'waptdeploy.exe')

    agent_status = setup_status = deploy_status = db_status = 'N/A'
    agent_style = setup_style = deploy_style = disk_space_style = 'style="color: red;"'

    setup_present, setup_version = get_wapt_exe_version(waptsetup)
    if setup_present:
        setup_style = ''
        if setup_version is not None:
            setup_status = setup_version
        else:
            setup_status = 'ERROR'

    agent_present, agent_version = get_wapt_exe_version(waptagent)
    agent_sha256 = None
    if agent_present:
        if agent_version is not None:
            agent_status = agent_version
            agent_sha256 = sha256_for_file(waptagent)
            if Version(agent_version) >= Version(setup_version):
                agent_style = ''
        else:
            agent_status = 'ERROR'

    deploy_present, deploy_version = get_wapt_exe_version(waptdeploy)
    if deploy_present:
        deploy_style = ''
        if deploy_version is not None:
            deploy_status = deploy_version
        else:
            deploy_status = 'ERROR'

    try:
        db_status = 'OK'
        if get_db_need_upgrade():
            db_status = 'NEED UPGRADE (%s)' % get_db_version()
    except Exception:
        db_status = 'ERROR'

    try:
        space = get_disk_space(app.conf['wapt_folder'])
        if not space:
            raise Exception('Disk info not found')
        percent_free = (space[0] * 100) / space[1]
        if percent_free >= 20:
            disk_space_style = ''
        disk_space_str = '%.2f %% free' % percent_free
    except Exception as e:
        disk_space_str = 'error, %s' % str(e)

    if os.path.isfile(waptsetup):
        waptsetup_tis_url = 'wapt/waptagent/waptsetup.exe'
    else:
        waptsetup_tis_url = 'https://wapt.tranquil.it/wapt/releases/wapt-%s/waptsetup.exe' % full_version

    if os.path.isfile(waptdeploy):
        waptdeploy_url = 'wapt/waptagent/waptdeploy.exe'
    else:
        waptdeploy_url = 'https://wapt.tranquil.it/wapt/releases/wapt-%s/binaries/windows/waptdeploy.exe' % full_version

    host_url = request.host_url.rsplit('/',1)[0]
    host_url = host_url.replace('http://',request.headers.get('X-Forwarded-Proto','https')+'://')

    json_config_urls = [ '%s/wapt/conf.d/%s' % (host_url,f['filename']) for f in get_all_json_configs(os.path.join(app.conf['wapt_folder'],'conf.d'))]
    waptagent_folder =  os.path.join(app.conf['wapt_folder'], 'waptagent')
    agents_list = get_agents_list_symlink(waptagent_folder)

    data = {
        'wapt': {
            'server': {'status': full_version.split('-')[0]},
            'agent': {'status': agent_status, 'style': agent_style, 'sha256': agent_sha256},
            'setup': {'status': setup_status, 'style': setup_style, 'url': waptsetup_tis_url},
            'deploy': {'status': deploy_status, 'style': deploy_style, 'url': waptdeploy_url},
            'db': {'status': db_status},
            'disk_space': {'status': disk_space_str, 'style': disk_space_style},
            'wapt_edition': get_wapt_edition(),
            'hostname': get_forwarded_url(request),
            'json_config_urls': json_config_urls,
            'agents_list': agents_list,
            'ip': '{{ip}}',
            'auto_create_waptagent_from_config': app.conf['auto_create_waptagent_from_config']
        }
    }
    return render_template('index.html', data=data)

def get_agents_list_symlink(waptagent_folder):
    #tis-waptagent.el8.x86_64.rpm
    #tis-waptagent-ubuntu-18-amd64.deb
    agents = []
    for target in os.listdir(waptagent_folder):
        if os.path.islink(os.path.join(waptagent_folder, target)):
            try:
                parts, ext = os.path.splitext(target)
                if ext == '.exe':
                    platform = 'windows'
                elif (ext == '.rpm') or (ext == '.deb'):
                    platform = 'linux'
                elif ext == '.pkg':
                    platform = 'macos'
                else:
                    platform = ''
                str_len = len(target) - len('tis-waptagent') - 1
                agent = {'filename': target,
                         'distribution': target[-str_len:],
                         'platform': platform
                        }
                agents.append(agent)
            except:
                pass
    return agents

def sign_host_csr(host_certificate_csr, lifetime_days = None):
    """Sign the CSR with server key and return a certificate for further host auth on nginx server

    Args:
        host_certificate_csr ()

    Returns:
        SSLCertificate: Signed certificate

    """
    host_cert = None
    if (app.conf['clients_signing_key'] and app.conf['clients_signing_certificate'] and
            os.path.isfile(app.conf['clients_signing_key']) and os.path.isfile(app.conf['clients_signing_certificate'])):
        signing_key = SSLPrivateKey(app.conf['clients_signing_key'])
        signing_cert = SSLCertificate(app.conf['clients_signing_certificate'])
        if lifetime_days is None:
            lifetime_days = app.conf['client_certificate_lifetime']
        if lifetime_days < 1:
            lifetime_days = 1
        if app.conf['clients_signing_crl_url']:
            crl_urls = ensure_list(app.conf['clients_signing_crl_url'])
        else:
            crl_urls = None
        host_cert = signing_cert.build_certificate_from_csr(host_certificate_csr, signing_key, validity_days=lifetime_days, crl_urls=crl_urls)
    return host_cert


def check_host_cert(host_certificate):
    """Check if host_certificate is trusted by the server own CA,
    ie it has been signed by the server key

    Args:
        host_certificate (SSLCertificate): host certificate to check.

    Returns
        list: cert chain if OK. Raise exception if not approved. None if no certificate is defined on the server.

    """
    if app.conf['clients_signing_certificate'] and os.path.isfile(app.conf['clients_signing_certificate']):
        server_ca = SSLCABundle(app.conf['clients_signing_certificate'])
        return server_ca.check_certificates_chain(host_certificate)
    return None


def revoke_cert(host_certificate):
    if (app.conf['clients_signing_key'] and app.conf['clients_signing_certificate'] and app.conf['clients_signing_crl'] and
            os.path.isfile(app.conf['clients_signing_key']) and os.path.isfile(app.conf['clients_signing_certificate'])):
        signing_key = SSLPrivateKey(app.conf['clients_signing_key'])
        signing_cert = SSLCertificate(app.conf['clients_signing_certificate'])
        server_ca = SSLCABundle(app.conf['clients_signing_certificate'], trust_first=True)
        try:
            if server_ca.check_certificates_chain(host_certificate):
                crl = SSLCRL(app.conf['clients_signing_crl'], cakey=signing_key, cacert=signing_cert)
                crl.revoke_cert(host_certificate, crl_ttl_days=app.conf['clients_signing_crl_days'])
                crl.save_as_pem()
            else:
                logger.warning('certificate %s can not be revoked by this clients signing CA %s' % (host_certificate.cn,app.conf['clients_signing_certificate']))
        except Exception as e:
            logger.warning('certificate %s can not be revoked by this clients signing CA %s: %s' % (host_certificate.cn, app.conf['clients_signing_certificate'], repr(e)))

        update_all_crls(app.conf)
    return None

def make_auth_token_from_session():
    token_gen = get_secured_token_generator()
    token_data = dict(
        user=session.get('user','').lower(),
        auth_method=session.get('auth_method'),
        auth_date=session.get('auth_date'),
        server_uuid=get_server_uuid(),
        )
    return token_gen.dumps(token_data)



@app.route('/add_host_kerberos', methods=['HEAD', 'POST'])
@app.route('/add_host', methods=['HEAD', 'POST'])
def register_host():
    """Add a new host into database, and return registration info
    If path is add_host_kerberos, assume there is (already validated by NGINX) computer kerberos SPN in the user part of Authorization http header else
    if path is add_host, requires a valid user/password combination in Authorization http header
    If data contains a host_certificate_csr, sign it a returns it to the caller.


    Specific headers:
        'X-Signature':  base64 encoded signature of payload. Can be checked with public key in host_certificate_csr
                        required if 'allow_unsigned_status_data' conf is False
        'X-Signer': indication of signer (not authoritative)
        'Content-Encoding' : if == 'gzip', payload in decompressed before process
        'authorization' : basic auth
        'X-Forwarded-For' :


    Args:
        posted body (json encoded dict):
            uuid
            host_info.computer_fqdn
            host_certificate_csr
            host_certificate

    Returns:

    """

    # we have to be sure that nginx has filtered out the bogus authorization header set by nginx kerberos module
    if session.get('auth_method') == 'kerb' and not request.path in ('/add_host_kerberos'):
        return authenticate()

    if request.method == 'HEAD':
        return ''

    try:
        starttime = time.time()

        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        data = ujson.loads(raw_data)
        if not data:
            raise Exception('register_host: No data supplied')

        uuid = data['uuid']
        if not uuid:
            raise Exception('register_host: No uuid supplied')
        logger.info('Trying to register host %s' % (uuid,))

        # get request signature
        signature_b64 = request.headers.get('X-Signature', None)
        if signature_b64:
            signature = base64.b64decode(signature_b64)
        else:
            signature = None
        if not signature and not app.conf['allow_unsigned_status_data']:
            raise Exception('register_host: Missing signature')

        # Registering a host requires authentication; Either signatue is Ok or Kerberos or basic
        authenticated_user = None
        registration_auth_user = None

        # 'host' is for pre wapt pre 1.4
        computer_fqdn = data.get('computer_fqdn')
        if computer_fqdn is None:
            computer_fqdn = (data.get('host_info', None) or data.get('host', {})).get('computer_fqdn', None)
        if not computer_fqdn:
            computer_fqdn = ''
            #raise Exception('register_host: No computer_fqdn supplied')

        auth_result = None
        host_cert = None

        # with nginx kerberos module, auth user name is stored as Basic auth in the
        # 'Authorisation' header with password 'bogus_auth_gss_passwd'
        if request.path == '/add_host_kerberos' and app.conf['use_kerberos']:
            auth = request.authorization
            if not auth.username.split('@')[0].endswith('$'):
                raise EWaptAuthenticationFailure('register_host : the account %s is not a machine account ' % auth.username)
            if auth and auth.password == 'bogus_auth_gss_passwd' and auth.username:
                authenticated_user = auth.username.lower().replace('$', '').split('@')[0]
                dns_domain = auth.username.split('@')[1].lower()
                authenticated_user = '%s.%s' % (authenticated_user, dns_domain)
                logger.debug('Kerberos authenticated user %s for %s' % (authenticated_user, computer_fqdn))
                registration_auth_user = 'kerb:%s' % authenticated_user
                auth_date = datetime.datetime.utcnow().isoformat()
                logger.debug('User %s authenticated using kerberos' % (authenticated_user,))
                auth_result = dict(auth_method='kerb', user=authenticated_user, auth_date=auth_date)
            else:
                authenticated_user = None

        # kerberos has failed or kerberos is not enabled, we try
        if not authenticated_user:
            # get authentication from enabled "registration_auth_methods" param.
            auth_result = check_auth(request=request, session=session, methods=app.conf['registration_auth_methods'])
            if auth_result:
                # assume authenticated user is the fqdn provided in the data
                logger.debug('Authenticated registration for %s with user %s' % (computer_fqdn, auth_result['user']))
                authenticated_user = auth_result['user']
                session.update(**auth_result)
                registration_auth_user = '%s:%s' % (auth_result['auth_method'], auth_result['user'])

            # if certificate is properly signed, we can trust it without using database
            if not auth_result and 'host_certificate' in data and app.conf.get('trust_signed_host_certificate',False):
                try:
                    untrusted_host_cert = SSLCertificate(crt_string=data['host_certificate'].encode('utf8'))
                    if (untrusted_host_cert.issuer_subject_hash != untrusted_host_cert.subject_hash and
                            (untrusted_host_cert.cn.lower() == computer_fqdn.lower() or untrusted_host_cert.cn.lower() == uuid.lower())):
                        # we can check if issuer is myself...
                        cert_chain = check_host_cert(untrusted_host_cert)
                        if cert_chain:
                            host_cert = untrusted_host_cert
                            authenticated_user = computer_fqdn
                            registration_auth_user = 'Cert:%s' % host_cert.cn

                except Exception as e:
                    logger.warning('Unable to trust supplied host certificate: %s' % (repr(e),))
                    host_cert = None

        # db stuff
        with WaptDB():
            with wapt_db.atomic():
                if not host_cert and not authenticated_user:
                    existing_host = Hosts.select(Hosts.host_certificate, Hosts.computer_fqdn).where(Hosts.uuid == uuid).first()
                    if existing_host and existing_host.host_certificate:
                        # check if existing record, and in this case, check signature with existing certificate
                        host_cert = SSLCertificate(crt_string=existing_host.host_certificate)

                # if data is properly signed by a trusted certificate, consider we are authenticated.
                if host_cert and not authenticated_user:
                    try:
                        authenticated_user = host_cert.verify_content(sha256_for_data(raw_data), signature)
                    except (InvalidSignature, SSLVerifyException):
                        authenticated_user = None

                if not authenticated_user and app.conf['allow_unauthenticated_registration']:
                    logger.warning('Unauthenticated registration for %s' % computer_fqdn)
                    # assume authenticated user is the fqdn provided in the data
                    authenticated_user = computer_fqdn  # request.headers.get('X-Forwarded-For',None)
                    registration_auth_user = 'None:%s' % request.headers.get('X-Forwarded-For', None)

                if not authenticated_user:
                    # use basic auth
                    return authenticate()

            if not authenticated_user:
                raise EWaptAuthenticationFailure('register_host : Missing authentication header')

            if not app.conf['allow_unauthenticated_registration']:
                logger.debug('Unauthenticated computer %s with user %s ' % (computer_fqdn, authenticated_user,))
                # check that authenticated user matches the CN of the certificate supplied in post data
                supplied_host_cert = SSLCertificate(crt_string=data['host_certificate'].encode('utf8'))
##                if request.path == '/add_host_kerberos' :
##                    if computer_fqdn.lower() != auth_result['user'].lower():
##                        raise EWaptAuthenticationFailure('register_host : Mismatch hostname %s / %s' % (computer_fqdn.lower() , auth_result['user'].lower()))
                if not (supplied_host_cert.cn.lower() == computer_fqdn.lower() or supplied_host_cert.cn.lower() == uuid.lower()):
                    raise EWaptAuthenticationFailure('register_host : Mismatch between certificate Certificate commonName %s and supplied fqdn or uuid %s / %s' % (supplied_host_cert.cn, computer_fqdn, uuid))
##                if supplied_host_cert.subject_alt_names[0].lower() != computer_fqdn.lower():
##                    raise EWaptAuthenticationFailure('register_host : Mismatch hostname and subject_alt_names %s / %s' % (supplied_host_cert.subject_alt_names[0].lower(), computer_fqdn.lower()))
            else:
                supplied_host_cert = None

            data['last_seen_on'] = datetime2isodate()

            # sign the CSR if present
            valid_auth = authenticated_user is not None  # auth_result and auth_result['auth_method'] in ['admin','passwd','ldap','kerb']
            if 'host_certificate_signing_request' in data and valid_auth:
                host_certificate_csr = SSLCertificateSigningRequest(csr_pem_string=data['host_certificate_signing_request'])
                if host_certificate_csr.cn.lower() == computer_fqdn.lower() or host_certificate_csr.cn.lower() == uuid.lower():
                    logger.info('Authenticated user: %s , Registration auth user: %s Issuing a signed certificate for %s' % (authenticated_user, registration_auth_user, host_certificate_csr.cn))
                    signed_host_cert = sign_host_csr(host_certificate_csr)
                    if signed_host_cert:
                        # we append server client CA for reference
                        data['host_certificate'] = b'\n'.join([signed_host_cert.as_pem(),open(app.conf['clients_signing_certificate'],'rb').read()])
                else:
                    logger.warning('Authenticated user: %s , Registration auth user %s asking a signed certificate for %s but registering as %s or %s. Refused.' % (authenticated_user, registration_auth_user, host_certificate_csr.cn, computer_fqdn, uuid))

                if registration_auth_user:
                    data['registration_auth_user'] = registration_auth_user

            db_data = update_host_data(data, app.conf)
            # return back signed host certificate
            if 'host_certificate_signing_request' in data and valid_auth:
                db_data['host_certificate'] = data['host_certificate']

            result = db_data
            result['server_uuid'] = get_server_uuid()
            message = 'register_host'


            # generate an auth token for immediate websocket connection
            token_gen = get_secured_token_generator()
            token_data = {'uuid': uuid, 'server_uuid': get_server_uuid(), 'purpose':'websocket'}
            result['authorization_token'] = token_gen.dumps(token_data)

            # add to current sid list
            if app.ws_connections is not None:
                if not uuid in app.ws_connections:
                    app.ws_connections[uuid] = [None, datetime.datetime.utcnow()]
                else:
                    ws = app.ws_connections[uuid]
                    ws[WS_TIMESTAMP] = datetime.datetime.utcnow()

            return make_response(result=result, msg=message, request_time=time.time() - starttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('add_host failed %s' % (repr(e)))
        time.sleep(3)
        return make_response_from_exception(e)


@app.route('/update_host', methods=['HEAD', 'POST'])
@require_wapt_db
def update_host():
    """Update localstatus of computer, and return known registration info
    Requires a base64 encoded signature in X-Signature http header (unless allow_unsigned_status_data config is True)
    This signature is checked using the host certificate stored in Hosts.host_certificate field in database.
    Data is supplied as a JSon (optionnaly gzipped) POST data.
    Required keys are:
        uuid
        host_info.computer_fqdn

    """
    if request.method == 'HEAD':
        return ''

    try:
        starttime = time.time()

        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        data = ujson.loads(raw_data)
        if not data:
            raise Exception('update_host: No data supplied')

        uuid = data['uuid']
        if not uuid:
            raise Exception('update_host: No uuid supplied')

        # 'host' is for pre wapt pre 1.4
        computer_fqdn = data.get('computer_fqdn')
        if computer_fqdn is None:
            computer_fqdn = (data.get('host_info', None) or data.get('host', {})).get('computer_fqdn', None)
        if not computer_fqdn:
            computer_fqdn = ''
            #raise Exception('update_host: No computer_fqdn supplied')

        logger.info('updating host status %s (%s), data:%s' % (uuid, computer_fqdn, list(data.keys())))

        # get request signature
        signature_b64 = request.headers.get('X-Signature', None)
        if signature_b64:
            signature = base64.b64decode(signature_b64)
        else:
            signature = None

        if not signature and not app.conf['allow_unsigned_status_data']:
            raise Exception('update_host: Missing signature')

        # check if data is for the registered host
        existing_host = Hosts.select(Hosts.uuid, Hosts.host_certificate, Hosts.computer_fqdn).where(Hosts.uuid == uuid).first()

        if not existing_host:
            raise EWaptMissingCertificate(
                'You try to update status of an unknown host %s (%s). Please register first.' % (uuid, computer_fqdn))

        if not app.conf['allow_unsigned_status_data']:
            if not existing_host.host_certificate:
                raise EWaptUnknownHost('Host certificate in not known for host %s, aborting' % (uuid,))

            host_cert = SSLCertificate(crt_string=existing_host.host_certificate.encode('utf8'))
            # certificate CN must at least match host uuid
            if not (host_cert.cn.lower() == uuid.lower() or host_cert.cn.lower() == computer_fqdn.lower()):
                raise EWaptUnknownHost('Supplied uuid or hostname does not match known certificate CN, aborting')

            if signature:
                logger.debug('About to check supplied data signature with certificate %s' % host_cert.cn)
                try:
                    host_cert_cn = host_cert.verify_content(sha256_for_data(raw_data), signature)
                except Exception as e:
                    # for pre 1.5 wapt clients
                    host_cert_cn = 'unknown'
                    if not app.conf['allow_unsigned_status_data']:
                        raise e
                logger.info('Data successfully checked with certificate CN %s for %s' % (host_cert_cn, uuid))
                if existing_host and not (host_cert_cn.lower() == existing_host.computer_fqdn.lower() or host_cert_cn.lower() == existing_host.uuid.lower()):
                    raise Exception('update_host: mismatch between host certificate DN %s and existing host hostname %s' % (host_cert_cn, existing_host.computer_fqdn))
            else:
                raise Exception('update_host: Invalid request. Missing signature')

        # be sure to not update host certificate
        if 'host_certificate' in data and not app.conf['allow_unsigned_status_data']:
            del data['host_certificate']

        data['last_seen_on'] = datetime2isodate()
        db_data = update_host_data(data, app.conf)

        result = db_data
        result['server_uuid'] = get_server_uuid()
        message = 'update_host'

        # add to current sid list
        if app.ws_connections is not None:
            if not uuid in app.ws_connections:
                app.ws_connections[uuid] = [None, datetime.datetime.utcnow()]
            else:
                ws = app.ws_connections[uuid]
                ws[WS_TIMESTAMP] = datetime.datetime.utcnow()

        return make_response(result=result, msg=message, request_time=time.time() - starttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('update_host failed %s' % (repr(e)))
        return make_response_from_exception(e)

def get_repo_packages() -> WaptLocalRepo:
    """Returns local repository instance for this server wapt_folder dir

    Returns:
        WaptLocalRepo
    """
    return WaptLocalRepo(app.conf['wapt_folder'])

def allowed_file(filename: str) -> str:
    """Ensure filename is safe and ascii only
    """
    return all(ord(c) < 128 for c in filename) and filename == secure_filename(filename) and os.path.splitext(filename)[1] in ALLOWED_EXTENSIONS

@app.route('/api/v3/packages')
@requires_auth(['admin', 'view'])
@require_wapt_db
def localrepo_packages():
    try:
        start_time = time.time()
        query = ensure_list(request.args.get('q', None), allow_none=True)
        depends = ensure_list(request.args.get('dependencies', []), allow_none=True)
        needed_sections = ensure_list(request.args.get('sections', ['base', 'wsus']), allow_none=True)
        newest_only = request.args.get('newest_only', '1') != '0'
        local_repo = get_repo_packages()
        # Don't scan all the packages if we only want the hosts
        if needed_sections == ['host']:
            packages = []
        else:
            packages = local_repo.search(searchwords=query, newest_only=newest_only, sections=needed_sections, dependencies_list=depends)
        if not needed_sections or 'host' in needed_sections:
            host_uuids = []
            searchwords = '%' + request.args.get('q', '') + '%'
            if depends != []:
                host_uuids = [h for h in list(HostGroups.select(HostGroups.host, Hosts.computer_fqdn).where(HostGroups.group_name << depends).join(Hosts, 'LEFT JOIN', on=(HostGroups.host == Hosts.uuid)).
                where((Hosts.uuid ** searchwords) | (Hosts.computer_fqdn ** searchwords) | (Hosts.description ** searchwords)).distinct().tuples())]
            else:
                host_uuids = [h for h in list(HostGroups.select(HostGroups.host, Hosts.computer_fqdn).join(Hosts, 'LEFT JOIN', on=(HostGroups.host == Hosts.uuid)).
                where((Hosts.uuid ** searchwords) | (Hosts.computer_fqdn ** searchwords) | (Hosts.description ** searchwords)).distinct().tuples())]
            for host in host_uuids:
                package_file_path = os.path.join(app.conf['wapt_folder']+'-host', host[0] + '.wapt')
                if not os.path.isfile(package_file_path):
                    continue
                pe = PackageEntry(waptfile=package_file_path)
                pe.name = host[1]
                packages.append(pe)


        return make_response(packages,
                             msg='%s packages in local repository matching the keywords %s' % (len(packages), query),
                             request_time=time.time() - start_time)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('localrepo_packages %s' % (repr(e)))
        return make_response_from_exception(e)

@app.route('/api/v3/packages_list')
@requires_auth(['admin', 'view'])
@require_wapt_db
def packages_list():
    """Returns list of available packages from database

    Returns:
        dict : result

    """
    try:
        start_time = time.time()
        exclude_sections = ensure_list(request.args.get('exclude_sections', 'unit,host,profile'))
        where = True

        packages = Packages.select(Packages.package.distinct())
        if request.args.get('available_only','true') == 'true':
            where = (Packages.is_available == True)
        if exclude_sections:
            packages = packages.where(where & (~ Packages.section.in_(exclude_sections)))
        packages = [package[0] for package in packages.tuples()]
        return make_response(packages,
                             msg='%s packages_list' % (len(packages)),
                             request_time=time.time() - start_time)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('packages_list failed %s' % (repr(e)))
        return make_response_from_exception(e)

@app.route('/api/v3/get_unavailable_packages', methods=['GET', 'POST'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def get_unavailable_packages():
    """Returns list of unavailable packages from the database

    Args:
        names (list): list of unavailable packages names to get.

    Returns:
        dict : result

    """
    try:
        start_time = time.time()
        names = ujson.loads(request.data)

        where = Packages.is_available == False
        if names:
            where = where & Packages.package.in_(names)

        packages = list(Packages.select().where(where).distinct(Packages.package).dicts())
        return make_response(packages,
                             msg='%s unavailable package(s) found' % (len(packages)),
                             request_time=time.time() - start_time)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('get_unavailable_packages failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/known_packages')
@requires_auth(['admin', 'view'])
@require_wapt_db
def known_packages():
    """Returns list of known packages metadata from database

    Args:
        exclude_sections (csv str): packages section to exclude from result.
                                    default 'unit,host,profile' are excluded

    Returns:
        dict : result

    """
    try:
        start_time = time.time()
        available_only = request.args.get('available_only','0') != '0'
        exclude_sections = ensure_list(request.args.get('exclude_sections', 'unit,host,profile'))
        where = None
        if available_only:
            where = Packages.is_available == True

        if exclude_sections:
            crit = ~ Packages.section.in_(exclude_sections)
            if where is not None:
                where = where & crit
            else:
                where = crit

        packages = Packages.select(Packages.package,
            SQL("target_os as target_oss"),
            SQL("array_agg(distinct section) as sections"),
            SQL("array_agg(distinct maturity) as maturities"),
            SQL("array_agg(distinct name) as names"),
            SQL("array_agg(distinct is_available) as availabilities"),
            fn.MAX(Packages.signature_date).alias('last_signature_date'))
        if where:
            packages = packages.where(where)
        list_packages = list(packages.group_by(Packages.package,Packages.target_os).order_by(1,2).dicts())
        return make_response(list_packages,
                             msg='%s known packages%s' % (len(list_packages),' available on this server' if available_only else ''),
                             request_time=time.time() - start_time)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('known_packages failed %s' % (repr(e)))
        return make_response_from_exception(e)

@app.route('/api/v3/latest_installed_package_version', methods=['GET'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def latest_installed_package_version():
    try:
        start_time = time.time()
        package_name = request.args.get('package', None)

        if not package_name:
            raise Exception("No package name given")
        versions = (HostPackagesStatus.select(HostPackagesStatus.version).where(HostPackagesStatus.package == package_name) |
            Packages.select(Packages.version).where(Packages.package == package_name)).dicts()

        versions_sorted = sorted([Version(v['version']) for v in versions])

        if not versions:
            return make_response(success=False, msg='No registered version found for package %s' % (package_name),
                             request_time=time.time() - start_time)

        version = str(versions_sorted[-1])
        return make_response(version, msg='Latest registered package %s version is %s' % (package_name,version),
                             request_time=time.time() - start_time)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('latest_installed_package_version failed %s' % (repr(e)))
        return make_response_from_exception(e)


def read_trusted_certificates(trusted_certificates_folder=None):
    """Loads the certificates files from trusted_certificates_folder.
    Trusted certificate is the first certificate of each PEM file in trusted_certificates_folder
    Other certificates in PEM files can be used to build certificate chain

    Returns:
        SSLCABundle : trusted certs are in trusted dict attribute (indexed by fingerprint)
    """
    if not trusted_certificates_folder:
        return None
    if not os.path.isdir(trusted_certificates_folder):
        raise Exception('Trusted certificates folder %s does not exists' % trusted_certificates_folder)

    cabundle = SSLCABundle()
    cabundle.add_pems(trusted_certificates_folder, trust_first=True)
    return cabundle


def check_valid_signer(package, cabundle):
    """Check if the signer of package is in the trusted list of certificates in cabundle.

    Args:
        package (PackageEntry): package to check signer
        cabundle (SSLCABundle) : buncle of CA trusted certificate.

    Returns:
        list of SSLCertificate : trusted certificate chain.
    """
    with WaptDB():
        signer_certs = package.package_certificates()
        if package.has_file('setup.py'):
            # check if certificate has code_signing extended attribute
            if not signer_certs or not signer_certs[0].is_code_signing:
                raise EWaptForbiddden('The package %s contains setup.py code but has not been signed with a proper code_signing certificate' % package.package)
        if cabundle is not None:
            trusted_chain = cabundle.check_certificates_chain(signer_certs, check_is_trusted=True)
        else:
            trusted_chain = signer_certs

        # If it's host package, check if host is known to trust one of the signer cert in cert chain.
        if package.section == 'host':
            host = Hosts.select(Hosts.host_capabilities).where(Hosts.uuid == package.package).first()
            # get the list of trusted packages signers for this host
            if host.host_capabilities:
                packages_trusted_ca_fingerprints = host.host_capabilities['packages_trusted_ca_fingerprints']
                if packages_trusted_ca_fingerprints:
                    trusted_cert = None
                    for signer_cert in trusted_chain:
                        # SSLCertificate.get_fingerprint('sha256').hexdigest()
                        if signer_cert.fingerprint in packages_trusted_ca_fingerprints:
                            trusted_cert = signer_cert
                            logger.info('Package %s is trusted for signer %s issued by %s' % (package.package, signer_cert.cn, signer_cert.issuer_dn))
                            break

                    if not trusted_cert:
                        raise EWaptForbiddden("Host with UUID %s does not trust this host package's signer certificate (sha256 %s). Package rejected by server." % (package.package, trusted_chain[0].fingerprint))

        return trusted_chain

def user_is_admin():
    user_acls = session.get('user_acls',{})
    return user_acls.get('admin') and (user_acls['admin'] == True or get_server_uuid() in user_acls['admin'])


def is_allowed_package_for_user(username: str, user_acls: dict, entry: PackageEntry, package_cert_chain_fingerprints: list) -> bool:
    """Check if package if package can be edited by user given its name, section and cert chain

    Args:
        username (str): user name
        user_acls (dict) : acl_name:list or bool
        entry (PackageEntry) : package to check acl against
        package_cert_chain_fingerprints (list): list of sha1 of certificate chain of package signature

    Returns:
        bool

    Raises:
        EWaptForbiddden: if package edition for the user is not allowed


    """
    if not username:
        raise EWaptForbiddden('No username')
    if not isinstance(entry,PackageEntry):
        raise EWaptForbiddden('No package')
    if not isinstance(user_acls,dict):
        raise EWaptForbiddden('Bad user acls, must be a dict')
    if not isinstance(package_cert_chain_fingerprints,list):
        raise EWaptForbiddden('Bad package signature certificate chain, must be a list')

    # check admin -> always ok if server_uuid matches
    if user_acls.get('admin') and (user_acls['admin'] == True or get_server_uuid() in user_acls['admin']):
        return True

    if entry.section == 'host':
        # ok if edit_host_package == True or one of allowed perimeter is in package's cert chain
        if user_acls.get('edit_host_package') and (
                user_acls.get('edit_host_package') == True or set(user_acls.get('edit_host_package')) & set(package_cert_chain_fingerprints)):
            return True
        else:
            raise EWaptForbiddden('User %s is not allowed to upload host package %s. Add any of %s perimeters in acl edit_host_package.' % (username,entry.package,package_cert_chain_fingerprints))

    else:
        # ok if edit_<section>_package == True or one of allowed name patterns matches package name
        allowed_names = user_acls.get('edit_%s_package' % entry.section)
        if allowed_names == True:
            return True
        elif allowed_names:
            for name_pattern in allowed_names:
                if fnmatch.fnmatch(entry.package,name_pattern):
                    return True
            raise EWaptForbiddden('User %s is not allowed to upload package name %s. Add package name to acl edit_%s_package ' % (username,entry.package,entry.section))
        else:
            raise EWaptForbiddden('User %s is not allowed to upload package name %s. Add package name to acl edit_%s_package ' % (username,entry.package,entry.section))

    return False


@app.route('/api/v3/get_stripped_package')
@requires_auth(['admin', 'view'])
@require_wapt_db
def get_stripped_package():
    """Get a stripped down package with only the wapt metadata / signatures / setup.py

    Args:
        package_uuid (str): package_uuid of th epackage to get

    Returns:
        Response: json message with keys:
                    'success' (bool)
                    'error_code' if success is False
                    'msg' (str)
                    'result' (data) : zipped content of the Package
                    'request_time' (float)
    """
    if 'package_uuid' in request.args:
        package_uuid = request.args['package_uuid']
    else:
        raise EWaptMissingParameter('Parameter uuid is missing')

    r = get_repo_packages()
    pe = r.get_package_by_uuid(package_uuid)
    if pe:
        stripped_down_data = pe.get_stripped_package()
        return Response(
            stripped_down_data,
            mimetype="application/zip",
            headers={"Content-disposition":
                     "attachment; filename={}".format(pe.filename)})
    else:
        return abort(404, description="Resource not found")

def read_package(packagefile: PostDataStream, trusted_signers: SSLCABundle, acls: dict, move_to_repository: bool = True, tmp_upload_dir=None):
    """Reads the stream of a zipped package file and return the entry if
    upload of this kind of package (signer, type, name) is allowed for this user.

    Args:
        packagefile (file): opened file like object seeked to the start of a wapt package (Zip file stream)
        trusted_signers (SSLCABundle): if not None, check the if package signer is authorized here
        acls (dict): current user acls
        move_to_repository (bool): if True, move the package to final repository and register it in database as available

    Returns:
        PackageEntry

    """
    target = None
    entry = None
    try:
        wapt_folder = app.conf['wapt_folder']
        if tmp_upload_dir is None:
            tmp_upload_dir = wapt_folder
        tmp_target = tempfile.mktemp(dir=tmp_upload_dir, prefix='wapt')+'.tmp'
        packagefile.save(tmp_target)
        # test if package is OK.
        entry = PackageEntry(waptfile=tmp_target)
        target_filename = entry.make_package_filename(with_md5sum=True)
        if not allowed_file(target_filename):
            raise EWaptForbiddden('Package filename / name is forbidden : %s' % target_filename)

        # check if package signer is authorized
        trusted_chain = check_valid_signer(entry, trusted_signers)
        if not trusted_chain:
            raise EWaptForbiddden('Package %s is not signed properly' % target_filename)
        logger.info('Package is trusted: %s' % trusted_chain)

        trusted_fingerprints = [c.fingerprint for c in trusted_chain]
        # check if package section or package name is authorized for this user
        if not is_allowed_package_for_user(session['user'],acls,entry,trusted_fingerprints ):
            raise EWaptForbiddden('User %s is not allowed to modify this package' % (entry.package))

        logger.debug('Saved package %s into %s' % (entry.asrequirement(), tmp_target))

        entry.filename = target_filename

        if move_to_repository:
            if entry.section == 'host':
                target = os.path.join(wapt_folder+'-host', target_filename)
            else:
                target = os.path.join(wapt_folder, target_filename)

            if os.path.isfile(target):
                os.unlink(target)
            logger.debug('Renaming package %s into %s' % (tmp_target, target))

            shutil.move(tmp_target, target)

            # for caller to get completed PackageEntry.
            entry.localpath = target

            # fix context on target file (otherwith tmp context is carried over)
            #logger.debug(subprocess.check_output('chcon -R -t httpd_sys_content_t %s' % target,shell=True))
            if entry.section == 'host':
                (added, removed) = HostGroups.sync_from_host_package(entry)
                if added or removed:
                    # immediate feedback in waptconsole
                    Hosts.update(host_status='TO-UPGRADE').where(Hosts.uuid == entry.package).execute()
            else:
                (rec, _isnew) = Packages.update_from_control(entry, is_available = True)

        return entry

    except Exception as e:
        logger.debug(traceback.format_exc())
        if entry:
            logger.critical('Error uploading package uuid %s: %s' % (entry.package_uuid, e,))
        else:
            logger.critical('Error uploading packages: %s' % (e,))

        if os.path.isfile(tmp_target):
            os.unlink(tmp_target)
        raise


@app.route('/api/v3/merge_stripped_package', methods=['HEAD', 'POST'])
@requires_auth(['admin', 'edit_host_package', 'edit_base_package', 'edit_group_package',
                'edit_unit_package', 'edit_wsus_package', 'edit_profile_package', 'edit_selfservice_package'])
@require_wapt_db
def merge_stripped_package():
    """Handle the streamed upload of multiple packages

    Args:
        POST x-form-data body is zipped package data

    Returns:
        Response: json message with keys:
                    'success' (bool)
                    'error_code' if success is False
                    'msg' (str)
                    'result' (data)
                    'request_time' (float)
    """
    try:
        starttime = time.time()
        old_package_uuid = request.args.get('old_package_uuid')
        if not old_package_uuid:
            raise EWaptMissingParameter('missing old_package_uuid parameter')

        # load trusted signers
        trusted_signers = read_trusted_certificates(app.conf.get('trusted_signers_certificates_folder'))
        acls = get_user_acls(session['user'])

        r = get_repo_packages()
        pe = r.get_package_by_uuid(old_package_uuid)
        if pe:
            wapt_folder = app.conf['wapt_folder']
            tmp_build_dir = tempfile.mktemp(dir=wapt_folder, prefix='wapt')+'.tmp'
            os.makedirs(tmp_build_dir)
            try:
                # read new stripped down newly signed packages from POST data stream
                package_stream = PostDataStream(request.stream)
                try:
                    stripped_down_package = read_package(package_stream,trusted_signers,acls,move_to_repository=False,tmp_upload_dir=tmp_build_dir)
                finally:
                    package_stream = None
                # unzip
                pe.unzip_package(tempfile.mktemp(dir=tmp_build_dir, prefix='wapt')+'.tmp')
                # merge stripped into unzipped sources
                pe.merge_stripped_package(stripped_package_filename = stripped_down_package.localpath)
                # check if merge is ok
                if trusted_signers:
                    pe.check_package_signature(trusted_signers)
                pe.build_package(target_directory=tmp_build_dir)
                # checksum has changed
                pe.filename = pe.make_package_filename(with_md5sum=True)
                new_pe_localpath = os.path.join(r.localpath,pe.filename)
                # move file to repository
                shutil.copyfile(pe.localpath,new_pe_localpath)
                # change local path in entry
                pe.localpath = new_pe_localpath
                # index the new file
                r._append_package_to_index(pe)
                # add it to DB
                (rec, _isnew) = Packages.update_from_control(pe, is_available = True)
                # extract icon
                r._extract_icon(pe)
                success = True
            finally:
                # cleanup
                shutil.rmtree(tmp_build_dir)
            msg = _('Stripped down package {} -> {}').format(old_package_uuid,repr(pe))
        else:
            success = False
            msg = 'old_package_uuid {} not found'.format(old_package_uuid)

        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))

        spenttime = time.time() - starttime
        return make_response(success=success,
                             result=pe.package_uuid,
                             msg=msg,
                             request_time=spenttime)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('merge_stripped_package failed %s' % (repr(e)))
        return make_response_from_exception(e)

@app.route('/api/v3/upload_packages', methods=['HEAD', 'POST'])
@requires_auth(['admin', 'edit_host_package', 'edit_base_package', 'edit_group_package',
                'edit_unit_package', 'edit_wsus_package', 'edit_profile_package', 'edit_selfservice_package'])
@require_wapt_db
def upload_packages():
    """Handle the streamed upload of multiple packages

    Args:
        scanpackages : if 1, rescan packages, else only append new packages to packages index
        POST body are packages file handles

    Returns:
        Response: json message with keys:
                    'success' (bool)
                    'error_code' if success is False
                    'msg' (str)
                    'result' (data)
                    'request_time' (float)
    """

    if request.method == 'HEAD':
        return ''

    try:
        starttime = time.time()
        done = []
        errors = []
        errors_msg = []
        packages_index_result = None

        # load trusted signers
        trusted_signers = read_trusted_certificates(app.conf.get('trusted_signers_certificates_folder'))
        acls = get_user_acls(session['user'])

        if request.method == 'POST':
            if request.files:
                files = request.files
                # multipart upload
                logger.info('Upload of %s packages' % len(files))
                for fkey in files:
                    try:
                        packagefile = request.files[fkey]
                        logger.info('Uploading package file : %s' % fkey)
                        if packagefile and allowed_file(packagefile.filename):
                            new_package = read_package(packagefile, trusted_signers,acls)
                            # check if package signer is auhtorized for the hos
                            done.append(new_package)
                    except Exception as e:
                        logger.critical('Error uploading %s : %s' % (fkey, e))
                        errors.append(fkey)
                        errors_msg.append(str(e))
            else:
                # streamed upload
                packagefile = PostDataStream(request.stream)
                # check if package signer is auhtorized for the hos
                new_package = read_package(packagefile, trusted_signers,acls)
                done.append(new_package)

            if [e for e in done if e.section != 'host']:
                logger.debug('Update package index')
                if request.args.get('scanpackages', '1') == '1':
                    repo = get_repo_packages()
                    packages_index_result = repo.update_packages_index(proxies=get_http_proxies(app.conf))
                    if packages_index_result['errors']:
                        errors_msg.extend(packages_index_result['errors'])
                else:
                    packages_index_result = {'processed': [], 'kept': None, 'errors': None, 'packages_filename': None}
                    repo = get_repo_packages()
                    for pe in done:
                        packages_index_result['processed'].append(repo._append_package_to_index(pe))
            else:
                packages_index_result = None

        else:
            pass

        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))

        spenttime = time.time() - starttime
        return make_response(success=len(errors) == 0 and len(done) > 0,
                             result=dict(done=done, errors=errors, packages_index_result=packages_index_result),
                             msg=_('{} Packages uploaded, {} errors. Errors: {}').format(len(done), len(errors), '\n'.join(errors_msg)),
                             request_time=spenttime)

    except EWaptForbiddden as e:
        logger.debug(traceback.format_exc())
        logger.critical('upload_packages failed %s' % (repr(e)))
        return make_response_from_exception(e)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('upload_packages failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/upload_file', methods=['HEAD', 'POST'])
@requires_auth(['admin'])
def upload_file():
    if request.method == 'HEAD':
        return ''
    try:
        starttime = time.time()
        if request.method == 'POST':
            wapt_files_folder = os.path.join(app.conf['wapt_folder'],'files')
            if not os.path.isdir(wapt_files_folder):
                os.makedirs(wapt_files_folder)
            tmp_target = tempfile.mktemp(dir=wapt_files_folder, prefix='wapt')+'.tmp'

            # streamed upload
            logger.info('Start streaming down file to %s' % tmp_target)
            filestream = PostDataStream(request.stream)
            filestream.save(tmp_target)

            if request.args.get('filename'):
                target = os.path.join(wapt_files_folder,secure_filename(request.args.get('filename')))

                if os.path.isfile(target):
                    os.unlink(target)
                logger.info('Renaming package %s into %s' % (tmp_target, target))
                shutil.move(tmp_target, target)
            else:
                target = tmp_target

        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))

        spenttime = time.time() - starttime
        return make_response(success=os.path.isfile(target),
                             result=target,
                             msg=_('{} uploaded. Size {}').format(target,os.stat(target).st_size),
                             request_time=spenttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('upload_packages failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/upload_host', methods=['HEAD', 'POST'])
@app.route('/api/v3/upload_hosts', methods=['HEAD', 'POST'])
@requires_auth(['admin', 'edit_host_package'])
@require_wapt_db
def upload_host():
    """Handle the upload of multiple host packages

    Args:

    Returns:
        Response: json message with keys:
                    'success' (bool)
                    'error_code' if success is False
                    'msg' (str)
                    'result' (data)
                    'request_time' (float)
    """
    if request.method == 'HEAD':
        return ''
    try:
        starttime = time.time()
        done = []
        errors = []
        if request.method == 'POST':
            # load trusted signers
            trusted_signers = read_trusted_certificates(app.conf.get('trusted_signers_certificates_folder'))

            #
            files = list(request.files.keys())
            logger.info('Upload of %s host packages' % len(files))
            for fkey in files:
                hostpackagefile = request.files[fkey]
                logger.debug('uploading host file : %s' % fkey)
                if hostpackagefile and allowed_file(hostpackagefile.filename):
                    filename = secure_filename(hostpackagefile.filename)
                    wapt_host_folder = os.path.join(app.conf['wapt_folder'] + '-host')
                    target = os.path.join(wapt_host_folder, filename)
                    tmp_target = tempfile.mktemp(dir=wapt_host_folder, prefix='wapt')
                    with wapt_db.atomic() as trans:
                        try:
                            # if encrypted host packages, store the clear copy in a protected area for further edit...
                            if app.conf['encrypt_host_packages']:
                                ref_target = os.path.join(app.conf['wapt_folder'] + '-hostref', filename)
                                hostpackagefile.save(ref_target)
                                entry = PackageEntry(waptfile=ref_target)
                            else:
                                # write directly clear zip to tmp_targert
                                hostpackagefile.save(tmp_target)
                                entry = PackageEntry(waptfile=tmp_target)

                            # check if package signer is authorized
                            trusted_chain = check_valid_signer(entry, trusted_signers)
                            if not trusted_chain:
                                raise EWaptForbiddden('Package %s is not signed properly' % entry.make_package_filename())
                            logger.info('Package is trusted: %s' % trusted_chain)

                            trusted_fingerprints = [c.fingerprint for c in trusted_chain]
                            # check if package section or package name is authorized for this user
                            if not is_allowed_package_for_user(session['user'],session['user_acls'],entry,trusted_fingerprints ):
                                raise EWaptForbiddden('User %s is not allowed to modify this package' % (entry.package))


                            # update database for filtering on direct dependencies of a host
                            HostGroups.sync_from_host_package(entry)

                            # get host cert to encrypt package with public key
                            if app.conf['encrypt_host_packages']:
                                host_id = entry.package
                                host = Hosts.select(Hosts.uuid, Hosts.computer_fqdn, Hosts.host_certificate) \
                                    .where((Hosts.uuid == host_id) | (Hosts.computer_fqdn == host_id)) \
                                    .dicts().first()

                                # write encrypted package content
                                with open(tmp_target, 'wb') as encrypted_package:
                                    with open(ref_target, 'rb') as f:
                                        package_data = f.read()
                                    if host and host['host_certificate'] is not None:
                                        host_cert = SSLCertificate(crt_string=host['host_certificate'])
                                        encrypted_package.write(host_cert.encrypt_fernet(package_data))
                                    else:
                                        encrypted_package.write(package_data)

                            if os.path.isfile(target):
                                os.unlink(target)
                            shutil.move(tmp_target, target)
                            # fix context on target file (otherwith tmp context is carried over)
                            #logger.debug(subprocess.check_output('chcon -R -t httpd_sys_content_t %s' % target,shell=True))

                            done.append(filename)

                        except Exception as e:
                            logger.debug(traceback.format_exc())
                            logger.critical('Error uploading package %s: %s' % (filename, e))
                            errors.append(filename)
                            trans.rollback()
                            if os.path.isfile(tmp_target):
                                os.unlink(tmp_target)

        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
                repositories.update_file_tree_of_files(username=session.get('user'))
        else:
            pass

        spenttime = time.time() - starttime
        return make_response(success=len(errors) == 0,
                             result=dict(done=done, errors=errors),
                             msg=_('{} Host packages uploaded, {} errors').format(len(done), len(errors)),
                             request_time=spenttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('upload_host failed %s' % (repr(e)))
        return make_response_from_exception(e, status='201')


@app.route('/api/v3/upload_waptsetup', methods=['HEAD', 'POST'])
@app.route('/upload_waptsetup', methods=['HEAD', 'POST'])
@requires_auth(['admin'])
@require_wapt_db
def upload_waptsetup():
    """Handle the upload of customized waptagent.exe into wapt repository
    """
    if request.method == 'HEAD':
        return ''
    logger.debug('Entering upload_waptsetup')
    tmp_target = None
    try:
        if request.method == 'POST':
            file = request.files['file']
            set_default = request.args.get('set_default','') == '1'
            if file and (file.filename.startswith('waptsetup') or file.filename.startswith('waptagent')):
                filename = secure_filename(file.filename)
                tmp_target = os.path.join(app.conf['wapt_folder'], secure_filename('tmp.' + filename))
                target = os.path.join(app.conf['wapt_folder'], secure_filename(filename))
                file.save(tmp_target)
                if not os.path.isfile(tmp_target):
                    result = dict(status='ERROR', message=_('Problem during upload'))
                else:
                    shutil.move(tmp_target, target)
                    waptagent_default_fn = os.path.join(app.conf['wapt_folder'],'waptagent.exe')
                    if set_default and waptagent_default_fn != target:
                        if os.path.exists(waptagent_default_fn):
                            os.remove(waptagent_default_fn)
                        try:
                            os.symlink(target,waptagent_default_fn)
                        except:
                            shutil.copyfile(target,waptagent_default_fn)
                    result = dict(status='OK', message=_('{} uploaded').format((filename,)))
                    if app.conf['auto_create_waptagent_from_config']:
                        stuff_waptsetup()
                    global licences_list
                    if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
                        repositories.update_file_tree_of_files(username=session.get('user'))
            else:
                result = dict(status='ERROR', message=_('Wrong file name (version conflict?)'))
        else:
            result = dict(status='ERROR', message=_('Unsupported method'))
    except:
        e = sys.exc_info()
        if tmp_target and os.path.isfile(tmp_target):
            os.unlink(tmp_target)
        result = dict(status='ERROR', message=_('unexpected: {}').format((e,)))
    return Response(response=ujson.dumps(result),
                    status=200,
                    mimetype='application/json')


@app.route('/api/v3/change_password', methods=['HEAD', 'POST'])
@requires_auth(['admin'])
def change_password():
    """Handle change of admin master password"""
    if request.method == 'POST':
        try:
            post_data = request.get_json()
            if 'user' in post_data and 'password' in post_data:
                if check_auth(username=post_data['user'], password=post_data['password'], session=session, methods=['admin'], request=request):
                    # change master password
                    if 'new_password' in post_data and post_data['user'] == app.conf['wapt_user']:
                        if len(post_data['new_password']) < app.conf.get('min_password_length', 10):
                            raise EWaptForbiddden('The password must be at least %s characters' % app.conf.get('min_password_length', 10))
                        if app.conf['check_password_complexity']:
                            error = waptlicences.check_password_complexity(post_data['new_password'],True)
                            if error:
                                raise EWaptForbiddden('The password does not patch requirements: %s' % error)
                        change_admin_password(post_data['new_password'])
                        msg = 'Password for %s updated successfully' % post_data['user']
                    else:
                        raise EWaptMissingParameter('Bad or missing parameter')
                else:
                    raise EWaptAuthenticationFailure('Bad user or password')
            else:
                raise EWaptMissingParameter('Missing parameter')
            return make_response(result=msg, msg=msg, status=200)
        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.critical('change_password failed %s' % (repr(e)))
            time.sleep(3)
            return make_response_from_exception(e)


@app.route('/api/v3/login', methods=['HEAD', 'POST', 'GET'])
@app.route('/api/v3/login_kerberos', methods=['HEAD', 'POST', 'GET'])
@app.route('/login', methods=['HEAD', 'POST', 'GET'])
@add_logs
def login():
    if request.method == 'HEAD':
        return ''
    result = None
    starttime = time.time()
    try:
        user = None
        password = None
        post_data = None

        if request.method == 'POST':
            if request.headers.get('Content-Encoding') == 'gzip':
                raw_data = zlib.decompress(request.data)
            else:
                raw_data = request.data
            post_data = ujson.loads(raw_data)
            if post_data is not None:
                # json auth from waptconsole
                user = post_data.get('user')
                password = post_data.get('password')
        else:
            user = request.args.get('user')
            password = request.args.get('password')

        auth_result = check_auth(username=user, password=password, request=request, methods=app.conf['login_auth_methods'])

        if not auth_result:
            return authenticate()

        session.clear()
        # if basic auth, user was in authorization header
        user = auth_result['user']
        user_acls = {}

        result = dict(
            auth_result=auth_result,
            server_uuid=get_server_uuid(),
            version=__version__,
            server_domain=get_dns_domain(),
            edition=get_wapt_edition(),
        )

        with WaptDB():
            user_data = WaptUsers.get_or_none(name=user)
            # In case of ssl auth, check if the user certificate Sha1 passed by NGINX in the header matches the certificate we know already in DB.
            if auth_result['auth_method'] == 'ssl' and (not user_data or not user_data.user_fingerprint_sha1 or user_data.user_fingerprint_sha1.lower() != request.headers.get('X-Ssl-Client-Sha1', None).lower()):
                result = None
                raise Exception('Client certificate does not match user account')

            # we should autocreate admin. If auth_method=='ldap', it means user is in the wapt_admin_group_dn and wapt_admin_group
            # so we can autocreate it
            if not user_data and auth_result['auth_method'] == 'ldap' and ensure_list(app.conf['default_ldap_users_acls']):
                user_data, _user_created = WaptUsers.get_or_create(name=user)
                if _user_created:
                    user_data.created_on = datetime.datetime.utcnow()
                    user_data.created_by = 'AUTO-'+auth_result['auth_method']

            if user_data:
                user_data.last_login_timestamp = datetime.datetime.utcnow()
                user_data.save()

                # add ACL on new users
                (user_acls_rec, _created) = WaptUserAcls.get_or_create(username=user)
                if _created:
                    user_acls_rec.created_on = datetime.datetime.utcnow()
                    user_acls_rec.created_by = 'AUTO-' + auth_result['auth_method']
                    # if comes from ldap, and we have defined default ACL for user from ldap
                    if auth_result['auth_method'] == 'ldap' and app.conf['default_ldap_users_acls']:
                        user_acls_rec.acls = {}
                        for name in ensure_list(app.conf['default_ldap_users_acls'],allow_none=True):
                            user_acls_rec.acls[name] = True
                    user_acls_rec.save()

                user_acls = get_user_acls(username = user)

                try:
                    result['hosts_count'] = Hosts.select(fn.COUNT(Hosts.uuid)).tuples().first()[0]  # pylint: disable=no-value-for-parameter
                except:
                    result['hosts_count'] = None

        result['user_acls'] = user_acls

        get_token = (post_data and post_data.get('get_token',False) or request.args.get('get_token','0')=='1')
        if 'token' in app.conf['default_auth_methods'] or get_token:
            token_gen = get_secured_token_generator()
            token_data = dict(
                user=auth_result['user'],
                auth_method=auth_result['auth_method'],
                auth_date=auth_result['auth_date'],
                user_acls=user_acls,
                server_uuid=get_server_uuid(),
            )
            result['token'] = token_gen.dumps(token_data)
        else:
            result['token'] = None

        # we don't return a cookie if we have explicitly asked for a token in requests
        if 'session' in app.conf['default_auth_methods'] and not get_token:
            session['user_acls'] = user_acls
            session.update(**auth_result)

        # if we are authenticated but not with SSL and a temporary self-signed user certificate is pushed,
        # send back a temporary signed certificate by the computer CA for next auth.
        if post_data:
            client_csr_pem = post_data.get('client_csr')
            if client_csr_pem:
                client_csr = SSLCertificateSigningRequest(csr_pem_string = client_csr_pem)
                if client_csr.cn.lower() == user.lower():
                    temp_cert_lifetime_days = max(1,round(app.conf['session_lifetime'] / (24 * 3600)))  # session_lifetime is in seconds
                    signed_client_certificate = sign_host_csr(client_csr, lifetime_days = temp_cert_lifetime_days)
                    if signed_client_certificate:
                        # we append server client CA for reference
                        result['client_certificate'] =  b'\n'.join([signed_client_certificate.as_pem(),open(app.conf['clients_signing_certificate'],'rb').read()])
                else:
                    logger.warning('Ignoring CSR. CN of CSR "%s" is not matching user "%s"' % (client_csr.cn, user))

        msg = 'Authentication OK'
        spenttime = time.time() - starttime

        return make_response(result=result, msg=msg, status=200, request_time=spenttime)
    except Exception as e:
        if 'auth_token' in session:
            del session['auth_token']
        #logger.debug(traceback.format_exc())
        logger.critical('login failed %s' % (repr(e)))
        time.sleep(3)
        return make_response_from_exception(e)


@app.route('/api/v3/logout', methods=['HEAD', 'POST', 'GET'])
def logout():
    if request.method == 'HEAD':
        return ''
    session.clear()
    resp =  make_response(result=None, msg='logout', status=200)
    resp.delete_cookie('session')
    resp.set_cookie('session', '', expires=0)
    return resp


@app.route('/api/v3/login_self_service', methods=['POST'])
@app.route('/login_self_service', methods=['POST'])
@require_wapt_db
def login_self_service():
    """ Logs user in and returns the groups they belong to. For the option use_server_auth. """

    try:
        starttime = time.time()
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        post_data = ujson.loads(raw_data)
        if not post_data:
            raise Exception('login_self_service: No data supplied')

        user = post_data.get('user')
        password = post_data.get('password')
        groups = post_data.get('groups')

        uuid = post_data['uuid']
        if not uuid:
            raise Exception('login_self_service: No uuid supplied')

        computer_fqdn = (post_data.get('host_info', None) or post_data.get('host', {})).get('computer_fqdn', None)

        # get request signature
        signature_b64 = request.headers.get('X-Signature', None)
        if signature_b64:
            signature = base64.b64decode(signature_b64)
        else:
            signature = None

        if not signature:
            raise Exception('login_self_service: Missing signature')

        existing_host = Hosts.select(Hosts.uuid, Hosts.host_certificate, Hosts.computer_fqdn).where(Hosts.uuid == uuid).first()

        if not existing_host or not existing_host.host_certificate:
            raise EWaptMissingCertificate('You are trying to login from an unknown host {} ({}). Please register first.'.format(uuid, computer_fqdn))

        host_cert = SSLCertificate(crt_string=existing_host.host_certificate.encode('utf8'))
        if 'computer_fqdn' in post_data and not (host_cert.cn.lower() == computer_fqdn.lower() or host_cert.cn.lower() == uuid.lower()):
            raise EWaptUnknownHost('Supplied hostname or uuid does not match known certificate CN, aborting')

        logger.debug('About to check supplied data signature with certificate %s' % host_cert.cn)
        try:
            host_cert_cn = host_cert.verify_content(sha256_for_data(raw_data), signature)
        except Exception:
            # for pre 1.5 wapt clients
            host_cert_cn = 'unknown'
            if not app.conf['allow_unsigned_status_data']:
                raise
        logger.info('Data successfully checked with certificate CN %s for %s' % (host_cert_cn, uuid))
        if existing_host and not (host_cert_cn.lower() == existing_host.computer_fqdn.lower() or host_cert_cn.lower() == existing_host.uuid.lower()):
            raise Exception('login_self_service: mismatch between host certificate DN %s and existing host hostname %s' % (host_cert_cn, existing_host.computer_fqdn))

        if user and password and groups:
            global licences_list
            if licences_list.is_enterprise:
                istoken = False
                token_secret_key = app.conf['token_secret_key']
                if token_secret_key:
                    try:
                        token_gen = get_secured_token_generator(token_secret_key)
                        max_age = app.conf['token_lifetime']
                        token_data = token_gen.loads(password, max_age=max_age)
                        if 'kerb' in token_data['auth_method']:
                            istoken = True
                    except Exception as e:
                        logger.debug('Token verification failed : %s' % repr(e))

                if istoken:
                    result = auth_module_ad.check_credentials_ad(app.conf, token_data['user'], password, [], groups,use_service_account=True)
                else:
                    result = auth_module_ad.check_credentials_ad(app.conf, user, password, [], groups)
                msg = 'Self service authentication'
                spenttime = time.time() - starttime
                return make_response(result=result, msg=msg, status=200, request_time=spenttime)
            else:
                msg = 'Not enterprise'
                spenttime = time.time() - starttime
                return make_response(result={"success": False, 'groups': [], 'error': True, 'msg': msg}, msg=msg, status=200, request_time=spenttime)
        else:
            msg = 'Missing arguments'
            spenttime = time.time() - starttime
            return make_response(result={"success": False, 'groups': [], 'error': True, 'msg': msg}, msg=msg, status=200, request_time=spenttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('login self-service failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/packages_delete', methods=['HEAD', 'POST'])
@requires_auth(['admin', 'edit_host_package', 'edit_base_package', 'edit_group_package'])
@require_wapt_db
def packages_delete():
    """Removes a list of packages by filenames
    After removal, the repository package index "Packages" is updated.

    Args:
        POST body is a json list of packages filenames

    """
    if request.method == 'HEAD':
        return ''
    errors = []
    deleted = []


    # load trusted signers
    trusted_signers = read_trusted_certificates(app.conf.get('trusted_signers_certificates_folder'))

    if request.method == 'POST':
        filenames = request.get_json()
        listpackage = []
        for filename in filenames:
            try:
                if not allowed_file(filename):
                    raise EWaptForbiddden('Bad filename: %s' % filename)
                package_path = os.path.join(app.conf['wapt_folder'], secure_filename(filename))
                if os.path.isfile(package_path):
                    entry = PackageEntry(waptfile=package_path)

                    # check if package signer is authorized
                    trusted_chain = check_valid_signer(entry, trusted_signers)
                    if not trusted_chain:
                        raise EWaptForbiddden('Package %s is not signed properly' % entry.make_package_filename())
                    logger.info('Package is trusted: %s' % trusted_chain)

                    trusted_fingerprints = [c.fingerprint for c in trusted_chain]
                    # check if package section or package name is authorized for this user
                    if not is_allowed_package_for_user(session['user'],session['user_acls'],entry,trusted_fingerprints ):
                        raise EWaptForbiddden('User %s is not allowed to delete this package' % (entry.package))

                    listpackage.append(entry)
                    os.unlink(package_path)
                    Packages.update(is_available=False).where(Packages.package_uuid == entry.package_uuid).execute()
                    deleted.append(filename)
                else:
                    errors.append(filename)
            except Exception:
                errors.append(filename)

        # cleanup packages icons
        repo = get_repo_packages()
        update_result = repo.update_packages_index(proxies=get_http_proxies(app.conf))
        for package in listpackage:
            icons_path = os.path.abspath(os.path.join(repo.localpath, 'icons'))
            icon_fn = os.path.join(icons_path, "%s.png" % package.package)
            if os.path.isfile(icon_fn):
                remicon = True
                for c in update_result['kept']:
                    if package.package in c:
                        remicon = False
                        pass
                if remicon:
                    try:
                        os.unlink(icon_fn)
                    except:
                        pass
    else:
        pass
    msg = ['%s packages deleted' % len(deleted)]

    global licences_list
    if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
        repositories.update_file_tree_of_files(username=session.get('user'))

    if errors:
        msg.append('ERROR : %s packages could not be deleted' % len(errors))
    return make_response(result={'deleted':deleted,'errors':errors}, msg='\n'.join(msg), status=200)


@app.route('/wapt/')
def wapt_listing():
    return render_template(
        'listing.html', dir_listing=os.listdir(app.conf['wapt_folder']))


@app.route('/wapt/<string:input_package_name>')
def get_wapt_package(input_package_name):
    package_name = secure_filename(input_package_name)
    r = send_from_directory(app.conf['wapt_folder'], package_name)
    if 'content-length' not in r.headers:
        r.headers.add_header(
            'content-length', int(os.path.getsize(os.path.join(app.conf['wapt_folder'], package_name))))
    return r


@app.route('/wapt/icons/<string:iconfilename>')
def serve_icons(iconfilename):
    """Serves a png icon file from /wapt/icons/ test waptserver"""
    iconfilename = secure_filename(iconfilename)
    icons_folder = os.path.join(app.conf['wapt_folder'], 'icons')
    r = send_from_directory(icons_folder, iconfilename)
    if 'content-length' not in r.headers:
        r.headers.add_header(
            'content-length', int(os.path.getsize(os.path.join(icons_folder, iconfilename))))
    return r


@app.route('/static/css/<string:fn>')
@app.route('/static/fonts/<string:fn>')
@app.route('/static/img/<string:fn>')
@app.route('/static/js/<string:fn>')
def serve_static(fn):
    """Serve"""
    rootdir = os.path.join(wapt_root_dir, 'waptserver', 'static')
    if fn is not None:
        fn = request.path.split('/')[2:]
        r = send_from_directory(os.path.join(rootdir, secure_filename(fn[0])), secure_filename(fn[-1]))
        if 'content-length' not in r.headers:
            r.headers.add_header(
                'content-length', int(os.path.getsize(os.path.join(rootdir, fn))))
        return r


@app.route('/wapt-host/<string:input_package_name>')
def get_host_package(input_package_name):
    """Returns a host package (in case there is no apache static files server)"""
    # TODO straighten this -host stuff
    host_folder = app.conf['wapt_folder'] + '-host'
    package_name = secure_filename(input_package_name)
    r = send_from_directory(host_folder, package_name)
    if 'Content-Length' not in r.headers:
        r.headers.add_header(
            'Content-Length', int(os.path.getsize(os.path.join(host_folder, package_name))))
    return r


@app.route('/wapt-group/<string:input_package_name>')
def get_group_package(input_package_name):
    """Returns a group package (in case there is no apache static files server)"""
    # TODO straighten this -group stuff
    group_folder = app.conf['wapt_folder'] + '-group'
    package_name = secure_filename(input_package_name)
    r = send_from_directory(group_folder, package_name)
    # on line content-length is not added to the header.
    if 'content-length' not in r.headers:
        r.headers.add_header(
            'content-length',
            os.path.getsize(
                os.path.join(
                    group_folder +
                    '-group',
                    package_name)))
    return r


@app.route('/ping')
@app.route('/api/v3/ping')
def ping():
    global full_version
    if request.method == 'HEAD':
        return ''

    result=dict(
        uuid=get_server_uuid(),
        date=datetime2isodate(),
        )
    # only session cookie, to reduce ddos risks.
    if request.args.get('with_version')=='1' and get_request_auth_http_status(request,session,['session'])==200:
        result['version'] = __version__
        result['full_version'] = full_version

    return make_response(result = result)


@app.route('/api/v3/hosts_sid', methods=['GET','POST'])
def get_hosts_sid():
    if request.remote_addr != '127.0.0.1':
        return Response('forbidden',status=401)

    if request.method == 'GET':
        host_uuids = ensure_list(request.args.get('host_uuids'))
    else:
        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data
        post_data = ujson.loads(raw_data)
        host_uuids = post_data.get('host_uuids')

    if host_uuids:
        result = { k: v for k,v in app.ws_connections.items() if k in host_uuids and v[WS_SID] is not None }
    else:
        result = { k: v for k,v in app.ws_connections.items() if v[WS_SID] is not None }
    return make_response(
        msg=_('%s Connected hosts SID') % len(result), result=result
    )


@app.route('/api/v3/reset_hosts_sid', methods=['GET', 'HEAD', 'POST'])
@requires_auth(['admin', 'view'])
def reset_hosts_sid():
    """Launch a separate thread to check all reachable IP and update database with results.
    """
    if request.method == 'HEAD':
        return ''
    try:
        if not socketio:
            raise Exception('socketio unavailable')

        # in case a POST is issued with a selection of uuids to scan.
        if request.json is not None:
            uuids = request.json.get('uuids', None) or None
        else:
            uuids = None

        if uuids is not None:
            message = _('Hosts connection reset launched for %s host(s)' % len(uuids))
        else:
            message = _('Hosts connection reset launched for all hosts')

        def target(uuids):
            logger.debug('Reset websockets SID and timestamps of hosts')
            if uuids:
                for uuid in uuids:
                    if uuid in app.ws_connections:
                        ws = app.ws_connections[uuid]
                        # force a wapt ping/pong to trigger a get token
                        oldsid = ws[WS_SID]
                        ws[WS_SID] = None
                        ws[WS_TIMESTAMP] = datetime.datetime.utcnow()
                        if oldsid:
                            socketio.emit('wapt_ping', room=oldsid)

            else:
                for ws in app.ws_connections.values():
                    ws[WS_SID] = None
                    ws[WS_TIMESTAMP] = datetime.datetime.utcnow()
                socketio.emit('wapt_ping')

        socketio.start_background_task(target=target, uuids=uuids)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('reset_hosts_sid failed %s' % (repr(e)))
        return make_response_from_exception(e)
    return make_response(msg=message)

@app.route('/api/v3/update_hosts_sid_table')
def update_hosts_sid_table():
    if request.remote_addr != '127.0.0.1':
        return Response('forbidden',status=401)

    if not update_ws_db:
        raise Exception('update_hosts_sid_table socketio unavailable')

    with wapt_db:
        result = update_ws_db()
    return make_response(msg='HostWebsocket table updated. %s connected hosts' % result,result=result)

def get_broadcast_mac(host_uuids):
    """Returns dict of {broadcast: [mac1,mac2..]} for machines with uuid in host_uuids list
    """
    with wapt_db:
        hosts_data = Hosts\
            .select(Hosts.uuid, SQL("host_networking->'interfaces' as interfaces"), SQL("host_info->'networking' as host_info_networking")) \
            .where(Hosts.uuid.in_(host_uuids)) \
            .dicts()

        result = {}
        for host in hosts_data:
            interfaces = host.get('interfaces',[]) or []
            if host.get('host_info_networking'):
                interfaces.extend(host.get('host_info_networking',[]))
            for interface in interfaces:
                mac = interface.get('mac','')
                if len(mac) != 17:
                    continue
                addr = interface.get('addr',[   {'addr': '0.0.0.0',
                                                 'broadcast': '255.255.255.255',
                                                 'connected': False,
                                                 'netmask': '0.0.0.0'}])
                for a in addr:
                    if not isinstance(a,dict):
                        broadcast = "255.255.255.255"
                    else:
                        broadcast = a.get('broadcast')
                    if not broadcast or not '.' in broadcast:
                        continue
                    if not broadcast in result:
                        result[broadcast] = [mac]
                    else:
                        result[broadcast].append(mac)
        return result

@app.route('/api/v3/hosts_mac_broadcast', methods=['GET','POST'])
@requires_auth(['admin', 'view', 'trigger_host_action'])
@require_wapt_db
def hosts_mac_broadcast():
    if request.method == 'GET':
        host_uuids = ensure_list(request.args.get('uuids'))
    else:
        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data
        post_data = ujson.loads(raw_data)
        host_uuids = post_data.get('uuids')

    data = get_broadcast_mac(host_uuids)
    return make_response(msg=_('Broadcast address and mac address list for %s hosts' % len(host_uuids)), result=data)

@app.route('/api/v3/trigger_wakeonlan', methods=['HEAD', 'POST'])
@requires_auth(['admin', 'trigger_host_action'])
@require_wapt_db
def trigger_wakeonlan():
    if request.method == 'HEAD':
        return ''
    try:
        uuids = request.get_json()['uuids']
        broadcast_macs = get_broadcast_mac(uuids)

        all_macs = []
        for macs in broadcast_macs.values():
            if macs:
                all_macs.extend(macs)
        unique_macs = list(set(all_macs))

        for port in app.conf['wol_port'].split(','):
            logger.debug('Send wakeonlan packet for %s on broadcast port %s' % (unique_macs,port))
            wakeonlan.send_magic_packet(*unique_macs, port=int(port.strip()))
            for broadcast_ip,macs in broadcast_macs.items():
                logger.debug('Send wakeonlan packet for %s on broadcast %s:%s' % (macs,broadcast_ip,port))
                wakeonlan.send_magic_packet(
                    * macs,
                    ip_address = broadcast_ip,
                    port=int(port.strip()))
        msg = _('Wakeonlan packets sent to {} machines.').format(len(uuids))
        return make_response(all_macs,
                             msg=msg,
                             success=True)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('trigger_wakeonlan failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/waptagent_version')
@app.route('/api/v2/waptagent_version')
@requires_auth(['admin', 'view'])
def waptagent_version():
    try:
        start = time.time()
        waptagent = os.path.join(app.conf['wapt_folder'], 'waptagent.exe')
        agent_present, agent_version = get_wapt_exe_version(waptagent)
        waptagent_timestamp = None
        agent_sha256 = None
        if agent_present and agent_version is not None:
            agent_sha256 = sha256_for_file(waptagent)
            waptagent_timestamp = datetime2isodate(
                datetime.datetime.fromtimestamp(
                    os.path.getmtime(waptagent)))

        waptsetup = os.path.join(app.conf['wapt_folder'], 'waptagent', 'waptsetup.exe')
        setup_present, setup_version = get_wapt_exe_version(waptsetup)
        waptsetup_timestamp = None
        if setup_present and setup_version is not None:
            waptsetup_timestamp = datetime2isodate(
                datetime.datetime.fromtimestamp(
                    os.path.getmtime(waptsetup)))

        if agent_present and setup_present and Version(
                agent_version) >= Version(setup_version):
            msg = 'OK : waptagent.exe %s >= waptsetup %s' % (
                agent_version, setup_version)
        elif agent_present and setup_present and Version(agent_version) < Version(setup_version):
            msg = 'Problem : waptagent.exe %s is older than waptsetup %s, and must be regenerated.' % (
                agent_version, setup_version)
        elif not agent_present and setup_present:
            msg = 'Problem : waptagent.exe not found. It should be compiled from waptconsole.'
        elif not setup_present:
            msg = 'Problem : waptsetup.exe not found on repository.'

        result = dict(
            waptagent_version=agent_version,
            waptagent_sha256=agent_sha256,
            waptagent_timestamp=waptagent_timestamp,
            waptsetup_version=setup_version,
            waptsetup_timestamp=waptsetup_timestamp,
            request_time=time.time() - start,
        )
    except Exception as e:
        return make_response_from_exception(e)

    return make_response(result=result, msg=msg, status=200)


@app.route('/api/v3/trigger_cancel_task')
@requires_auth(['admin', 'trigger_host_action'])
def host_cancel_task():
    if not socketio:
        raise Exception('socketio not available')
    return proxy_host_request(request, 'trigger_cancel_task')


@app.route('/api/v3/get_current_user_acls')
@requires_auth(['admin', 'view'])
@require_wapt_db
def get_current_user_acls():
    """Return dict of acls for the current user
    """
    try:
        username = session.get('user')
        acls = get_user_acls(username)
        session['user_acls'] = acls
        msg = 'Acls for logged in user {}'.format(username)

    except Exception as e:
        return make_response_from_exception(e)

    return make_response(result=acls, msg=msg, status=200)

@app.route('/api/v3/users_acls', methods=['HEAD', 'GET'])
@requires_auth(['admin'])
@require_wapt_db
def get_all_users_acls():
    """Return dict of acls for the current user
    """
    try:
        acls = list(WaptUserAcls.select(WaptUserAcls,
            WaptUsers.id.alias('_waptusers_id'),
            WaptUsers.name.alias('_waptusers_name'),
            WaptUsers.description.alias('_waptusers_description'),
            WaptUsers.user_certificate.alias('_waptusers_user_certificate'),
            WaptUsers.user_fingerprint.alias('_waptusers_user_fingerprint'),
            WaptUsers.user_fingerprint_sha1.alias('_waptusers_user_fingerprint_sha1'),
            WaptUsers.last_login_timestamp.alias('_waptusers_last_login_timestamp'),
            WaptUsers.auth_method.alias('_waptusers_auth_method')
        ).join(WaptUsers,'LEFT JOIN',on=(WaptUsers.name == WaptUserAcls.username)).order_by(WaptUsers.name).dicts())
        # add a bool to flag user with local password in htpasswd
        htpasswd_users = list_user_password_htpasswd()
        for acl in acls:
            if acl['_waptusers_name'] in htpasswd_users:
                acl['_is_local_user'] = True
        msg = 'Acls for all users'

    except Exception as e:
        return make_response_from_exception(e)

    return make_response(result=acls, msg=msg, status=200)

@app.route('/api/v3/global_config', methods=['HEAD', 'GET'])
@requires_auth(['admin','view'])
@require_wapt_db
def global_config():
    """Return dict of centralized parameters
    """
    try:
        result = {
            'server_uuid': app.conf['server_uuid'],
            'use_kerberos': app.conf['use_kerberos'],
            'waptwua_enable': app.conf['waptwua_enable'],
            'wads_enable': app.conf['wads_enable'],
            'reporting_enable': app.conf['reporting_enable'],
            'repo_rules_enable': app.conf['repo_rules_enable'],
            'repo_sync_enable': app.conf['repo_sync_enable'],
        }
        if app.conf['check_password_complexity']:
            result.update(
                min_password_length=app.conf['min_password_length'],
                check_password_complexity=app.conf['check_password_complexity']
            )
        if app.conf.get('secure_mode'):
            result['secure_mode'] = app.conf['secure_mode']
        msg = 'Server side global config'
    except Exception as e:
        return make_response_from_exception(e)

    return make_response(result=result, msg=msg, status=200)

@app.route('/api/v3/users_acls', methods=['POST','DELETE'])
@requires_auth(['admin'])
@require_wapt_db
def post_users_acls():
    """Return dict of acls for the current user
    """
    try:
        with wapt_db.atomic() as trans:
            try:
                # unzip if post data is gzipped
                if request.headers.get('Content-Encoding') == 'gzip':
                    raw_data = zlib.decompress(request.data)
                else:
                    raw_data = request.data

                post_data = ujson.loads(raw_data)
                if not post_data:
                    raise Exception('post_users_acls: No data supplied')

                result = {}
                msg = 'Records updated'

                def filter_row_attributes(row: dict,aclass) -> dict:
                    result = {}
                    for att,value in row.items():
                        if hasattr(aclass,att):
                            result[att] = value
                    return result


                for row in post_data:
                    # 0: usUnmodified, 1: usModified, 2: usInserted, 3: usDeleted
                    us = row['_update_status']
                    del row['_update_status']
                    id = row['id']
                    if us == 2:
                        if int(row['id']) < 0:
                            del row['id']
                        r = WaptUserAcls.insert(**filter_row_attributes(row,WaptUserAcls))
                    elif us == 3:
                        r = WaptUserAcls.delete().where(WaptUserAcls.id == row['id'])
                    else:
                        r = WaptUserAcls.update(**filter_row_attributes(row,WaptUserAcls)).where(WaptUserAcls.id == row['id'])

                    result[id] = r.execute()
                    if us in [1,2]:
                        r2 = WaptUsers.insert(
                                name=row['username'],
                                user_certificate=row.get('_waptusers_user_certificate'),
                                user_fingerprint=row.get('_waptusers_user_fingerprint'),
                                user_fingerprint_sha1=row.get('_waptusers_user_fingerprint_sha1'),
                                created_on=row['created_on'],
                                created_by=row['created_by']
                            ).on_conflict(
                                conflict_target='name',
                                update={
                                    'user_certificate':row.get('_waptusers_user_certificate'),
                                    'user_fingerprint':row.get('_waptusers_user_fingerprint'),
                                    'user_fingerprint_sha1':row.get('_waptusers_user_fingerprint_sha1'),
                                    'updated_on':row.get('updated_on',row.get('created_on')),
                                    'updated_by':row.get('updated_by',row.get('created_by'))
                                    }
                            )
                        r2.execute()

            except Exception as e:
                logger.debug(traceback.format_exc())
                logger.critical('Update alcs failed %s' % (repr(e)))
                trans.rollback()
                return make_response_from_exception(e)

        return make_response(result=result, msg=msg, status=200)

    except Exception as e:
        return make_response_from_exception(e)



@app.route('/api/v3/groups')
@app.route('/api/v1/groups')
@requires_auth(['admin', 'view'])
@require_wapt_db
def get_groups():
    """List of packages having section == group
    """
    try:
        groups = list(HostGroups.select(fn.DISTINCT(HostGroups.group_name).alias('package')).order_by(1).dicts())
        msg = '{} Packages for section group'.format(len(groups))

    except Exception as e:
        return make_response_from_exception(e)

    return make_response(result=groups, msg=msg, status=200)


@app.route('/api/v3/get_ad_groups')
@requires_auth(['admin', 'view'])
@require_wapt_db
def get_ad_groups():
    """list of active directory computers groups

    Result:
        Response {"msg": "4 active directory computers groups", "result": ["xxx", "yyy", "ttt"], "success": true, "request_time": null}
    """
    starttime = time.time()
    try:
        adgroups = Hosts.select(fn.unnest(Hosts.computer_ad_groups).alias('groups')).where(~Hosts.computer_ad_groups.is_null()).distinct()
        groups = [r[0] for r in list(Select(columns=[adgroups.c.groups]).from_(adgroups).distinct().tuples().bind(wapt_db))]
        msg = '{} active directory computers groups'.format(len(groups))

    except Exception as e:
        return make_response_from_exception(e)

    return make_response(result=groups, msg=msg, status=200, request_time=time.time() - starttime)


@app.route('/api/v3/get_ad_ou')
@requires_auth(['admin', 'view'])
@require_wapt_db
def get_ad_ou():
    """List all the OU registered by hosts
    """
    try:
        starttime = time.time()
        result = [r[0] for r in Hosts.select(
            Hosts.computer_ad_ou,
            fn.COUNT(Hosts.uuid))
            .where(
            ~Hosts.computer_ad_ou.is_null())
            .group_by(Hosts.computer_ad_ou)
            .tuples()
        ]

        message = 'AD OU DN List'
        return make_response(result=result, msg=message, request_time=time.time() - starttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('get_ad_ou failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/get_ad_ou_split')
@requires_auth(['admin', 'view'])
@require_wapt_db
def get_ad_ou_split():
    """List all the OU registered by hosts
    """
    try:
        starttime = time.time()
        req = [r[0] for r in Hosts.select(
            Hosts.computer_ad_ou,
            fn.COUNT(Hosts.uuid))
            .where(
            ~Hosts.computer_ad_ou.is_null())
            .group_by(Hosts.computer_ad_ou)
            .tuples()
        ]
        result = []
        for elem in req:
            if elem:
                try:
                    result.append(ldap3.utils.dn.parse_dn(elem))
                except:
                    # be tolerant to malformed OU set by clients
                    pass

        message = 'AD OU DN List'
        return make_response(result=result, msg=message, request_time=time.time() - starttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('get_ad_ou failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/get_ad_sites')
@requires_auth(['admin', 'view'])
@require_wapt_db
def get_ad_sites():
    """List all the AD Sites registered by hosts
    """
    try:
        starttime = time.time()
        result = [r[0] for r in Hosts.select(fn.DISTINCT(Hosts.computer_ad_site)).where(~Hosts.computer_ad_site.is_null()).tuples()]  # pylint: disable=too-many-function-args

        message = 'AD Sites List'
        return make_response(result=result, msg=message, request_time=time.time() - starttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('get_ad_sites failed %s' % (repr(e)))
        return make_response_from_exception(e)


def build_hosts_filter(model, filter_expr):
    """Legacy helper function to translate waptconsole <=1.3.11 hosts filter
        into peewee model where clause.
    Args:
        filter_expr (str) : field1,field4,field5:search_regexp
    """
    (search_fields, search_expr) = filter_expr.split(':', 1)
    if search_expr.startswith('not ') or search_expr.startswith('!'):
        not_filter = 1
        if search_expr.startswith('not '):
            search_expr = search_expr.split(' ', 1)[1]
        else:
            search_expr = search_expr[1:]
    else:
        not_filter = 0

    if search_fields.strip() and search_expr.strip():
        result = None
        for fieldname in ensure_list(search_fields):
            members = fieldname.split('.')
            rootfield = members[0]
            # external collections...
            if rootfield == 'installed_softwares':
                clause = Hosts.uuid.in_(HostSoftwares.select(HostSoftwares.host).where(
                    HostSoftwares.key.regexp(r'(?i)%s' % search_expr) | (HostSoftwares.name+' ('+HostSoftwares.version+')').regexp(r'(?i)%s' % search_expr)))
            elif rootfield == 'installed_packages':
                clause = Hosts.uuid.in_(HostPackagesStatus.select(HostPackagesStatus.host).where(HostPackagesStatus.package.iregexp(r'%s' % search_expr)))
            elif rootfield in model._meta.fields:
                if isinstance(model._meta.fields[rootfield], (JSONField, BinaryJSONField)):
                    if len(members) == 1:
                        clause = SQL("%s::text ~* '%s'" % (fieldname, search_expr))
                    else:
                        # (wapt->'waptserver'->'dnsdomain')::text ~* 'asfrance.lan'
                        clause = SQL("(%s->%s)::text ~* '%s'" % (rootfield, '->'.join(["'%s'" % f for f in members[1:]]), search_expr))
                        # model._meta.fields[members[0]].path(members[1:]).regexp(ur'(?i)%s' % search_expr)
                elif isinstance(model._meta.fields[rootfield], ArrayField):
                    clause = SQL("%s::text ~* '%s'" % (fieldname, search_expr))
                else:
                    if '|' not in search_expr:
                        clause = model._meta.fields[fieldname].iregexp(r'%s' % search_expr)
                    else:
                        search_names = [s_name.strip() for s_name in search_expr.split('|')]
                        condition = r'({})'.format(r'|'.join(search_names))
                        clause = model._meta.fields[fieldname].iregexp(r'%s' % condition)

            # else ignored...
            else:
                clause = None

            if result is None:
                result = clause
            else:
                if clause is not None:
                    result = result | clause
        if result is not None and not_filter:
            result = ~result  # pylint: disable=invalid-unary-operand-type
        return result
    else:
        raise Exception('Invalid filter provided in query. Should be f1,f2,f3:regexp ')


@app.route('/api/v3/hosts_delete', methods=['HEAD', 'POST'])
@requires_auth(['admin', 'unregister_host'])
@require_wapt_db
def hosts_delete():
    """Remove one or several hosts from Server DB and optionnally the host packages

    Args:
        uuids (list) : list of uuids to delete
        filter (csvlist of field:regular expression): filter based on attributes
        delete_packages (bool) : delete host's packages
        delete_inventory (bool) : delete host's inventory

    Returns:
        result (dict):

    """
    msg = []
    result = dict(files=[], records=[])

    if request.method == 'HEAD':
        return ''

    if request.method == 'POST':
        with wapt_db.atomic() as trans:
            try:
                # unzip if post data is gzipped
                if request.headers.get('Content-Encoding') == 'gzip':
                    raw_data = zlib.decompress(request.data)
                else:
                    raw_data = request.data

                post_data = ujson.loads(raw_data)
                if not post_data:
                    raise Exception('unregister_host: No data supplied')

                if 'uuids' in post_data:
                    query = Hosts.uuid.in_(ensure_list(post_data['uuids']))
                elif 'filter' in post_data:
                    query = build_hosts_filter(Hosts, post_data['filter'])
                else:
                    raise Exception('Neither uuid nor filter provided in query')

                if 'delete_packages' in post_data and post_data['delete_packages']:
                    selected = Hosts.select(Hosts.uuid, Hosts.computer_fqdn).where(query)
                    for host in selected:
                        result['records'].append(
                            dict(
                                uuid=host.uuid,
                                computer_fqdn=host.computer_fqdn))

                        if host.uuid:
                            uuid_hostpackage = os.path.join(app.conf['wapt_folder'] + '-host', host.uuid+'.wapt')

                        if host.computer_fqdn:
                            fqdn_hostpackage = os.path.join(app.conf['wapt_folder'] + '-host', host.computer_fqdn + '.wapt')

                        if os.path.isfile(uuid_hostpackage):
                            logger.debug('Trying to remove %s' % uuid_hostpackage)
                            if os.path.isfile(uuid_hostpackage):
                                os.remove(uuid_hostpackage)
                                result['files'].append(uuid_hostpackage)

                        if os.path.isfile(fqdn_hostpackage):
                            logger.debug('Trying to remove %s' % fqdn_hostpackage)
                            if os.path.isfile(fqdn_hostpackage):
                                os.remove(fqdn_hostpackage)
                                result['files'].append(fqdn_hostpackage)

                    msg.append(
                        '{} files removed from host repository'.format(len(result['files'])))

                if 'delete_inventory' in post_data and post_data['delete_inventory']:
                    if app.conf['clients_signing_crl']:
                        host_certs = Hosts.select(Hosts.host_certificate).where(query)
                        for host in host_certs:
                            if host.host_certificate:
                                cert = SSLCertificate(crt_string=host.host_certificate)
                                revoke_cert(cert)

                    remove_result = Hosts.delete().where(query).execute()

                    # force disconnect on connected remote hosts
                    if 'uuids' in post_data:
                        for uuid in ensure_list(post_data['uuids']):
                            if uuid in app.ws_connections:
                                ws = app.ws_connections[uuid]
                                oldsid = ws[WS_SID]
                                del app.ws_connections[uuid]
                                if oldsid:
                                    socketio.emit('wapt_ping', room=oldsid)

                    msg.append('{} hosts removed from DB'.format(remove_result))

            except Exception as e:
                logger.debug(traceback.format_exc())
                logger.critical('hosts_delete failed %s' % (repr(e)))
                trans.rollback()
                return make_response_from_exception(e)

        return make_response(result=result, msg='\n'.join(msg), status=200)


def build_fields_list(model, columns):
    """Returns (list of peewee fields,alias_map)
    """
    removed_columns = [
        'reachable',
        'listening_timestamp',
        'listening_address',
        'listening_protocol'
    ]

    result = []
    alias_map = {}
    for fname in columns:
        if fname in removed_columns:
            continue
        if isinstance(fname, SQL):
            result.append(fname)
        elif isinstance(fname,str) and fname in model._meta.fields:
            result.append(model._meta.fields[fname])
        elif fname == 'depends':
            # subquery with result aggregation.
            result.append(HostGroups.select(fn.STRING_AGG(HostGroups.group_name, ',')).where(
                (HostGroups.host_id == Hosts.uuid)
            ).alias('depends'))
        elif (not isinstance(fname,list) and fname.startswith('host_audit/')) or fname[0].startswith('host_audit/'):
            #if fname is a list, the first item is the expression, the second one is the alias for the result.
            if isinstance(fname,list):
                fname,alias = fname
            else:
                fname,alias = fname,fname

            if len(alias)>60:
                actual_alias = alias
                alias = hashlib.md5(alias.encode('utf8')).hexdigest()
                alias_map[alias] = actual_alias

            # extract the most up to date audit data
            #   field name is named "host_audit/section/key" where section and key are  the data to restrieve.
            _,section,key = fname.split('/',2)
            if '/' in key:
                key, subpath = key.split('/',1)
                parts = subpath.split('/')
                if len(parts)>1: # json...
                    path = ','.join(parts[1:])
                    value_fn = SQL("%s #>'{%s}' as \"%s\" " % (parts[0], path, alias))
                else: # plain field
                    value_fn = SQL("%s as \"%s\" " % (parts[0],alias))
            else:
                value_fn = SQL("""jsonb_build_object('value_id',value_id,'value_section',value_section,'value_key',value_key,'value',value) as \"%s\" """ % alias)
            field = HostAuditData.select(value_fn).where(
                    (HostAuditData.host == Hosts.uuid) &
                    (HostAuditData.value_section == section) &
                    (HostAuditData.value_key == key)).order_by(HostAuditData.value_date.desc()).limit(1).alias(alias)

            result.append(field)
        else:
            # jsonb sub fields.
            if isinstance(fname,list):
                fname,alias = fname
            else:
                fname,alias = fname,fname

            parts = fname.split('/')
            root = parts[0]
            if root in model._meta.fields:
                path = ','.join(parts[1:])
                result.append(SQL("%s #>'{%s}' as \"%s\" " % (root, path, alias)))
    return (result,alias_map)


@app.route('/api/v3/reported_managed_hosts_count', methods=['GET'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def reported_managed_hosts_count():
    """ Get the number of managed hosts reported"""
    return make_response(result=Hosts.select().where((Hosts.server_uuid == get_server_uuid()) | Hosts.server_uuid.is_null()).count())


@app.route('/api/v3/host_packages_filters', methods=['HEAD', 'GET'])
@requires_auth(['admin', 'view'])
@require_wapt_db
@gzipped
def get_host_packages_filters():
    """Returns a list of all packages filter strings
    for registered hosts.
    Allow to detect packages which are not needed anymore or
    filter external repo packages to match the current registred hosts.
    """
    if request.method == 'HEAD':
        return ''
    try:
        result = []
        msg = ''
        start_time = time.time()
        packages_filter_list = Hosts.select(Hosts.host_packages_filter).distinct().tuples()

        result = [r[0] for r in packages_filter_list]

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('failed %s' % (repr(e)))
        return make_response_from_exception(e)

    return make_response(result=result, msg=msg, status=200, request_time=time.time() - start_time)



@app.route('/api/v3/hosts', methods=['HEAD', 'GET'])
@app.route('/api/v1/hosts', methods=['HEAD', 'GET'])
@requires_auth(['admin', 'view'])
@require_wapt_db
@gzipped
def get_hosts():
    """Get registration data of one or several hosts

    Args:
        has_errors (0/1): filter out hosts with packages errors
        need_upgrade (0/1): filter out hosts with outdated packages
        groups (csvlist of packages) : hosts with packages
        ad_groups (csvlist of ad group names) : hosts which belong to these AD groups
        ad_site (ad site name) : hosts which belong to these AD site
        columns (csvlist of columns) :
        uuid (csvlist of uuid): <uuid1[,uuid2,...]>): filter based on uuid
        filter (csvlist of field):regular expression: filter based on attributes
        not_filter (0,1):
        limit (int) : 1000
        trusted_certs_sha256 (csvlist): filter out machines based on their trusted package certs

    Returns:
        result (dict): {'records':[],'files':[]}

        query:
          uuid=<uuid>
        or
          filter=<csvlist of fields>:regular expression
    """
    if request.method == 'HEAD':
        return ''
    try:
        result = []
        msg = ''
        start_time = time.time()
        if request.method == 'GET':
            default_columns = ['reachable',
                               'host_status',
                               'last_update_status',
                               'wapt_version',
                               'computer_fqdn',
                               'dnsdomain',
                               'description',
                               'connected_users',
                               'manufacturer',
                               'productname',
                               'serialnr',
                               'last_seen_on',
                               'mac_addresses',
                               'connected_ips',
                               'uuid',
                               'server_uuid',
                               'host_info/domain_controller',
                               'host_info/domain_name',
                               'host_info/domain_controller_address',
                               'depends',
                               'computer_type',
                               'os_name',
                               'os_version',
                               'registration_auth_user',
                               'platform',
                               'repositories',
                               'host_capabilities',
                               'host_networking',
                               ]

            # minimum columns
            if request.args.get('no_default_columns')=='1':
                columns = []
            else:
                columns = ['uuid',
                           'server_uuid',
                           'host_status',
                           'last_seen_on',
                           'last_update_status',
                           'computer_fqdn',
                           'computer_name',
                           'description',
                           'dnsdomain',
                           'connected_users',
                           'registration_auth_user',
                           'wapt_version',
                           'host_capabilities',
                           'host_networking',
                           ]

            cols_arg = request.args.get(
                    'columns',
                    default_columns)
            if isinstance(cols_arg,str) and cols_arg.startswith('['):
                other_columns = json.loads(cols_arg)
            else:
                other_columns = ensure_list(cols_arg)


            # add request columns
            for fieldname in other_columns:
                if not fieldname in columns:
                    columns.append(fieldname)

            # add columns_software_version columns
            columns_software_version = []
            for col_software_normalized_name in columns:
                if isinstance(col_software_normalized_name,str) and col_software_normalized_name.startswith('software_version#'):
                    columns_software_version.append(col_software_normalized_name)


            not_filter = request.args.get('not_filter', '0') == '1'

            query = ((Hosts.server_uuid == get_server_uuid()) | Hosts.server_uuid.is_null())

            def and_query(q, not_filter=False):
                if query is None:
                    if not_filter:
                        return ~(q)
                    else:
                        return q
                else:
                    if not_filter:
                        return query & ~(q)
                    else:
                        return query & q

            user_acls = get_user_acls(session['user'])
            perimeters = has_any_right(user_acls,['admin','view'])
            if perimeters:
                if isinstance(perimeters,list):
                    certs_sub = None
                    for cert_fingerprint in perimeters:
                        if certs_sub is None:
                            certs_sub = Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
                        else:
                            certs_sub = certs_sub | Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
                    query = and_query(certs_sub)

                # build filter
                if 'query' in request.args:
                    if 'query_id' in request.args:
                        query = and_query(Hosts.uuid.in_(get_corresponding_column(request.args.get('query_id'), request.args.get('query'))))
                    else:
                        logger.critical('Missing query id')


                if 'platform' in request.args:
                    query = and_query(Hosts.platform.in_(ensure_list(request.args['platform'])))
                if 'uuid' in request.args:
                    query = and_query(Hosts.uuid.in_(ensure_list(request.args['uuid'])))
                elif 'filter' in request.args:
                    query = and_query(build_hosts_filter(Hosts, request.args['filter']), not_filter)

                # TODO : pylint does not like this block... raises 'Uninferable' object is not iterable
                if 'groups' in request.args:
                    groups = ensure_list(request.args.get('groups', ''))
                    in_group = HostGroups.select(HostGroups.host).where(HostGroups.group_name << groups)
                    query = and_query(Hosts.uuid << in_group, not_filter)

                if 'trusted_certs_sha256' in request.args:
                    trusted_certs_sha256 = ensure_list(request.args.get('trusted_certs_sha256', ''))
                    certs_sub = None
                    for cert_fingerprint in trusted_certs_sha256:
                        if certs_sub is None:
                            certs_sub = Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
                        else:
                            certs_sub = certs_sub | Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
                    query = and_query(certs_sub, not_filter)



                if 'ad_site' in request.args:
                    query = and_query(Hosts.computer_ad_site == request.args.get('ad_site'), not_filter)

                if 'ad_groups' in request.args and request.args.get('ad_groups'):
                    groups = request.args.get('ad_groups').split('||')
                    or_list = Hosts.computer_ad_groups.contains(groups[0])
                    for group in groups[1:]:
                        or_list = or_list | Hosts.computer_ad_groups.contains(group)
                    query = and_query(or_list, not_filter)

                if 'organizational_unit' in request.args:
                    ou_list = request.args.get('organizational_unit').split('||')
                    if request.args.get('include_childs_ou', '1') == '1':
                        or_list = Hosts.computer_ad_ou.endswith(ou_list[0])
                        for ou in ou_list[1:]:
                            or_list = or_list | Hosts.computer_ad_ou.endswith(ou)
                        query = and_query(or_list)
                    else:
                        query = and_query(Hosts.computer_ad_ou.in_(ou_list))

                if 'has_errors' in request.args and request.args['has_errors']:
                    query = and_query(Hosts.host_status == 'ERROR')
                if 'need_upgrade' in request.args and request.args['need_upgrade']:
                    query = and_query(Hosts.host_status.in_(['ERROR', 'TO-UPGRADE']))

                only_reachable = 'reachable' in request.args and (request.args['reachable'] == '1')
                if only_reachable:
                    query = and_query(Hosts.uuid.in_(connected_uuids()))

            else:
                query = Hosts.uuid.is_null()

            limit = int(request.args.get('limit', 1000))


            fields,alias_map = build_fields_list(Hosts, columns)

            lookup_columns = {
                'remote_reboot_allowed': SQL("(wapt_status->'remote_reboot_allowed') as remote_reboot_allowed"),
                'remote_shutdown_allowed' : SQL("(wapt_status->'remote_shutdown_allowed') as remote_shutdown_allowed"),
                'main_ip' : SQL("coalesce(host_networking#>'{main_ip,0}',host_info#>'{main_ip,0}') as main_ip"),
            }

            for (column,f) in lookup_columns.items():
                if request.args.get('no_default_columns') != '1' or column in columns:
                    fields.append(f)

            order_by = request.args.get('order_by')
            order_direction = request.args.get('order_direction',1)

            if order_by and order_by.isidentifier() and order_by in Hosts._meta.fields:
                if order_direction=='1': # descending
                    sql_order_by= SQL('"%s" desc NULLS LAST' % order_by)
                else:
                    sql_order_by= SQL('"%s" NULLS FIRST' % order_by)
            else:
                sql_order_by= SQL('last_seen_on desc NULLS LAST')

            req = Hosts.select(*fields).limit(limit).order_by(sql_order_by)
            if query:
                req = req.where(query)

            #logger.debug(str(req))

            request_dict = req.dicts()
            software_version_dicts = {}
            if columns_software_version:
                for name in columns_software_version:
                    software_normalized_name = base64.b64decode(name.split('software_version#')[-1]).decode('utf8')
                    software_version_cursor =  HostSoftwares.select(HostSoftwares.version ,HostSoftwares.host_id ).join(Normalization,'LEFT JOIN',
                    on=(HostSoftwares.key == Normalization.key) & (HostSoftwares.name == Normalization.original_name)).where( Normalization.normalized_name== software_normalized_name ).dicts()
                    if software_version_cursor:
                        for software_version in  software_version_cursor:
                            if not (software_version['host'],name) in software_version_dicts.keys() :
                                software_version_dicts[(software_version['host'],name)] = [' '+software_version['version']]
                            else:
                                software_version_dicts[(software_version['host'],name)].append(' '+software_version['version'])

            request_dict = []
            # filter ou unreachable
            # add connection status from ws_connections dict

            # TODO: get only needed connections ?
            # be sure to get ws_conenctions from the other websockets enabled server (in case of uwsgi)
            get_ws_connections()

            def remap_alias(host_dict):
                if alias_map:
                    for k,v in alias_map.items():
                        if k in host_dict:
                            host_dict[v] = host_dict[k]
                            del host_dict[k]

            if 'reachable' in columns or 'listening_timestamp' in columns or 'listening_address' in columns:
                for h in req.dicts():
                    remap_alias(h)
                    append_ws_status(h)
                    request_dict.append(h)
            else:
                if alias_map:
                    for h in req.dicts():
                        remap_alias(h)
                        request_dict.append(h)
                else:
                    request_dict = list(req.dicts())

            # merge version of installed software
            if software_version_dicts:
                for host in request_dict:
                   for name in columns_software_version:
                      try:
                        host[name] = software_version_dicts[(host['uuid'],name)]
                      except KeyError:
                        host[name] = []

            result = request_dict

            if 'uuid' in request.args:
                if len(result) == 0:
                    msg = 'No data found for uuid {}'.format(request.args['uuid'])
                else:
                    msg = 'host data fields {} returned for uuid {}'.format(
                        ','.join(columns),
                        request.args['uuid'])
            elif 'filter' in request.args:
                if len(result) == 0:
                    msg = 'No data found for filter {}'.format(
                        request.args['filter'])
                else:
                    msg = '{} hosts returned for filter {}'.format(
                        len(result),
                        request.args['filter'])
            else:
                if len(result) == 0:
                    msg = 'No data found'
                else:
                    msg = '{} hosts returned'.format(len(result))

        return make_response(
            result=result, msg=msg, status=200, request_time=time.time() - start_time)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('get_hosts failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/audit_data', methods=['GET', 'POST'])
@requires_auth(['admin', 'view', 'update_audit_data'])
@require_wapt_db
def audit_data():
    """

    Args:
        host_uuids (list of str): list of host uuid to query audit date for. If  None, returns all.
        limit (int): maximum number of records to return
        section (str) : if not none, returns only data from this section
        keys (list of str): keys to return. If None, returns all keys.
        latest_only (0/1): if 1, returns only the lastest information, else returns history too.
    """
    try:
        start_time = time.time()
        if request.method == 'GET':
            limit = int(request.args.get('limit','20000'))
            host_uuids = ensure_list(request.args.get('host_uuids'),allow_none=True)
            section = request.args.get('section')
            keys = ensure_list(request.args.get('keys'),allow_none=True)
            latest_only = request.args.get('latest_only','') == '1'
        else:
            # unzip if post data is gzipped
            if request.headers.get('Content-Encoding') == 'gzip':
                raw_data = zlib.decompress(request.data)
            else:
                raw_data = request.data
            post_data = ujson.loads(raw_data)
            limit = int(post_data.get('limit', '20000'))
            host_uuids = post_data.get('host_uuids')
            keys = ensure_list(post_data.get('keys'), allow_none=True)
            section = post_data.get('section')
            latest_only = str(post_data.get('latest_only')) == '1'

        where = True
        if host_uuids is not None:
            where = HostAuditData.host_id.in_(host_uuids)
        if section is not None:
            where = where & (HostAuditData.value_section == section)
        if keys is not None:
            where = where & (HostAuditData.value_key.in_(keys))

        if latest_only:
            query = HostAuditData.select(SQL('distinct on (host_id,value_section,value_key) t1.*,computer_fqdn,description'))
        else:
            query = HostAuditData.select(HostAuditData,Hosts.computer_fqdn,Hosts.description)

        query = query.join(Hosts,'left join')
        query = query.where(where).order_by(HostAuditData.host_id,HostAuditData.value_section,HostAuditData.value_key,HostAuditData.value_date.desc())

        if limit is not None:
            query = query.limit(limit)
        result = list(query.dicts())
        msg = '%s audit records' % len(result)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('audit_data failed %s' % (repr(e)))
        return make_response_from_exception(e)

    return make_response(result=result, msg=msg, status=200, request_time=time.time() - start_time)



@app.route('/api/v3/host_data')
@app.route('/api/v1/host_data')
@requires_auth(['admin', 'view'])
@require_wapt_db
def host_data():
    """
        Get additional data for a host
        query:
          uuid=<uuid>
          field=packages, dmi or softwares
    """
    try:
        start_time = time.time()

        # build filter
        if 'uuid' in request.args:
            uuid = request.args['uuid']
        else:
            raise EWaptMissingParameter('Parameter uuid is missing')

        if 'field' in request.args:
            field = request.args['field']
            if not field in list(Hosts._meta.fields.keys()) + ['installed_softwares', 'installed_packages', 'wsusupdates','audit_data', 'configurations']:  # pylint: disable=no-member
                raise EWaptMissingParameter('Parameter field %s is unknown' % field)
        else:
            raise EWaptMissingParameter('Parameter field is missing')

        if field == 'audit_data':
            data_key = request.args.get('data_key')
            data_section = request.args.get('data_section')
            latest_only = request.args.get('latest_only')

            limit = request.args.get('limit')

            where = HostAuditData.host == uuid

            if data_key is not None:
                where = where & (HostAuditData.data_key == data_key)
            if data_section is not None:
                where = where & (HostAuditData.data_section == data_section)

            if latest_only=='1':
                query = HostAuditData.select(SQL('distinct on (host_id,value_section,value_key) *'))
            else:
                query = HostAuditData.select()
            query = query.where(where).order_by(HostAuditData.host_id,HostAuditData.value_section,HostAuditData.value_key,HostAuditData.value_date.desc())

            if limit is not None:
                query = query.limit(limit)
            result = list(query.dicts())
        elif field == 'installed_softwares':
            result = list(HostSoftwares.select(HostSoftwares.id, HostSoftwares.host, HostSoftwares.software_id,
            SQL("""CASE WHEN (t1.name is null) or (t1.name = '') THEN t1.key ELSE t1.name END""").alias('name'),
            HostSoftwares.version, HostSoftwares.publisher, HostSoftwares.key, HostSoftwares.system_component,
            HostSoftwares.uninstall_string, HostSoftwares.install_date, HostSoftwares.install_location, Normalization.normalized_name)
            .join(Normalization,'LEFT JOIN',on=(HostSoftwares.key == Normalization.key) & (HostSoftwares.name == Normalization.original_name)).where(HostSoftwares.host == uuid).dicts())
        elif field == 'installed_packages':
            result = list(HostPackagesStatus.select().where(HostPackagesStatus.host == uuid).order_by(HostPackagesStatus.created_on.desc()).dicts())
        elif field == 'wsusupdates':
            local_status = HostWsus.alias('local_status')
            fields = [
                local_status.id.alias('local_status_id'),
                local_status.status.alias('local_status_status'),
                local_status.allowed.alias('local_status_allowed'),
                local_status.installed.alias('local_status_installed'),
                local_status.install_date.alias('local_status_install_date'),
                local_status.present.alias('local_status_present'),
                local_status.hidden.alias('local_status_hidden'),
                local_status.downloaded.alias('local_status_downloaded'),
                local_status.history.alias('local_status_history'),
                WsusUpdates.update_id,
                WsusUpdates.title,
                WsusUpdates.update_type,
                WsusUpdates.kbids,
                WsusUpdates.severity,
                WsusUpdates.changetime,
                WsusUpdates.product,
                WsusUpdates.classification,
                WsusUpdates.download_urls,
                WsusUpdates.min_download_size,
                WsusUpdates.max_download_size,
                WsusUpdates.superseded_update_ids,
                WsusUpdates.security_bulletin_ids,
                WsusUpdates.is_mandatory,
                WsusUpdates.reboot_behaviour,
                WsusUpdates.can_request_user_input,
                WsusUpdates.requires_network_connectivity,
                WsusUpdates.languages,
                WsusUpdates.created_on,
                WsusUpdates.updated_on,
                WsusUpdates.downloaded_on,
                WsusUpdates.is_beta,
                WsusUpdates.release_notes,
                WsusUpdates.is_uninstallable,
                WsusUpdates.update_type,
                WsusUpdates.installation_impact,
                WsusUpdates.uninstallation_impact,
            ]
            result = list(local_status.select(*fields).join(WsusUpdates, 'RIGHT OUTER JOIN').where(local_status.host == uuid).order_by(WsusUpdates.changetime).dicts())  # pylint: disable=no-member
        elif field == 'configurations':
            result = HostConfigs.select(HostConfigs.config, HostConfigs.priorities).where(HostConfigs.host == uuid).dicts().first()
        else:
            data = Hosts\
                .select(Hosts.uuid, Hosts.computer_fqdn, Hosts.fieldbyname(field))\
                .where(Hosts.uuid == uuid)\
                .dicts()\
                .first()

            if data is None:
                raise EWaptUnknownHost(
                    'Host {} not found in database'.format(uuid))
            result = data.get(field, None)
        msg = '{} data for host {}'.format(field, uuid)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('host_data failed %s' % (repr(e)))
        return make_response_from_exception(e)

    if result is None:
        msg = 'No {} data for host {}'.format(field, uuid)
        success = False
        error_code = 'empty_data'
    else:
        success = True
        error_code = None

    return make_response(result=result, msg=msg, success=success,
                         error_code=error_code, status=200, request_time=time.time() - start_time)


@app.route('/api/v3/hosts_count_for_package', methods=['GET'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def hosts_count_for_package():
    """Returns numer of hosts having the supplied package uuid

    Args:
        package_uuid (str)

    Returns:
        number of hosts having the supplied package_uuid installed

    """
    try:
        start_time = time.time()
        package_uuid = request.args.get('package_uuid', '')
        result = HostPackagesStatus.select(HostPackagesStatus.host).where(HostPackagesStatus.package_uuid == package_uuid).distinct().count()

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('host_data failed %s' % (repr(e)))
        return make_response_from_exception(e)

    return make_response(result=result, msg='Found {} Hosts having the package with uuid {}'.format(result, package_uuid), success=True,
                         error_code=None, status=200, request_time=time.time() - start_time)

@app.route('/api/v3/hosts_for_package', methods=['GET', 'POST'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def hosts_for_package():
    """Returns list of hosts having the supplied package names

    Args:
        package (str) : (can be a csv list of package names)
        limit (int)

    Returns:
        list of Hosts

    """
    if request.method == 'GET':
        limit = int(request.args.get('limit', '1000'))
        packages = ensure_list(request.args.get('package'))
    else:
        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        post_data = ujson.loads(raw_data)
        limit = post_data.get('limit', '1000')
        packages = post_data.get('package')

    def and_query(q, not_filter=False):
        if query is None:
            if not_filter:
                return ~(q)
            else:
                return q
        else:
            if not_filter:
                return query & ~(q)
            else:
                return query & q

    user_acls = get_user_acls(session['user'])
    perimeters = has_any_right(user_acls,['admin','view'])

    query = ((Hosts.server_uuid == get_server_uuid()) | Hosts.server_uuid.is_null())

    if perimeters:
        if isinstance(perimeters,list):
            certs_sub = None
            for cert_fingerprint in perimeters:
                if certs_sub is None:
                    certs_sub = Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
                else:
                    certs_sub = certs_sub | Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
            query = and_query(certs_sub)

        query = and_query(HostPackagesStatus.package.in_(packages))

        result = []
        for h in (HostPackagesStatus.select(
                HostPackagesStatus,
                Hosts.computer_name.alias('host_computer_name'),
                Hosts.computer_fqdn.alias('host_computer_fqdn'),
                Hosts.description.alias('host_description'),
                Hosts.connected_ips.alias('host_connected_ips'),
                Hosts.os_name.alias('host_os_name'),
                Hosts.connected_users.alias('host_connected_users')
                )
                .where(query)
                .join(Hosts, 'RIGHT OUTER JOIN')
                .limit(limit)
                .dicts()):
            append_ws_status(h,'host')
            result.append(h)
    else:
        result = []

    return make_response(msg=_('Hosts for packages %s, limit %s') % (packages, limit), result=result)

@app.route('/api/v3/hosts_for_wua', methods=['GET', 'POST'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def hosts_for_wua():
    """Returns list of hosts having the supplied WUA UUID

    Args:
        WUA_UUID (str) : (can be a csv list of package names)
        limit (int)

    Returns:
        list of Hosts

    """
    if request.method == 'GET':
        limit = int(request.args.get('limit', '1000'))
        wua_uuids = request.args.get('wua_uuids')
        need_distinct = request.args.get('merge', 'True') == 'True'
        wua_uuids = wua_uuids.split(',')
    else:
        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        post_data = ujson.loads(raw_data)
        limit = post_data.get('limit', '1000')
        wua_uuids = post_data.get('wua_uuids')
        need_distinct = post_data.get('merge', 'True') == 'True'

    where_clause = HostWsus.update_id.in_(wua_uuids) &  ((Hosts.server_uuid == get_server_uuid()) | Hosts.server_uuid.is_null())

    def and_query(q, not_filter=False):
        if where_clause is None:
            if not_filter:
                return ~(q)
            else:
                return q
        else:
            if not_filter:
                return where_clause & ~(q)
            else:
                return where_clause & q

    user_acls = get_user_acls(session['user'])
    perimeters = has_any_right(user_acls,['admin','view'])

    if perimeters:
        if isinstance(perimeters,list):
            certs_sub = None
            # add a subquery to check if host's authorized_certificates_sha256 list contains one of the allowed perimeters of the user.
            for cert_fingerprint in perimeters:
                if certs_sub is None:
                    certs_sub = Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
                else:
                    certs_sub = certs_sub | Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
            where_clause = and_query(certs_sub)

        if need_distinct:
            result = list(
                HostWsus
                    .select(HostWsus, Hosts)
                    .where(where_clause)
                    .join(Hosts, JOIN.INNER, on=(Hosts.uuid == HostWsus.host_id))
                    .distinct(Hosts.uuid)
                    .limit(limit)
                    .dicts()
            )
        else:
            result = list(
                HostWsus
                    .select(HostWsus, Hosts)
                    .where(where_clause)
                    .join(Hosts, JOIN.INNER, on=(Hosts.uuid == HostWsus.host_id))
                    .limit(limit)
                    .dicts()
            )
    else:
        # no rights
        result = []

    return make_response(msg=_('Hosts for wua %s, limit %s') % (str(wua_uuids), limit), result=result)

@app.route('/api/v3/hosts_for_normalization', methods=['GET', 'POST'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def hosts_for_normalization():
    """Returns list of hosts having the supplied WUA UUID

    Args:
        normalizations_original_names (str)
        limit (int)

    Returns:
        list of Hosts

    """
    start_time = time.time()
    if request.method == 'GET':
        limit = int(request.args.get('limit', '1000'))
        need_distinct = request.args.get('merge', '0')  != '0'
        normalizations_original_names = ensure_list(request.args.get('normalizations_original_names'))
    else:
        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        post_data = ujson.loads(raw_data)
        limit = post_data.get('limit', '1000')
        need_distinct = post_data.get('merge', 'False') == 'True'
        normalizations_original_names = post_data.get('normalizations_original_names')

    where_clause = Normalization.original_name.in_(normalizations_original_names)  &  ((Hosts.server_uuid == get_server_uuid()) | Hosts.server_uuid.is_null())

    def and_query(q, not_filter=False):
        if where_clause is None:
            if not_filter:
                return ~(q)
            else:
                return q
        else:
            if not_filter:
                return where_clause & ~(q)
            else:
                return where_clause & q

    user_acls = get_user_acls(session['user'])
    perimeters = has_any_right(user_acls,['admin','view'])

    if perimeters:
        # specific rights
        if isinstance(perimeters,list):
            certs_sub = None
            for cert_fingerprint in perimeters:
                if certs_sub is None:
                    certs_sub = Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
                else:
                    certs_sub = certs_sub | Hosts.authorized_certificates_sha256.contains(cert_fingerprint)
            where_clause = and_query(certs_sub)


        if need_distinct:
            result = []
            obj = (Normalization
                    .select(Normalization, Hosts.host_capabilities, Hosts.computer_fqdn, Hosts.uuid, Hosts.host_status, Hosts.audit_status, Hosts.description, Hosts.platform, Hosts.os_name, Hosts.registration_auth_user, Hosts.repositories,  HostSoftwares.name, HostSoftwares.version,
SQL("case when session_id is null then 'DISCONNECTED' else 'OK' end as reachable"))
                    .where(where_clause)
                    .join(HostSoftwares,'RIGHT JOIN',on=(HostSoftwares.key == Normalization.key) & (HostSoftwares.name == Normalization.original_name) )
                    .join(Hosts, JOIN.INNER, on=(Hosts.uuid == HostSoftwares.host_id))
                    .distinct(Hosts.uuid)
                    .limit(limit)
                    .dicts())
            for h in obj:
                append_ws_status(h)
                result.append(h)

        else:
            result = list(
                Normalization
                    .select(Normalization, Hosts.host_capabilities, Hosts.computer_fqdn, Hosts.uuid, Hosts.host_status, Hosts.audit_status, Hosts.description, Hosts.platform, Hosts.os_name, Hosts.registration_auth_user, Hosts.repositories,  HostSoftwares.name, HostSoftwares.version)
                    .where(where_clause)
                    .join(HostSoftwares,'RIGHT JOIN',on=(HostSoftwares.key == Normalization.key) & (HostSoftwares.name == Normalization.original_name) )
                    .join(Hosts, JOIN.INNER, on=(Hosts.uuid == HostSoftwares.host_id))
                    .limit(limit)
                    .dicts()
            )
    else:
        # no rights                                      h)
        result = []
    request_time = time.time() - start_time
    return make_response(msg=_('Hosts for normalization %s, limit %s') % (str(normalizations_original_names), limit), result=result, request_time=request_time)

@app.route('/api/v3/delete_unassigned_softwares', methods=['GET'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def delete_unassigned_softwares():
    """Delete unassigned softwares"""
    to_delete = str(request.args.get('to_delete'))

    with WaptDB():

        softwares_to_delete = Normalization.select(Normalization.key, Normalization.original_name).join(
        HostSoftwares, 'LEFT JOIN',on=(HostSoftwares.key == Normalization.key) & (HostSoftwares.name == Normalization.original_name)).where(
        HostSoftwares.name.is_null() & HostSoftwares.key.is_null()).tuples()

        nbr_softwares = 0
        list_key_softwares = []
        for row in softwares_to_delete:
            nbr_softwares += 1
            if to_delete == 'True':
                list_key_softwares.append(row[0])

        if to_delete == 'True':
            Normalization.delete().where(Normalization.key.in_(list_key_softwares)).execute()

    return make_response(result=nbr_softwares)


@app.route('/api/v3/packages_for_hosts', methods=['GET', 'POST'])
@requires_auth(['admin','view'])
@require_wapt_db
def packages_for_hosts():
    """Returns list of aggregated packages status

    Args:
        host_uuids (str) : (can be a csv list of hosts uuid)
        install_status (str): csv list of install_status to return
        sections (str): csv list of packages section to return. If empty, all sections
        reachable : if set and =1, returns only reachable hosts status

        limit (int)

    Returns:
        list of Hosts

    """
    if request.method == 'GET':
        limit = int(request.args.get('limit', '1000'))
        host_uuids = ensure_list(request.args.get('host_uuids'))
        reachable = request.args.get('reachable', '') == '1'
        install_status = ensure_list(request.args.get('install_status'))
        sections = ensure_list(request.args.get('sections'))
    else:
        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        post_data = ujson.loads(raw_data)
        limit = post_data.get('limit', '1000')
        host_uuids = post_data.get('host_uuids')
        reachable = str(post_data.get('reachable', '')) == '1'
        install_status = ensure_list(post_data.get('install_status'))
        sections = ensure_list(post_data.get('sections'))

    fields = []
    fields.extend([
        HostPackagesStatus.package,
        HostPackagesStatus.version,
        HostPackagesStatus.section,
        HostPackagesStatus.priority,
        HostPackagesStatus.package_uuid,
        HostPackagesStatus.install_status,
    ])

    fields.append(SQL('json_agg(uuid) as host_uuids'))
    fields.append(fn.COUNT(Hosts.uuid))
    query = HostPackagesStatus.select(*fields)
    where = HostPackagesStatus.host.in_(host_uuids)
    if install_status:
        where = where & HostPackagesStatus.install_status.in_(install_status)
    if sections:
        where = where & HostPackagesStatus.section.in_(sections)

    if reachable:
        where = where & HostPackagesStatus.host.in_(connected_uuids(host_uuids))

    result = []

    # TODO : merge 2 lists of hosts, one not connected, the other connected
    # as we can't do that by query.
    for h in (query.where(where)
            .join(Hosts, 'RIGHT OUTER JOIN')
            .group_by(SQL(','.join([str(i) for i in range(1, len(fields)-1)])))  # 1,2,3,4
            .limit(limit)
            .dicts()):
        result.append(h)

    return make_response(msg=_('Packages status for %s hosts, limit %s') % (len(host_uuids), limit), result=result)


@app.route('/api/v3/trusted_signers_certificates', methods=['GET', 'DELETE', 'POST'])
@requires_auth(['admin'])
def trusted_signers_certificates():
    if request.method == 'GET':
        try:
            start_time = time.time()
            # load trusted signers
            trusted = read_trusted_certificates(app.conf.get('trusted_signers_certificates_folder'))
            # add other known certificates
            known_ssl_path = app.conf.get('known_certificates_folder')
            if os.path.isdir(known_ssl_path):
                trusted.add_pems(known_ssl_path)
            trusted_certs = [{'cn': c.cn,
                              'issuer_cn': c.issuer_cn,
                              'subject_key_identifier': codecs.encode(c.subject_key_identifier or b'','hex'),
                              'authority_key_identifier': codecs.encode(c.authority_key_identifier or b'','hex'),
                              'fingerprint_sha1': codecs.encode(c.get_fingerprint('sha1'),'hex'),
                              'fingerprint': codecs.encode(c.get_fingerprint('sha256'),'hex'),
                              'pem': c.as_pem()} for c in list(trusted.trusted.values())]
            known_certificates = [{'cn': c.cn,
                                   'issuer_cn': c.issuer_cn,
                                   'subject_key_identifier': codecs.encode(c.subject_key_identifier or b'','hex'),
                                   'authority_key_identifier': codecs.encode(c.authority_key_identifier or b'','hex'),
                                   'fingerprint_sha1': codecs.encode(c.get_fingerprint('sha1'),'hex'),
                                   'fingerprint': codecs.encode(c.get_fingerprint('sha256'),'hex'),
                                   'pem': c.as_pem()} for c in trusted.certificates()]

            msg = 'Trusted: %s, Known: %s' % (','.join([c.cn for c in list(trusted.trusted.values())]), ','.join([c.cn for c in trusted.certificates()]))
            result = {'trusted': trusted_certs, 'certificates': known_certificates}
            return make_response(result=result, msg=msg, status=200, request_time=time.time() - start_time)

        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.critical('trusted_signers_certificates %s' % (repr(e)))
            return make_response_from_exception(e)
    elif request.method == 'POST':
        # one new trusted certificate in payload
        try:
            start_time = time.time()

            if request.headers.get('Content-Encoding') == 'gzip':
                raw_data = zlib.decompress(request.data)
            else:
                raw_data = request.data
            # try to read it to check consistency
            trusted = SSLCABundle()
            new_certs = trusted.add_certificates_from_pem(raw_data)
            cert_fn = os.path.join(app.conf.get('trusted_signers_certificates_folder'), codecs.encode(new_certs[0].get_fingerprint('sha1'),'hex')+'.crt')
            with open(cert_fn, 'w') as f:
                f.write(raw_data)
            msg = 'Added %s as trusted certificate (cn %s)' % (cert_fn, new_certs[0].cn)
            return make_response(result=trusted, msg=msg, status=200, request_time=time.time() - start_time)

        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.critical('trusted_signers_certificates %s' % (repr(e)))
            return make_response_from_exception(e)
    elif request.method == 'DELETE':
        try:
            start_time = time.time()
            # payload is a list of sha1 fingerprints of trusted certificates to remove from  trusted_signers_certificates_folder
            fingerprints = ujson.loads(request.args.get('fingerprints','[]'))
            if not isinstance(fingerprints, list):
                raise Exception('Bad argument')

            result = []
            for cert_fn in glob.glob(os.path.join(app.conf.get('trusted_signers_certificates_folder'), '*.crt')):
                cert = SSLCertificate(cert_fn)
                if codecs.encode(cert.get_fingerprint('sha1'),'hex') in fingerprints:
                    os.unlink(cert_fn)
                    result.append(cert_fn)

            msg = 'Removed trusted certificates %s' % (', '.join(result),)
            return make_response(result=result, msg=msg, status=200, request_time=time.time() - start_time)

        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.critical('trusted_signers_certificates %s' % (repr(e)))
            return make_response_from_exception(e)


@app.route('/api/v3/known_signers_certificates')
@requires_auth(['admin','view'])
@require_wapt_db
def known_signers_certificates():
    try:
        start_time = time.time()
        # load signers certificates from Packages index (extracted from uploaded packages)
        local_repo = get_repo_packages()
        signers_ca = local_repo.get_certificates()

        # append certificates trusted by all client computers (from each wapt/ssl directory)
        known_ssl_path = app.conf.get('known_certificates_folder')
        if os.path.isdir(known_ssl_path):
            signers_ca.add_pems(known_ssl_path)

        # append certificates from Users table
        for user_acls in WaptUserAcls.select(WaptUsers.user_certificate).join(WaptUsers,'LEFT JOIN',on=(WaptUsers.name == WaptUserAcls.username)).where(~WaptUsers.user_certificate.is_null()).dicts():
            pem_data = user_acls['user_certificate']
            signers_ca.add_certificates_from_pem(pem_data)

        msg = 'Known signers certificates from packages and users'
        result = {'certificates': signers_ca.certificates(), 'crls': signers_ca.crls}
        return make_response(result=result, msg=msg, status=200, request_time=time.time() - start_time)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('known_signers_certificates %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/usage_statistics')
@app.route('/api/v1/usage_statistics')
@requires_auth(['admin', 'view'])
@require_wapt_db
def usage_statistics():
    """returns some anonymous usage statistics to give an idea of depth of use"""
    try:
        start_time = time.time()
        host_data = Hosts.select(
            fn.COUNT(Hosts.uuid).alias('hosts_count'),
            fn.MIN(Hosts.last_seen_on).alias('oldest_query'),
            fn.MAX(Hosts.last_seen_on).alias('newest_query'),
        ).where(Hosts.server_uuid == get_server_uuid()).dicts().first()

        stats = {
            'hosts_count': host_data['hosts_count'],
            'oldest_query': host_data['oldest_query'],
            'newest_query': host_data['newest_query'],
            'packages_count_max': None,
            'packages_count_avg': None,
            'packages_count_ok': None,
            'hosts_count_has_error': None,
            'hosts_count_need_upgrade': None,
        }

    except:
        pass

    result = dict(
        uuid=app.conf['server_uuid'],
        platform=platform.system(),
        architecture=platform.architecture(),
        version=__version__,
        date=datetime2isodate(),
        edition=get_wapt_edition(),
        git_hash=git_hash,
    )
    result.update(stats)
    return make_response(msg=_('Anomnymous usage statistics'), result=result,request_time=time.time() - start_time)


@app.route('/api/v3/host_tasks_status', methods=['GET','POST'])
@requires_auth(['admin', 'view'])
@require_wapt_db
def host_tasks_status():
    """Proxy the get tasks status action to the client"""
    try:
        if not socketio:
            raise Exception('socketio unavailable')

        if request.method == 'POST':
            if request.headers.get('Content-Encoding') == 'gzip':
                raw_data = zlib.decompress(request.data)
            else:
                raw_data = request.data

            args = ujson.loads(raw_data)
        else:
            args = request.args

        uuid = args['uuid']

        client_tasks_timeout = float(args.get('client_tasks_timeout', app.conf['client_tasks_timeout']))
        start_time = time.time()
        ws = app.ws_connections.get(uuid)
        if ws and ws[WS_SID]:
            result = []

            def result_callback(data):
                result.append(data)

            request.sid = ws[WS_SID]
            socketio.emit('get_tasks_status', args, room=request.sid, callback=result_callback)

            start_waiting = time.time()
            while not result:
                if time.time() - start_waiting > client_tasks_timeout:
                    raise EWaptTimeoutWaitingForResult('Timeout, client did not send result within %s s' % client_tasks_timeout)
                socketio.sleep(0.1)
            # be sure to not eat cpu in case host return empty result
            if not 'running' in result:
                time.sleep(0.1)
            msg = 'Tasks status for %s' % uuid
            if result:
                if result[0]['success']:
                    return make_response(result[0]['result'],
                                         msg=msg,
                                         success=True,
                                         request_time=time.time() - start_time,
                                         )
                else:
                    return make_response(None,
                                         msg=result[0]['msg'],
                                         success=False,
                                         request_time=time.time() - start_time,
                                         )
            else:
                return make_response(None,
                                     msg='No answer',
                                     success=False,
                                     request_time=time.time() - start_time,
                                     )

        else:
            raise EWaptHostUnreachable('Host not connected, Websocket sid not in database')

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('host_tasks_status failed %s' % (repr(e)))
        return make_response_from_exception(e)


@app.route('/api/v3/trigger_host_action', methods=['HEAD', 'POST'])
@requires_auth(['admin', 'trigger_host_action', 'trigger_host_upgrade'])
@require_wapt_db
def trigger_host_action():
    """Proxy some single shot actions to the client using websockets"""
    try:
        if not socketio:
            raise Exception('socketio unavailable')

        action_data = request.get_json()
        if action_data:
            if isinstance(action_data, list):
                for action in action_data:
                    if not action.get('signature', None):
                        raise EWaptBadSignature('One or some actions not signed, not relayed')
            else:
                if not action_data.get('signature', None):
                    raise EWaptBadSignature('Action is not signed, not relayed')
        else:
            raise EWaptForbiddden('Invalid action')
        # TODO : check signer before proxying ?
        host = None

        if not isinstance(action_data, list):
            action_data = [action_data]

        ok = []
        client_errors = []
        server_errors = []
        other_server = []
        expected_result_count = 0

        def result_callback(data):
            if data.get('success', False):
                ok.append(data)
            else:
                client_errors.append(data)

        for action in action_data:
            if 'uuid' in action:
                where = Hosts.uuid == action['uuid']
            elif 'uuids' in action:
                where = Hosts.uuid.in_(action['uuids'])
            elif 'organizational_units' in action:
                ou_list = action['organizational_units']
                if action.get('include_childs_ou', '1') == '1':
                    or_list = Hosts.computer_ad_ou.endswith(ou_list[0])
                    for ou in ou_list[1:]:
                        or_list = or_list | Hosts.computer_ad_ou.endswith(ou)
                    where = (or_list)
                else:
                    where = (Hosts.computer_ad_ou.in_(ou_list))

            for host in Hosts.select(Hosts.uuid, Hosts.server_uuid).where(where).dicts():
                uuid = host['uuid']
                if host['server_uuid'] != app.conf['server_uuid']:
                    other_server.append(uuid)
                else:
                    # check if connected
                    ws = app.ws_connections.get(uuid)
                    if ws and ws[WS_SID]:
                        notify_server = action.get('notify_server', False)
                        #if notify_server:
                        #    socket_callback = result_callback
                        #else:
                        #    socket_callback = None
                        try:
                            request.sid = ws[WS_SID]
                            socketio.emit('trigger_host_action', action, room=request.sid)  # , callback = socket_callback)
                            if notify_server:
                                expected_result_count += 1
                            # notify console that action is in progress until client send it updated status.
                            # if notify_server and not uuid in notified_uuids:
                            #    notified_uuids.append(uuid)
                            #    Hosts.update(host_status='RUNNING').where(Hosts.uuid == uuid).execute()
                        except Exception as e:
                            server_errors.append('Error for %s: %s' % (uuid, e))
        msg_parts = []
        if expected_result_count:
            msg_parts.append('Actions launched: %s ' % expected_result_count)
        if client_errors:
            msg_parts.append('Client errors: %s ' % len(client_errors))
        if other_server:
            msg_parts.append('Skipped: %s' % len(other_server))
        if server_errors:
            msg_parts.append('Server errors: %s' %  len(server_errors))
        msg = ', '.join(msg_parts)
        return make_response([r.get('result', None) for r in (ok + client_errors)],
                             msg=msg,
                             success=len(client_errors) == 0 and len(server_errors) == 0)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('trigger_host_action failed %s' % (repr(e)))
        return make_response_from_exception(e)


def setup_logging(config=None):
    loglevel = config['loglevel']

    hdlr = logging.StreamHandler()
    hdlr.setFormatter(
        logging.Formatter('%(asctime)s [%(name)-15s] %(levelname)s %(message)s'))
    rootlogger = logging.getLogger()
    rootlogger.addHandler(hdlr)
    setloglevel(rootlogger, loglevel)

    for log in WAPTLOGGERS:
        sublogger = logging.getLogger(log)
        if sublogger:
            if 'loglevel_%s' % log in config:
                setloglevel(sublogger, config['loglevel_%s' % log])
            else:
                setloglevel(sublogger, loglevel)

            # if platform.system() == 'Linux':
            #    hdlr = logging.handlers.SysLogHandler('/dev/log')
            # else:
                #log_directory = os.path.join(wapt_root_dir, 'log')
                # if not os.path.exists(log_directory):
                #    os.mkdir(log_directory)
            #    hdlr = logging.FileHandler(os.path.join(log_directory, 'waptserver.log'))

            #hdlr = logging.StreamHandler()
            # hdlr.setFormatter(
            #    logging.Formatter('%(asctime)s [%(name)-15s] %(levelname)s %(message)s'))
            # sublogger.addHandler(hdlr)



@app.route('/api/v3/user_infos')
# @requires_auth(['admin','view'],methods=['session','admin','passwd','ldap','ssl','token','kerb'])
def user_infos():
    user_info = dict(
        user=session.get('user'),
        user_fingerprint=session.get('user_fingerprint'),
        user_fingerprint_sha1=session.get('user_fingerprint_sha1'),
        user_acls=session.get('user_acls'),
        auth_method=session.get('auth_method'),
        auth_date=session.get('auth_date'),
    )
    msg = 'Info on logged in user'
    return make_response(result=user_info, msg=msg)


@app.route('/api/v3/set_user_password',methods=['POST'])
@require_wapt_db
@requires_auth(['admin','view'],methods=['session','admin','passwd'])
def set_user_password():
    try:
        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        post_data = ujson.loads(raw_data)
        if not post_data:
            raise Exception('post_users_acls: No data supplied')

        new_username = post_data.get('username')
        new_password = post_data.get('password')
        old_password = post_data.get('old_password')
        if new_username == app.conf['wapt_user']:
            raise EWaptForbiddden('Changing wapt server admin password is not available here')

        if not new_username or not new_password:
            raise EWaptMissingParameter('Missing argument')
        if not old_password:
            old_password = None
        if not user_is_admin() and old_password is None:
            raise EWaptMissingParameter('Not admin, missing old password')
        user_rec = WaptUserAcls.get_or_none(WaptUserAcls.username==new_username)
        if not user_rec:
            raise EWaptForbiddden('Unknown user')
        existing = set_user_password_htpasswd(new_username,new_password,old_password)
        return make_response(dict(user=new_username,new_user=not existing),msg='Password changed successfully')
    except Exception as e:
        return make_response_from_exception(e)

@app.route('/api/v3/invalidate_user_password',methods=['POST'])
@require_wapt_db
@requires_auth(['admin','view'],methods=['session','admin','passwd'])
def invalidate_user_password():
    try:
        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        post_data = ujson.loads(raw_data)
        if not post_data:
            raise Exception('post_users_acls: No data supplied')

        username = post_data.get('username')
        password = post_data.get('password')
        if username == app.conf['wapt_user']:
            raise EWaptForbiddden('Changing wapt server admin password is not available here')

        if not username:
            raise EWaptMissingParameter('Missing argument')
        if not password:
            password = None
        if not user_is_admin() and password is None:
            raise EWaptMissingParameter('Not admin, missing password')
        existing = delete_user_password_htpasswd(username,password)
        return make_response(dict(user=username,existing=existing),msg='Password removed successfully')
    except Exception as e:
        return make_response_from_exception(e)



@app.route('/api/v3/get_auth_logs')
@requires_auth(['admin'])
@require_wapt_db
def get_auth_logs():
    """
    """
    start_time = time.time()
    try:
        pattern = request.args.get('pattern', '')
        forbidden = int(request.args.get('forbidden', -1))
        authenticated = int(request.args.get('authenticated', -1))
        columns = request.args.get('columns')
        if columns:
            query = LogsAPI.select(* [getattr(LogsAPI, c) for c in columns.split(',')] )
        else:
            query = LogsAPI.select()

        if pattern:
            where_clause = (LogsAPI.id.cast('text').contains(pattern) | LogsAPI.endpoint.contains(pattern) | LogsAPI.args.cast('text').contains(pattern) | LogsAPI.user.contains(pattern) \
            | LogsAPI.user_acls.contains(pattern) | LogsAPI.host.contains(pattern) | LogsAPI.remote_addr.contains(pattern) | LogsAPI.header_user_agent.contains(pattern) \
            | LogsAPI.method.contains(pattern) | LogsAPI.scheme.contains(pattern) | LogsAPI.full_path.contains(pattern) | LogsAPI.data.contains(pattern))
        else:
            where_clause = True
        if forbidden!=-1:
            where_clause = where_clause & (LogsAPI.forbidden == bool(forbidden))
        if authenticated!=-1:
            where_clause = where_clause & (LogsAPI.authenticated == bool(authenticated))
        query = query.where(where_clause)
        if request.args.get('order'):
            if request.args.get('descending','')=='1':
                query = query.order_by(request.args.get('order')).desc()
            else:
                query = query.order_by(request.args.get('order'))
        else:
            query = query.order_by(LogsAPI.id.desc())

        limit = int(request.args.get('limit', '200'))
        query = query.limit(limit)
        values = list(query.dicts())

        rows_count =  LogsAPI.select(fn.COUNT(LogsAPI.id)).where(where_clause).tuples()[0][0]

        result = {"values": values, "rows_count": rows_count}
        return make_response(result=result, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@app.route('/api/v3/clear_auth_logs')
@requires_auth(['admin'])
@require_wapt_db
def clear_auth_logs():
    """
    """
    start_time = time.time()
    try:
        return make_response(result={'rows_deleted': LogsAPI.delete().execute()}, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@app.route('/api/v3/licences', methods=['GET', 'DELETE', 'POST'])
@requires_auth(['admin'])
def licences():
    global licences_list
    start_time = time.time()
    if request.method == 'GET':
        try:
            msg = 'Licences'
            return make_response(result=licences_list.get_licences(), msg=msg, status=200, request_time=time.time() - start_time)

        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.critical('get licences %s' % (repr(e)))
            return make_response_from_exception(e)

    elif request.method == 'POST':
        # one new licence in payload
        try:
            if request.headers.get('Content-Encoding') == 'gzip':
                raw_data = zlib.decompress(request.data)
            else:
                raw_data = request.data

            (result,status,message) = licences_list.add_licence(raw_data)

            return make_response(
                result = result,
                success = (status == LicenceStatus.lsValid),
                msg = message,
                status=200,
                request_time=time.time() - start_time)

        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.critical('post licences %s' % (repr(e)))
            return make_response_from_exception(e)

    elif request.method == 'DELETE':
        try:
            # payload is a list of licence_nr to delete
            to_delete = ujson.loads(request.args.get('licences_nr','[]'))
            if not isinstance(to_delete, list):
                raise Exception('Bad argument')

            result = licences_list.remove_licences(to_delete)
            msg = 'Removed licences: %s' % (','.join([licence['licence_nr'] for licence in result]))
            return make_response(result=result, msg=msg, status=200, request_time=time.time() - start_time)

        except Exception as e:
            logger.debug(traceback.format_exc())
            logger.critical('delete licences %s' % (repr(e)))
            return make_response_from_exception(e)



@app.route('/api/v3/connected_wol_relays', methods=['GET'])
@requires_auth(['admin', 'trigger_host_action'])
@require_wapt_db
def get_connected_wol_relays():
    with WaptDB():
        wol_relays = [h for h in
                    Hosts.select(Hosts.uuid, Hosts.server_uuid, Hosts.host_info)
                        .where((SQL("(wapt_status->>'wol_relay')") == 'true'))
                        .dicts()
                        if host_is_reachable(h['uuid'])
                        ]
        return make_response(msg='%s connected wake-on-lan relays' % len(wol_relays),result=wol_relays)


def stuff_waptsetup():
    waptsetup_fn = os.path.join(app.conf['wapt_folder'], 'waptagent', 'waptsetup.exe')
    stuffed_waptsetup_fn = os.path.join(app.conf['wapt_folder'], 'waptsetup.exe.tmp')
    if os.path.isfile(stuffed_waptsetup_fn):
        os.remove(stuffed_waptsetup_fn)

    configs = get_all_json_configs(os.path.join(app.conf['wapt_folder'],'conf.d'))
    try:
        if configs and os.path.exists(waptsetup_fn):
            if waptlicences.stuff_exe_certificate(waptsetup_fn,stuffed_waptsetup_fn,json.dumps(configs)):
                try:
                    (_present,waptsetup_version) = get_wapt_exe_version(stuffed_waptsetup_fn)
                    if not waptsetup_version:
                        raise Exception('No version found in waptsetup executable.')
                    config_names = {c['name']: c['hash'] for c in configs}

                    for filename in glob.glob(os.path.join(app.conf['wapt_folder'],'waptsetup-*_*.exe')):
                        os.unlink(filename)
                    for config_name,hash in config_names.items():
                        waptsetup_config_fn = os.path.join(app.conf['wapt_folder'], 'waptsetup-%s_%s.exe' % ( secure_filename(config_name),hash))
                        if os.path.exists(waptsetup_config_fn):
                            os.unlink(waptsetup_config_fn)
                        try:
                            os.link(stuffed_waptsetup_fn, waptsetup_config_fn)
                        except:
                            shutil.copy2(stuffed_waptsetup_fn, waptsetup_config_fn)

                finally:
                    os.unlink(stuffed_waptsetup_fn)

    except Exception as e:
        logger.critical('Unable to stuff waptsetup.exe with configurations data: %s' % e)

@app.route('/api/v3/add_configurations', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def add_configurations():
    try:
        set_waptagent=request.args.get('set_waptagent')

        config_folder = os.path.join(os.path.dirname(app.conf['wapt_folder']), 'wapt', 'conf.d')
        configs = json.loads(request.data.decode('utf-8'))
        if set_waptagent and len(configs) != 1:
            raise Exception('If set_waptagent is given, only one agent configuration is supported.')

        result = []

        for conf in configs:
            config_name = conf['name']
            config_hash = conf.get('hash', None)
            config_filename = secure_filename(config_name+ (('_'+config_hash) if config_hash else '')+'.json')
            config_abspath = os.path.join(config_folder, config_filename)
            ## Remove the old conf
            for old_conf in glob.glob(os.path.join(config_folder, config_name + '_%s.json' % ('?'*len(config_hash),) )):
                os.remove(old_conf)
            ## write the new conf
            with open(config_abspath, 'w') as conf_file:
                conf_file.write(conf['content'])
            result.append(config_filename)

        if app.conf['auto_create_waptagent_from_config']:
            stuff_waptsetup()

        # create a legacy waptagent.exe link to the default stuffed conf
        waptsetup_fn = os.path.join(app.conf['wapt_folder'], 'waptagent', 'waptsetup.exe')
        if set_waptagent and len(configs)==1 and os.path.exists(waptsetup_fn):
            stuffed_waptsetup_fn = os.path.join(app.conf['wapt_folder'], 'waptagent.exe.tmp')
            if os.path.isfile(stuffed_waptsetup_fn):
                os.remove(stuffed_waptsetup_fn)
            if waptlicences.stuff_exe_certificate(waptsetup_fn,stuffed_waptsetup_fn,request.data):
                waptagent_default_fn = os.path.join(app.conf['wapt_folder'],'waptagent.exe')
                if os.path.exists(waptagent_default_fn):
                    os.remove(waptagent_default_fn)
                os.rename(stuffed_waptsetup_fn,waptagent_default_fn)

        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))
        return make_response(result=result)
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('add configuration %s' % (repr(e)))
        return make_response_from_exception(e)

@app.route('/api/v3/delete_configuration', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def delete_configuration():
    config_name = request.args.get('name', None)
    config_hash = request.args.get('hash', None)
    if not config_name:
        return make_response(msg='No configuration found to delete')
    config_folder = os.path.join(os.path.dirname(app.conf['wapt_folder']), 'wapt', 'conf.d')
    config_abspath = os.path.join(config_folder, config_name+ (('_'+config_hash) if config_hash else '*')+'.json')
    try:
        ## Remove the old conf
        for old_conf in glob.glob(config_abspath):
            os.remove(old_conf)
        stuff_waptsetup()
        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))
        return make_response()
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('add configuration %s' % (repr(e)))
        return make_response_from_exception(e)

@app.route('/api/v3/list_configurations', methods=['GET'])
@requires_auth(['admin','register_host'])
def list_configurations():
    return make_response(result=get_all_json_configs(os.path.join(app.conf['wapt_folder'],'conf.d')))


@app.route('/api/v3/delete_audit_datas', methods=['DELETE'])
@requires_auth(['admin'])
@require_wapt_db
def delete_audit_data():
    ids = ensure_list(request.args.get('ids'))

    with WaptDB():
        HostAuditData.delete().where(HostAuditData.id.in_(ids)).execute()

    return make_response()

@app.route('/api/v3/get_hash_json_content')
@requires_auth(['admin', 'view'])
def get_hash_json_content():
    hash_json_filename = os.path.join(wapt_root_dir,'waptbinaries.sha256')
    if not os.path.isfile(hash_json_filename):
        return make_response(success=False,msg='waptbinaries.sha256 not found')
    with open(hash_json_filename, 'r') as f:
        hashes_content = json.loads(f.read())
    return make_response(result=hashes_content, success=True)

@app.route('/api/v3/get_waptdeploy_exe/<ip>/waptdeploy.exe')
def get_waptdeploy_exe(ip=''):
    url = get_secondary_repo(ip,get_forwarded_url(request),repo_name='wapt')
    return redirect(url+'wapt/waptagent/waptdeploy.exe')

@app.route('/api/v3/get_waptagent_exe/<ip>/waptagent.exe')
def get_wapt_agent_exe(ip=''):
    url = get_secondary_repo(ip,get_forwarded_url(request),repo_name='wapt')
    return redirect(url+'wapt/waptagent.exe')

def get_corresponding_column(query_id, query_column):
    filter_query = ReportingQueries.select(ReportingQueries.query).where(ReportingQueries.id == query_id).dicts().first()
    if not filter_query:
        raise Exception("Query id %d not found" % (id))

    rows = ReportingQueries.raw(filter_query['query']).dicts()
    uuid_list = [row[query_column] for row in rows]

    return uuid_list

def get_hash256_exe(exe_filename):
    BUF_SIZE = 65536

    if os.path.isfile(exe_filename):
        hash_exe = hashlib.sha256()
        with open(exe_filename, 'rb') as exe:
            while True:
                data_exe = exe.read(BUF_SIZE)
                if not data_exe:
                    break
                hash_exe.update(data_exe)
        return hash_exe.hexdigest()
    else:
        return ''

@app.route('/api/v3/get_exe_hash', methods=['GET'])
@requires_auth(['admin'])
def get_exe_hash():
    ##TO DO : WAPT SETUP INSTEAD OF WAPTAGENT
    waptagent_exe = os.path.join(app.conf['wapt_folder'], 'waptagent.exe')
    waptdeploy_exe = os.path.join(app.conf['wapt_folder'], 'waptagent','waptdeploy.exe')
    wads_exe_64 = os.path.join(app.conf['wads_folder'], 'wads64.exe')
    wads_exe_32 = os.path.join(app.conf['wads_folder'], 'wads32.exe')

    return make_response(
      result={
        'waptagent.exe' : get_hash256_exe(waptagent_exe),
        'waptdeploy.exe' : get_hash256_exe(waptdeploy_exe),
        'wads64.exe' : get_hash256_exe(wads_exe_64),
        'wads32.exe' : get_hash256_exe(wads_exe_32)
        },
        success=True)

@app.route('/api/v3/import_templates', methods=['GET'])
@requires_auth(['admin'])
@require_wapt_db
def import_templates():
    start_time = time.time()
    wapt_db_connect()
    if not request.args.get('user'):
        return make_response(success=False, msg='No User provided')
    try:
        data = MessagesTemplate.select().where(MessagesTemplate.user == request.args.get('user'))
        data = list(data.dicts())
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)

@app.route('/api/v3/delete_templates', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def delete_templates():
    start_time = time.time()
    ##prepare_request_data(request)
    try:
        template_to_delete = ujson.loads(request.data)
        if not template_to_delete:
            raise Exception("HTML Template : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as db:
        try:
            cnt = MessagesTemplate.delete().where((MessagesTemplate.name == template_to_delete["name"]) & (MessagesTemplate.user == template_to_delete["user"])).execute()
            return make_response(cnt, success=True, request_time=time.time() - start_time)
        except Exception as e:
            db.rollback()
            return make_response_from_exception(e)

@app.route('/api/v3/save_templates', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def save_templates():
    start_time = time.time()
    try:
        html_messages = ujson.loads(request.data)
        if not html_messages:
            raise Exception("Html Templates : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)


    with wapt_db.atomic() as db:
        if "editing" in html_messages:
            MessagesTemplate.update(html_content=html_messages["html_content"]).where((MessagesTemplate.name == html_messages["name"]) & (MessagesTemplate.user == html_messages["user"])).execute()

            return make_response(success=True, request_time=time.time() - start_time)

        else:
            try:
                data = []
                MT = MessagesTemplate()
                if "name" in html_messages:
                    MT.name = html_messages["name"]
                if "user" in html_messages:
                    MT.user = html_messages["user"]
                if "html_content" in html_messages:
                    MT.html_content = html_messages["html_content"]

                MT.save()
                db.commit()
                return make_response(data, success=True, request_time=time.time() - start_time)
            except Exception as e:
                db.rollback()
                return make_response_from_exception(e)

def decode_agent_filename(full_filename, with_hash):
    file_hash = ''
    if with_hash:
        file_hash = sha256_for_file(full_filename)
    file_size = os.path.getsize(full_filename)
    local_path,filename = os.path.split(full_filename)


    result = {
        "filename": filename,
        "file_sha256" : file_hash,
        "file_size" : file_size,
    }

    if not filename.startswith('tis-waptagent') and not filename.startswith('waptsetup'):
        return result

    wapt_target_os = 'all'
    wapt_architecture = 'all'
    wapt_min_os_version = ''
    wapt_max_os_version = ''

    parts, ext = os.path.splitext(filename)
    parts = parts.split('-')

    if 'debian' in filename or 'ubuntu' in filename:
        # ex. tis-waptagent-2.2.1.11899-2d82654e-ubuntu-18-amd64.deb
        wapt_version = parts[2]
        wapt_githash = parts[3]

        os_type = filename.split('-')[4]
        os_version = filename.split('-')[5]
        os_arch = filename.split('-')[6].split(".")[0]
        wapt_target_os = os_type
        wapt_min_os_version = os_version
        wapt_max_os_version = os_version
        if os_arch == 'amd64':
            wapt_architecture = 'x64'
        elif os_arch == 'arm64':
            wapt_architecture = 'arm64'
        else:
            wapt_architecture = 'all'

    elif 'macos' in filename:
        # ex. tis-waptagent-2.2.2.12198-9d39e055-macos-all-intel
        wapt_version = parts[2]
        wapt_githash = parts[3]
        os_type = filename.split('-')[4]
        os_version = filename.split('-')[5]
        os_arch = filename.split('-')[6]
        wapt_target_os = 'macos'

    elif '.el7.' in filename:
        wapt_version = parts[2]
        wapt_githash = parts[3]
        os_type = 'redhat'
        os_version = '7'
        os_arch = filename.split('-')[3].split('.')[2]

        wapt_target_os = 'redhat_based'
        wapt_architecture = 'x64'
        wapt_min_os_version = '7'
        wapt_max_os_version = '7'

    elif '.el8.' in filename:
        wapt_version = parts[2]
        wapt_githash = parts[3]
        os_type = 'redhat'
        os_version = '8'
        os_arch = filename.split('-')[3].split('.')[2]

        wapt_target_os = 'redhat_based'
        wapt_architecture = 'x64'
        wapt_min_os_version = '8'
        wapt_max_os_version = '8'

    elif filename == 'waptsetup.exe':
        os_type = 'windows'
        os_version = '>6.1'
        os_arch = 'all'
        setup_present, setup_version = get_wapt_exe_version(full_filename)
        wapt_version = setup_version
        wapt_githash = ''

        wapt_target_os = 'windows'
        wapt_min_os_version = '6.1'

    else:
        return result

    result = {
        "filename": filename,
        "os_type" : os_type,
        "os_version" : os_version,
        "os_arch" : os_arch,
        "git_hash" : wapt_githash,
        "wapt_version" : wapt_version,
        "file_sha256" : file_hash,
        "file_size" : file_size,
        "wapt_target_os" : wapt_target_os,
        "wapt_architecture" : wapt_architecture,
        "wapt_min_os_version" : wapt_min_os_version,
        "wapt_max_os_version" : wapt_max_os_version
        }

    return result


@app.route('/api/v3/get_agents_info', methods=['GET'])
@requires_auth(['admin'])
def get_agents_info():
    with_hash = request.args.get('with_hash', 'False') == 'True'
    agents_folder =  app.conf['agents_folder']
    agents = []
    starttime = time.time()
    if os.path.isdir(agents_folder):
        for filename in glob.glob(os.path.join(agents_folder,'**')):
            if not os.path.isfile(filename):
                continue

            agents.append(decode_agent_filename(filename, with_hash))

        #wapt_agent_list_filename = os.path.join(app.conf['wapt_folder'], 'wapt_agent_list.json')
        #open(wapt_agent_list_filename,'w').write(jsondump(agents))
    else:
        os.makedirs(app.conf['agents_folder'])

    return make_response(result=agents, request_time=time.time() - starttime)

@app.route('/api/v3/upload_agents', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def upload_agents():
    try:
        starttime = time.time()

        agents_folder = os.path.join(app.conf['agents_folder'])

        if not os.path.isdir(agents_folder):
            os.makedirs(agents_folder)

        agent = request.args.get('agent')

        target = os.path.join(agents_folder,secure_filename(agent))
        if os.path.isfile(target):
            pass
        else:
            # streamed upload
            logger.info('Start streaming down file to %s' % target)
            filestream = PostDataStream(request.stream)
            filestream.save(target)

        if os.path.isfile(target):
            #tis-waptagent-2.4.0.13890-7d960b2a.el8.x86_64.rpm
            #tis-waptagent-2.4.0.13890-2d82654e-ubuntu-18-amd64.deb
            target_symlink, file_ext = os.path.splitext(target)
            if '-gui-' in os.path.basename(target):
                target_symlink = target_symlink.split('-', 4)
                str_len = len(target_symlink[0] + '-' + target_symlink[1] + '-' + target_symlink[2] + '-' + target_symlink[3])
                str_len = len(target) - str_len - 5 - len(file_ext) # 5 = hash part of file - separators char
                target_symlink = target_symlink[0] + '-' + target_symlink[1] + '-' + target_symlink[2] + target[-str_len:]
            else:
                target_symlink = target_symlink.split('-', 3)
                str_len = len(target_symlink[0] + target_symlink[1] + target_symlink[2])
                str_len = len(target) - str_len - 7 - len(file_ext) # 7 = hash part of file - separator char
                target_symlink = target_symlink[0] + '-' + target_symlink[1] + target[-str_len:]
            if os.path.basename(target_symlink) == secure_filename(os.path.basename(target_symlink)):
                if os.path.exists(target_symlink):
                    os.remove(target_symlink)
                logger.info('Making symlink %s to %s' % (target_symlink, target))
                os.symlink(target,target_symlink)
            else:
                logger.info('Symlink %s not made due to unsecure name' % (target_symlink))
            #tis-waptagent.el8.x86_64.rpm
            #tis-waptagent-ubuntu-18-amd64.deb

        spenttime = time.time() - starttime
        global licences_list
        if licences_list.is_enterprise and Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
            repositories.update_file_tree_of_files(username=session.get('user'))
        return make_response(success=os.path.isfile(target),
                             result=target,
                             msg=_('{} uploaded. Size {}').format(target,os.stat(target).st_size),
                             request_time=spenttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('upload agent (%s) failed %s' % (agent ,repr(e)))
        return make_response_from_exception(e)

@app.route('/api/v3/delete_agents', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def delete_agents():
    try:
        starttime = time.time()

        agents_folder = os.path.join(app.conf['agents_folder'])
        if not os.path.isdir(agents_folder):
            os.makedirs(agents_folder)

        agents = request.get_json()
        deleted = 0

        for agent in agents:
            target = os.path.join(agents_folder,secure_filename(agent))
            if os.path.isfile(target):
                logger.info("Deleting agent %s" % target)
                os.remove(target)
                deleted += 1
            else:
                logger.warning("Trying to delete agent %s not present on server" % target)

        spenttime = time.time() - starttime

        return make_response(msg='{} agents deleted.'.format(deleted), request_time=spenttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('delete agents (%s) failed %s' % (agents ,repr(e)))
        return make_response_from_exception(e)

@app.route('/api/v3/update_hosts_audit_data', methods=['HEAD', 'POST'])
@requires_auth(['admin','update_audit_data'])
@require_wapt_db
def update_hosts_audit_data():
    """Update table host_audit_data with the supplied json posted payload
    Can handle data for different host uuid.

    data is a json list of dicr with keys
        ['host','value_date','value_section','value_key','value','expiration_date']
        optional :
            'value_id' (int): remote id of the value (in host's local status table for example)
            'id' (int):

        'host' is the uuid of the host and *must* exist in Host table.
        'value' is a json string data.

    Returns:


    """
    if request.method == 'HEAD':
        return ''

    try:
        starttime = time.time()

        # unzip if post data is gzipped
        if request.headers.get('Content-Encoding') == 'gzip':
            raw_data = zlib.decompress(request.data)
        else:
            raw_data = request.data

        audit_data = ujson.loads(raw_data)
        if not audit_data:
            raise Exception('update_hosts_audit_data: No data supplied')

        if not isinstance(audit_data,list):
            raise Exception('Invalid audit_data')


        # check if audit_data is a list of dicts (easier client side) or a list of list (smaller/faster)
        if isinstance(audit_data[0],list):
            fields = audit_data[0]
            append_created_on = not 'created_on' in fields

            # removed not known columns if any
            dropped_columns_idx = []
            columns = []
            for i in range(len(fields)-1,-1,-1):
                if fields[i] == 'host':
                    fields[i] = 'host_id'

                if not fields[i] in HostAuditData._meta.columns:
                    dropped_columns_idx.append(i)
                else:
                    columns.insert(0,fields[i])

            if append_created_on:
                columns.append('created_on')

            missing_columns = list(set(['host_id','value_id','value_section','value_key','value','value_date','created_on']) - set(columns))
            if missing_columns:
                raise Exception('Missing mantaory columns: %s' % missing_columns)

            now = datetime.datetime.utcnow()
            if dropped_columns_idx or append_created_on:
                for row in audit_data[1:]:
                    if dropped_columns_idx:
                        for col_idx in dropped_columns_idx:  # already in reversed order
                            row.pop(col_idx)
                    if append_created_on:
                        row.append(now)

            result = []
            res_cursor = (HostAuditData.insert_many(audit_data[1:],columns)
                .on_conflict('update', conflict_target = 'id', preserve=['host_id','value_id','value_section','value_key','value','value_date','created_on'])
                .on_conflict('update', conflict_target = ['host_id','value_id'], preserve=['host_id','value_id','value_section','value_key','value','value_date','created_on'])
                ).execute()
            result = [t[0] for t in res_cursor]

        elif isinstance(audit_data[0],dict):
            result = []
            for row in audit_data:
                result.append(row.get('value_id'))
                if 'host' in row:
                    row['host_id'] = row.pop('host')
                keys = [k for k in row.keys()]
                for k in keys:
                    if not k in HostAuditData._meta.columns:
                        del row[k]
                row['created_on'] = datetime.datetime.utcnow()

            res_cursor = (HostAuditData.insert_many(audit_data)
                .on_conflict('update', conflict_target = 'id', preserve=['host_id','value_id','value_section','value_key','value','value_date','created_on'])
                .on_conflict('update', conflict_target = ['host_id','value_id'], preserve=['host_id','value_id','value_section','value_key','value','value_date','created_on'])
                ).execute()
            result = [t[0] for t in res_cursor]

        # remove obsolete
        HostAuditData.delete().where((HostAuditData.expiration_date != '') & (HostAuditData.expiration_date <= datetime2isodate(datetime.datetime.utcnow())) ).execute()
        message = '%s records inserted' % len(result)
        return make_response(result=result, msg=message, request_time=time.time() - starttime)

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('update_host failed %s' % (repr(e)))
        return make_response_from_exception(e)

def update_server_params():
    global full_version

    server_params = dict(
            version=__version__,
            full_version=full_version,
            uuid=get_server_uuid(),
            application_root=app.conf['application_root'],
            edition=wapt_edition,
            platform=platform.system(),
            architecture=platform.architecture(),
            fqdn = socket.getfqdn().lower()
            )

    with WaptDB():
        for key,value in server_params.items():
            (v, created) = ServerAttribs.get_or_create(key=key)
            v.value = value
            v.save()

# app mode
if __name__ == '__main__':
    usage = """\
    %prog [-c configfile] [--devel] [action]

    WAPT Server daemon.

    action is either :
      <nothing> : run service in foreground

    """
    git_hash = get_revision_hash()
    wapt_edition = get_wapt_edition()
    full_version = get_wapt_full_version()

    parser = OptionParser(usage=usage, version='waptserver.py ' + __version__)
    parser.add_option(
        '-c',
        '--config',
        dest='configfile',
        default=waptserver.config.DEFAULT_CONFIG_FILE,
        help='Config file full path (default: %default)')
    parser.add_option('-l', '--loglevel', dest='loglevel', default=None, type='choice',
                      choices=['debug',   'warning', 'info', 'error', 'critical'],
                      metavar='LOGLEVEL', help='Loglevel (default: warning)')
    parser.add_option('-d', '--devel', dest='devel', default=False, action='store_true',
                      help='Enable debug mode (for development only)')
    parser.add_option('--waptbasedir', dest='waptbasedir', default=wapt_root_dir,
                      help='Wapt installation base dir (default: %default)')

    for log in WAPTLOGGERS:
        parser.add_option('--loglevel_%s' % log, dest='loglevel_%s' % log, default=None, type='choice',
                          choices=['debug', 'warning', 'info', 'error', 'critical'],
                          metavar='LOGLEVEL', help='Loglevel %s (default: warning)' % log)

    (options, args) = parser.parse_args()
    wapt_root_dir = options.waptbasedir

    app.config['CONFIG_FILE'] = options.configfile
    app.conf.update(**waptserver.config.load_config(options.configfile))
    # merge command line options
    for att in options.__dict__:
        if att in app.conf and getattr(options, att) is not None:
            app.conf[att] = getattr(options, att)

    app.config['SECRET_KEY'] = app.conf.get('secret_key')
    app.config['APPLICATION_ROOT'] = app.conf.get('application_root', '')
    app.config['SESSION_COOKIE_PATH'] = app.config['APPLICATION_ROOT'] or '/'
    app.config['PERMANENT_SESSION_LIFETIME'] = datetime.timedelta(seconds=app.conf.get('session_lifetime'))


    setup_logging(app.conf)
    logger.info('Using config file %s' % options.configfile)

    logger.info('Using openssl %s' % cryptography.hazmat.backends.openssl.backend.openssl_version_text())

    # write an empty file (not client ssl protected) for console config test
    if not os.path.exists(os.path.join(app.conf['wapt_folder'],'ping')):
        with open(os.path.join(app.conf['wapt_folder'],'ping'),'a') as f:
            pass

    # monkey patch for greenlet and define a socketio on top of app
    from waptserver.server_socketio import socketio, proxy_host_request, update_ws_db

    if wsus:
        # add socketio targets to trigger wsus actions on hosts
        from waptserver import wsus_socketio
        app.register_blueprint(wsus_socketio.wsus_socketio)

    logger.info('Load database configuration')
    load_db_config(app.conf)

    # update huey sqlite storage filename based on configuration
    for task_def in app.enterprise_tasks:
        task_def.huey.storage_kwargs['filename'] = app.conf.get('wapt_huey_db')
        task_def.huey.storage.filename = app.conf.get('wapt_huey_db')
        task_def.huey.storage.close()

    # check wapt directories
    if not os.path.exists(app.conf['wapt_folder']):
        logger.warning('Folder missing : %s. Creating it' % app.conf['wapt_folder'])
        os.makedirs(app.conf['wapt_folder'])

    if not os.path.exists(app.conf['wapt_folder'] + '-host'):
       logger.warning('Folder missing : %s-host. Creating it' % app.conf['wapt_folder'])
       os.makedirs(app.conf['wapt_folder'] + '-host')

    if not os.path.exists(os.path.join(app.conf['wapt_folder'],'conf.d')):
       logger.warning('Folder missing : %s/conf.d. Creating it' % app.conf['wapt_folder'])
       os.makedirs(os.path.join(app.conf['wapt_folder'],'conf.d'))

    with WaptDB():
        if not get_db_version():
            try:
                init_db(drop=False,server_config=app.conf)
            except Exception as e:
                logger.critical('Unable to init DB: %s' % (repr(e)))
        try:
            upgrade_db_structure(server_config=app.conf)
        except Exception as e:
            logger.critical('Unable to upgrade DB structure: %s' % (repr(e)))
            raise

        # update server params for reporting
        update_server_params()

    logger.info('Using licences from %s' % licences_list.json_path)
    logger.info('Is enterprise: %s' % licences_list.is_enterprise())
    logger.info('Licences count: %s' % licences_list.hosts_count)

    logger.info('Waptserver starting...')
    port = app.conf['waptserver_port']
    licences_list.conf = app.conf

    if len(args)>0:
        action = args[0]
        if action =='update-crls':
            print(update_all_crls(app.conf))
            sys.exit(0)
        else:
            print('Actions: update-crls')


    # initialize socketio layer
    if socketio:
        logger.info('Starting socketio/wsgi server on port %s' % (port,))
        socketio.run(app, host=app.conf['wapt_bind_interface'], log=logger,  port=port, debug=options.devel, log_output=True, use_reloader=options.devel, max_size=app.conf['max_clients'])
    else:
        # debug wsgi mode mode
        logger.info('Starting wsgi debug server...')
        app.run(host=app.conf['wapt_bind_interface'], port=port, debug=options.devel)

    logger.info('Waptserver stopped')
else:
    # initialize WSGI app only
    # config filename is defined in  uwsgi config file (so refer to himself...)
    git_hash = get_revision_hash()
    wapt_edition = get_wapt_edition()
    full_version = get_wapt_full_version()

    app.config['CONFIG_FILE'] = os.environ['CONFIG_FILE']
    app.conf.update(**waptserver.config.load_config(app.config['CONFIG_FILE']))
    setup_logging(app.conf)
    app.config['SECRET_KEY'] = app.conf.get('secret_key')
    app.config['APPLICATION_ROOT'] = app.conf.get('application_root', '')
    app.config['SESSION_COOKIE_PATH'] = app.config['APPLICATION_ROOT'] or '/'
    app.config['PERMANENT_SESSION_LIFETIME'] = datetime.timedelta(seconds=app.conf.get('session_lifetime'))
    load_db_config(app.conf)

    licences_list.conf = app.conf

    # write an empty file (not client ssl protected) for console config test
    if not os.path.exists(os.path.join(app.conf['wapt_folder'],'ping')):
        with open(os.path.join(app.conf['wapt_folder'],'ping'),'a') as f:
            pass

    app.ws_connections = None

    # update huey sqlite storage filename based on configuration
    for task_def in app.enterprise_tasks:
        task_def.huey.storage_kwargs['filename'] = app.conf.get('wapt_huey_db')
        task_def.huey.storage.filename = app.conf.get('wapt_huey_db')
        task_def.huey.storage.close()

