#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2024  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import logging
import time
import traceback
import datetime

from waptserver.app import app, WS_SID, WS_TIMESTAMP

from waptutils import ensure_list

from waptserver.utils import EWaptAuthenticationFailure, EWaptForbiddden
from waptserver.utils import EWaptTimeoutWaitingForResult
from waptserver.common import get_secured_token_generator, get_server_uuid
from waptserver.common import make_response, make_response_from_exception

from waptserver.model import wapt_db, Hosts, HostWebsocket

from waptserver.auth import get_cn_from_x_ssl_client_auth

from flask import request, session
from flask_socketio import SocketIO
from flask_socketio import emit

logger = logging.getLogger('waptws')

# chain SocketIO server
socketio = SocketIO(app, logger=logger, engineio_logger=logger, cors_allowed_origins='*')

def update_ws_host_connection_from_db(uuid):
    with wapt_db:
        h = Hosts.select(Hosts.host_certificate).where(Hosts.uuid==uuid).tuples().first()
        if h:
            app.ws_connections[uuid] = [None,datetime.datetime.utcnow()] # uuid
            return True
        else:
            return False

def update_ws_db():
    with wapt_db.atomic():
        result = 0
        now = datetime.datetime.utcnow()
        HostWebsocket.delete().where(HostWebsocket.timestamp < now - datetime.timedelta(days=180)).execute()
        HostWebsocket.update(session_id=None,timestamp=now).where(HostWebsocket.server_uuid==get_server_uuid(), ~ HostWebsocket.session_id.is_null()).execute()
        for uuid,ws in app.ws_connections.items():
            if ws[WS_SID]:
                result += 1
            HostWebsocket.insert(
                host_id=uuid,
                server_uuid=get_server_uuid(),
                session_id=ws[WS_SID],
                timestamp= ws[WS_TIMESTAMP]
            ).on_conflict(
                conflict_target='host_id',
                update={
                    'host_id':uuid,
                    'server_uuid':get_server_uuid(),
                    'session_id': ws[WS_SID],
                    'timestamp':ws[WS_TIMESTAMP]
                    }
            ).execute()

    return result

def proxy_host_request(request, action):
    """Proxy a waptconsole action to wapt clients using websockets

    Args:
        uuid: can be a list or a single uuid
        notify_user: 0/1
        notify_server: 0/1

    Returns:
        dict:
            'result':
                'success' (list)
                'errors' (list)
            'msg' (str)
            'success'(bool)
            'request_time' (float)
            'error_code'
    """
    try:
        start_time = time.time()
        all_args = {k: v for k, v in request.args.items()}
        if request.json:
            all_args.update(request.json)

        uuids = ensure_list(all_args['uuid'])
        del(all_args['uuid'])
        timeout = float(request.args.get('timeout', app.conf.get('clients_read_timeout', 5)))

        result = dict(success=[], errors=[])
        tasks = []
        for uuid in uuids:
            try:
                if uuid in app.ws_connections:
                    ws = app.ws_connections[uuid]
                    msg = ''
                    args = dict(all_args)

                    def emit_action(sid, uuid, action, action_args, timeout=5):
                        try:
                            got_result = []

                            def result_callback(data):
                                got_result.append(data)

                            logger.debug('Emit %s to %s' % (action, uuid))
                            socketio.emit(action, action_args, room=sid, callback=result_callback)
                            # wait for asynchronous answer...
                            wait_loop = timeout * 20
                            while not got_result:
                                wait_loop -= 1
                                if wait_loop < 0:
                                    raise EWaptTimeoutWaitingForResult('Timeout, client did not send result within %s s' % timeout)
                                socketio.sleep(0.05)

                            # action succedded
                            result['success'].append(
                                dict(
                                    uuid=uuid,
                                    msg=msg,
                                    result=got_result[0],
                                ))
                        except Exception as e:
                            result['errors'].append(
                                dict(
                                    uuid=uuid,
                                    msg='%s' % repr(e),
                                ))
                    # if connected (and sid is known..)
                    if ws[WS_SID]:
                        logger.info(
                            'Launching %s with args %s for %s at address %s...' %
                            (action, all_args, uuid, ws[WS_SID]))

                        tasks.append(socketio.start_background_task(
                            emit_action, sid=ws[WS_SID], uuid=uuid, action=action, action_args=args))

                else:
                    result['errors'].append(
                        dict(
                            uuid=uuid,
                            msg='Host %s is not registered or not connected wia websockets' % uuid,
                        ))
            except Exception as e:
                result['errors'].append(
                    dict(
                        uuid=uuid,
                        msg='Host %s error %s' % (uuid, repr(e)),
                    ))

        # wait for all background tasks to terminate or timeout...
        # assume the timeout should be longer if more hosts to perform
        wait_loop = timeout * len(tasks)
        while True:
            running = [t for t in tasks if t.is_alive()]
            if not running:
                break
            wait_loop -= 1
            if wait_loop < 0:
                break
            socketio.sleep(0.05)

        msg = ['Success : %s, Errors: %s' % (len(result['success']), len(result['errors']))]
        if result['errors']:
            msg.extend(['%s: %s' % (e['computer_fqdn'], e['msg'])
                        for e in result['errors']])

        return make_response(result,
                             msg='\n- '.join(msg),
                             success=len(result['success']) > 0,
                             request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


# SocketIO Callbacks / handlers
@socketio.on('trigger_update_result')
def on_trigger_update_result(result):
    """Return from update on client"""
    logger.debug('Trigger Update result : %s (uuid:%s)' % (result, request.args['uuid']))
    # send to all waptconsole warching this host.
    socketio.emit('trigger_update_result', result, room=request.args['uuid'], include_self=False)


@socketio.on('trigger_upgrade_result')
def on_trigger_upgrade_result(result):
    """Return from the launch of upgrade on a client"""
    logger.debug('Trigger Upgrade result : %s (uuid:%s)' % (result, request.args['uuid']))
    socketio.emit('trigger_upgrade_result', result, room=request.args['uuid'], include_self=False)


@socketio.on('trigger_install_packages_result')
def on_trigger_install_packages_result(result):
    logger.debug('Trigger install result : %s (uuid:%s)' % (result, request.args['uuid']))
    socketio.emit('trigger_install_packages_result', result, room=request.args['uuid'], include_self=False)


@socketio.on('trigger_remove_packages_result')
def on_trigger_remove_packages_result(result):
    logger.debug('Trigger remove result : %s (uuid:%s)' % (result, request.args['uuid']))
    socketio.emit('trigger_remove_packages_result', result, room=request.args['uuid'], include_self=False)


@socketio.on('reconnect')
@socketio.on('connect')
def on_waptclient_connect():
    try:
        ws = None
        uuid = None
        # first we try to get an authenticated uuid from the request header
        # this uuid is set by nginx based on client certificate tls auth
        auth_uuid = get_cn_from_x_ssl_client_auth(request)

        if auth_uuid:
            uuid = auth_uuid
        else:
            uuid = request.args.get('uuid', None)
            if not uuid:
                uuid = request.headers.get('uuid', None)

        if not uuid:
            raise EWaptForbiddden('Missing source host uuid')

        # we need to check that uuid is known by this server

        # if we are on a uwsgi type setup, we need to check/get the uuid and certificate from the DB
        if not uuid in app.ws_connections and not update_ws_host_connection_from_db(uuid):
            raise EWaptForbiddden('Unknown host %s' % uuid)

        ws = app.ws_connections[uuid]

        allow_unauthenticated_connect = app.conf.get('allow_unauthenticated_connect', False)

        # if we are not yet authenticated with SSL, we use the old method with token in request arg or in request header
        if not auth_uuid and not allow_unauthenticated_connect:
            try:
                client_token = request.args.get('token',None)
                if not client_token:
                    authorization = request.headers.get('authorization')
                    if authorization:
                        (bearer,client_token) = authorization.split(' ',1)
                        if bearer != 'Bearer' or not client_token:
                            raise EWaptForbiddden('Bad header authorization token')

                if not client_token:
                    raise EWaptAuthenticationFailure('No authorization token')

                token_gen = get_secured_token_generator()
                max_age = app.conf['token_lifetime']
                token_data = token_gen.loads(client_token, max_age=max_age)
                auth_uuid = token_data.get('uuid', None)
                if not auth_uuid:
                    raise EWaptAuthenticationFailure('Bad host UUID')
                if token_data['server_uuid'] != get_server_uuid():
                    raise EWaptAuthenticationFailure('Bad server UUID')
            except Exception as e:
                raise EWaptAuthenticationFailure('SocketIO connection not authorized, invalid token: %s' % e)
            logger.info('Socket.IO connection from wapt client sid %s (uuid: %s)' % (request.sid, auth_uuid))
        else:
            logger.info('Unauthenticated Socket.IO connection from wapt client sid %s (uuid: %s)' % (request.sid, uuid))

        # update the SID in memory
        ws[WS_SID] = request.sid
        ws[WS_TIMESTAMP] = datetime.datetime.utcnow()

        if auth_uuid:
            session['uuid'] = auth_uuid
        else:
            session['uuid'] = uuid
            session['unsafe_uuid'] = True

        return True

    except Exception as e:
        if ws:
            ws[WS_SID] = None
            ws[WS_TIMESTAMP] = datetime.datetime.utcnow()

        if 'uuid' in session:
            session.pop('uuid')

        logger.warning('SocketIO connection refused for uuid %s, sid %s: %s, instance %s' % (uuid, request.sid, e, app.conf.get('application_root')))
        return False


@socketio.on('wapt_pong')
def on_wapt_pong():
    uuid = None
    try:
        uuid = session.get('uuid')
        if not uuid:
            logger.critical('SocketIO %s connected but no host uuid in session: asking connected host to reconnect' % (request.sid))
            emit('wapt_force_reconnect')
            return False
        else:
            logger.debug('Socket.IO pong from wapt client sid %s (uuid: %s)' % (request.sid, session.get('uuid', None)))
            # stores sid in database

        if not uuid in app.ws_connections:
            emit('wapt_force_reconnect')
            return False

        ws = app.ws_connections[uuid]
        ws[WS_SID] = request.sid
        ws[WS_TIMESTAMP] = datetime.datetime.utcnow()

        session['uuid'] = uuid
        return True

    except Exception:
        logger.critical('SocketIO pong error for uuid %s and sid %s : %s, instance: %s' % (uuid, request.sid, traceback.format_exc(), app.conf.get('application_root')))
        return False


@socketio.on('disconnect')
def on_waptclient_disconnect():
    uuid = session.get('uuid', None)
    logger.info('Socket.IO disconnection from wapt client sid %s (uuid: %s)' % (request.sid, uuid))

    # clear sid in database
    if uuid in app.ws_connections:
        ws = app.ws_connections[uuid]
        if ws[WS_SID] == request.sid and (not ws[WS_TIMESTAMP] or ws[WS_TIMESTAMP] < datetime.datetime.utcnow()):
            ws[WS_SID] = None
            ws[WS_TIMESTAMP] = datetime.datetime.utcnow()
            if 'uuid' in session:
                session.pop('uuid')
        else:
            logger.info('Socket.IO for %s has already got a new SID %s replacing SID to disconnect %s. Keep it' % (uuid, ws[WS_SID], request.sid))
            ws[WS_TIMESTAMP] = datetime.datetime.utcnow()

    return True

@socketio.on_error()
def on_wapt_socketio_error(e):
    logger.critical('Socket IO : An error has occurred for sid %s, uuid:%s : %s' % (request.sid, request.args.get('uuid', None), repr(e)))

# end websockets
