#!/opt/wapt/bin/python
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2024  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import sys
import os

if "__file__" in locals():
    sys.path.insert(0, os.path.realpath(os.path.join( os.path.dirname(os.path.realpath(__file__)),'..')))

import waptserver.config
from waptutils import *
from waptserver.utils import *
from waptserver.model import *
from playhouse.migrate import *
from optparse import OptionParser
import configparser
import logging
import requests
import glob
__version__ = '1.5.0'
usage = """
This script if aimed at triggering package upgrade through direct http call
for client waptagent <=1.4. It is usefull for triggering upgrades from earlier
waptagent version to wapt 1.5 which uses websockets.

Note :
client version<=1.4 are polled from the server using http requests.
client version >=1.5 use websockets from clients.

trigger_install_all.py [-c configfile] [-l loglevel] [-t timeout] package
"""


try:
    wapt_root_dir = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'


DEFAULT_CONFIG_FILE = os.path.join(wapt_root_dir, 'conf', 'waptserver.ini')
config_file = DEFAULT_CONFIG_FILE

# setup logging
logger = logging.getLogger()
logging.basicConfig()

if __name__ == '__main__':
    parser = OptionParser(usage=usage, version='waptserver.py ' + __version__)
    parser.add_option(
        '-c',
        '--config',
        dest='configfile',
        default=DEFAULT_CONFIG_FILE,
        help='Config file full path (default: %default)')
    parser.add_option(
        '-l',
        '--loglevel',
        dest='loglevel',
        default='info',
        type='choice',
        choices=[
            'debug',
            'warning',
            'info',
            'error',
            'critical'],
        metavar='LOGLEVEL',
        help='Loglevel (default: warning)')
    parser.add_option(
        '-t',
        '--timeout',
        dest='timeout',
        default=0.5,
        type='float',
        help='Timeout (default: %default)')

    (options, packages) = parser.parse_args()
    if not packages:
        print(parser.usage)
        sys.exit(1)

    conf = waptserver.config.load_config(options.configfile)
    load_db_config(conf)

    timeout = 0.5

    hosts = Hosts.select(Hosts.uuid, Hosts.computer_fqdn,
                         Hosts.listening_address, Hosts.listening_port, Hosts.connected_ips, Hosts.wapt_status).where(~Hosts.listening_protocol.is_null() and ~Hosts.connected_ips.is_null())
    for host in hosts:
        if Version(host.wapt_status['wapt-exe-version']) < Version('1.5.0'):
            logger.info('processing %s' % host.computer_fqdn)
            args = {}
            args['uuid'] = host.uuid
            args['protocol'] = host.listening_protocol
            args['port'] = host.listening_port
            args['package'] = ','.join(packages)
            logger.debug(host.connected_ips)
            for address in ensure_list(host.connected_ips):
                try:
                    args['address'] = address
                    logger.info('sending update command')
                    client_result = requests.get('http://%(address)s:8088/register.json?uuid=%(uuid)s&force=1&notify_server=1' % args,
                                                 proxies={'http': None, 'https': None}, verify=False, timeout=options.timeout).text
                    try:
                        client_result = json.loads(client_result)
                        logger.debug(client_result)
                    except ValueError:
                        if 'Restricted access' in client_result:
                            logger.error('Forbidden %s' % host.computer_fqdn)
                        else:
                            logger.error('Error %s : %s' % repr(e))

                    logger.info('sending install command')
                    client_result = requests.get('http://%(address)s:8088/install.json?uuid=%(uuid)s&package=%(package)s&force=1' % args,
                                                 proxies={'http': None, 'https': None}, verify=False, timeout=options.timeout).text
                    try:
                        client_result = json.loads(client_result)
                        logger.debug(json.dumps(client_result, indent=True))
                        logger.info('OK')
                        break
                    except ValueError:
                        if 'Restricted access' in client_result:
                            logger.error('Forbidden %s' % host.computer_fqdn)
                        else:
                            logger.error('Error %s : %s' % repr(e))
                except (requests.ConnectTimeout, requests.ConnectionError, requests.ReadTimeout):
                    logger.error('No answer %s' % address)
