#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2024  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
import os
import sys

if "__file__" in locals():
    sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))

try:
    wapt_root_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'


from waptutils import setloglevel
from waptserver.common import update_all_crls

import waptserver.config

from optparse import OptionParser
import logging

logger = logging.getLogger()

__doc__ = """\
%prog <wapt_directory>

    Sign CRL of hosts pki if needed (expire in less than 2 days, or force)
    Merge CRL of hosts, CRLs of all certificates in packages, and not expired CRLs in 'ssl_additional_crls' directory.
    Put the resulting CRLs bundle in 'ssl_client_crls' file for use by NGINX.
"""


def main():
    parser = OptionParser(usage=__doc__)
    parser.add_option("-f", "--force",    dest="force",    default=False, action='store_true', help="Force (default: %default)")
    parser.add_option("-l", "--loglevel", dest="loglevel", default=None, type='choice',  choices=['debug', 'warning', 'info', 'error', 'critical'], metavar='LOGLEVEL', help="Loglevel (default: warning)")
    parser.add_option("-p", "--proxy",    dest="proxy",    default=None, help="http proxy (default: %default)")
    parser.add_option(
        '-c',
        '--config',
        dest='configfile',
        default=waptserver.config.DEFAULT_CONFIG_FILE,
        help='Config file full path (default: %default)')
    (options, args) = parser.parse_args()

    conf = waptserver.config.load_config(options.configfile)

    loglevel = options.loglevel

    if len(logger.handlers) < 1:
        hdlr = logging.StreamHandler(sys.stderr)
        hdlr.setFormatter(logging.Formatter(
            '%(asctime)s %(levelname)s %(message)s'))
        logger.addHandler(hdlr)

    if loglevel:
        setloglevel(logger, loglevel)
    else:
        setloglevel(logger, 'warning')

    crls_updated = update_all_crls(conf,force=options.force)

    if crls_updated and os.name == 'posix':
        logger.info('Set CRLs file ownership to wapt')
        import pwd
        pwd_entry = pwd.getpwnam('wapt')
        uid, gid = pwd_entry.pw_uid, pwd_entry.pw_gid
        if os.path.isfile(conf['clients_signing_crl']):
            os.chown(conf['clients_signing_crl'], uid, gid)  # pylint: disable=no-member
        if os.path.isfile(conf['ssl_client_crls']):
            os.chown(conf['ssl_client_crls'], uid, gid)  # pylint: disable=no-member


if __name__ == "__main__":
    main()
