# -*- coding: utf-8 -*-
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2024  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import os
import sys

if "__file__" in locals():
    sys.path.insert(0, os.path.realpath(os.path.join( os.path.dirname(os.path.realpath(__file__)),'..')))
# there is a security bug in this module. known files are by default ['/etc/mime.types', '/etc/... even on winddows
# this allow a regular user to trigger memory overflow if he create a huge file /etc/mime.type
import mimetypes
mimetypes.knownfiles.clear()

from waptservice.waptservice_common import WaptUpgrade, WaptUpdate, WaptUpdateServerStatus, WaptCleanup, WaptDownloadPackage, WaptLongTask
from waptservice.waptservice_common import WaptAuditPackage, WaptPackageRemove, WaptDownloadIcon, WaptDownloadUpgrade

from waptservice.waptservice_common import WaptClientUpgrade, WaptServiceRestart, WaptNetworkReconfig, WaptPackageInstall
from waptservice.waptservice_common import forbidden, authenticate, allow_local, render_wapt_template
from waptservice.waptservice_common import WaptEvents, WaptRegisterComputer, WaptPackageForget
from waptservice.waptservice_common import waptconfig, WAPTLOGGERS
from waptpackage import PackageEntry, WaptLocalRepo
from flask import request, Flask, Response, send_from_directory, session, g, redirect, url_for
from waptservice.waptservice_socketio import WaptSocketIOClient
from setuphelpers import Version
import setuphelpers
from common import Wapt
import common
import random,string
from waptutils import setloglevel, ensure_list, ensure_unicode, jsondump, LogOutput, get_time_delta, is_between_two_times, run, makepath
from waptutils import harakiri, is_local_user, user_config_directory, user_is_local_admin, user_is_member_of
from waptutils import EWaptAuthException
from waptutils import datetime2isodate

if sys.platform=='linux':
    from waptutils import LinuxImpersonate

import ctypes
import datetime
import copy
import traceback
import queue
import threading
import json
import gc
import sqlite3
import logging
from functools import wraps
from werkzeug.utils import secure_filename
import jinja2
import hashlib
from waitress import serve
from optparse import OptionParser
from waptutils import __version__
import time
import base64
import glob
import auth_module_ad


v = (sys.version_info.major, sys.version_info.minor)
if v[0] != 3:
    raise Exception('wapt-get supports only Python 3, not %d.%d' % v)

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
except NameError:
    wapt_root_dir = 'c:/tranquilit/wapt'

if sys.platform.startswith('linux') or sys.platform.startswith('darwin'):
    if 'PYTHONPATH' in os.environ:
        del os.environ['PYTHONPATH']
    if 'PYTHONHOME' in os.environ:
        del os.environ['PYTHONHOME']

if not sys.platform == 'win32':
    import setproctitle
    setproctitle.setproctitle('waptservice')

if sys.platform == 'win32':
    import win32api
    import pythoncom
    import win32security
    from setuphelpers import local_users
else:
    import pam
    from setuphelpers_unix import get_groups


if sys.platform == 'win32':
    from waptservice.enterprise import WaptWUADowloadTask, WaptWUAInstallTask, waptwua_api
    from waptwua.client import WaptWUA, WaptWUAParams
    import pywintypes

from waptservice.enterprise import WaptRunSessionSetup
from waptservice.repositories import WaptSyncRepo


try:
    from flask_babel import Babel
    try:
        from flask_babel import gettext
    except ImportError:
        gettext = (lambda s: s)

except:
    gettext = (lambda s: s)
    Babel = None

# i18n
_ = gettext

logger = logging.getLogger('waptservice')
tasks_logger = logging.getLogger('wapttasks')


def format_isodate(isodate):
    """Pretty format iso date like : 2014-01-21T17:36:15.652000

    >>> format_isodate('2014-01-21T17:36:15.652000')
    '2014-01-21 17:36:15'
    """
    return isodate.replace('T', ' ')[0:19]
    #dateutil.parser.parse(isodate).strftime("%d/%m/%Y %H:%M:%S")


def beautify(c):
    """return pretty html"""
    join = "".join
    if c is None:
        return ""
    elif isinstance(c, (datetime.datetime, datetime.date)):
        return c.isoformat()
    elif isinstance(c, int):
        return '{}'.format(c)
    elif isinstance(c, float):
        return '{:.3}'.format(c)
    elif isinstance(c, str):
        return jinja2.Markup(c.replace('\r\n', '<br>').replace('\n', '<br>'))
    elif isinstance(c, str):
        return jinja2.Markup(ensure_unicode(c).replace('\r\n', '<br>').replace('\n', '<br>'))
    elif isinstance(c, PackageEntry):
        return jinja2.Markup('<a href="%s">%s</a>' % (url_for('package_details', package=c.asrequirement()), c.asrequirement()))
    elif isinstance(c, dict) or (hasattr(c, 'keys') and callable(c.keys)):
        rows = []
        try:
            for key in list(c.keys()):
                rows.append('<li><b>{}</b>: {}</li>'.format(beautify(key), beautify(c[key])))
            return jinja2.Markup('<ul>{}</ul>'.format(join(rows)))
        except:
            pass
    elif isinstance(c, (list, tuple)):
        if c:
            rows = ['<li>{}</li>'.format(beautify(item)) for item in c]
            return jinja2.Markup('<ul>{}</ul>'.format(join(rows)))
        else:
            return ''
    else:
        return jinja2.Markup("<pre>{}</pre>".format(ensure_unicode(c)))


app = Flask(__name__)
app.config['PROPAGATE_EXCEPTIONS'] = True
# ephemeral cookies key
app.config['SECRET_KEY'] = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(64))

app.name = 'waptservice'

if sys.platform == 'win32':
    app.register_blueprint(waptwua_api)


app.jinja_env.filters['beautify'] = beautify  # pylint: disable=no-member
app.waptconfig = waptconfig

if Babel:
    app_babel = Babel(app)

    @app_babel.localeselector
    def get_locale():
        browser_lang = request.accept_languages.best_match(['en', 'fr'])
        user_lang = session.get('lang', browser_lang)
        return user_lang

    @app_babel.timezoneselector
    def get_timezone():
        user = getattr(g, 'user', None)
        if user is not None:
            return user.timezone

else:
    app_babel = None


def apply_host_settings(waptconfig):
    """Apply waptservice / waptexit specific settings
    """
    wapt = Wapt(config_filename=waptconfig.config_filename)
    try:
        if waptconfig.max_gpo_script_wait is not None and wapt.max_gpo_script_wait != waptconfig.max_gpo_script_wait:
            logger.info('Setting max_gpo_script_wait to %s' % waptconfig.max_gpo_script_wait)
            wapt.max_gpo_script_wait = waptconfig.max_gpo_script_wait
        if waptconfig.pre_shutdown_timeout is not None and wapt.pre_shutdown_timeout != waptconfig.pre_shutdown_timeout:
            logger.info('Setting pre_shutdown_timeout to %s' % waptconfig.pre_shutdown_timeout)
            wapt.pre_shutdown_timeout = waptconfig.pre_shutdown_timeout
        if waptconfig.hiberboot_enabled is not None and wapt.hiberboot_enabled != waptconfig.hiberboot_enabled:
            logger.info('Setting hiberboot_enabled to %s' % waptconfig.hiberboot_enabled)
            wapt.hiberboot_enabled = waptconfig.hiberboot_enabled
    except Exception as e:
        logger.critical('Unable to set shutdown policies : %s' % e)


def apply_waptwua_settings(waptconfig):
    """Apply waptwua service specific settings
    """
    wapt = Wapt(config_filename=waptconfig.config_filename)
    if wapt.is_enterprise():
        try:
            c = WaptWUA(wapt)
            c.apply_waptwua_settings_to_host()
        except Exception as e:
            logger.critical('Unable to set waptwua policies : %s' % e)


def scoped_wapt():
    """Flask request contextual cached Wapt instance access"""
    if not hasattr(g, 'wapt'):
        g.wapt = Wapt(config_filename=waptconfig.config_filename)
        if sys.platform == 'win32':
            apply_host_settings(waptconfig)
    # apply settings if changed at each wapt access...
    elif g.wapt.reload_config_if_updated():
        # apply waptservice / waptexit specific settings
        if sys.platform == 'win32':
            apply_host_settings(waptconfig)
    #logger.debug('returning a scoped Wapt instance for flask')
    return g.wapt


@app.teardown_appcontext
def close_connection(exception):
    try:
        local_wapt = getattr(g, 'wapt', None)
        if local_wapt is not None and local_wapt._waptdb and local_wapt._waptdb.transaction_depth > 0:
            try:
                local_wapt._waptdb.commit()
                local_wapt._waptdb = None
            except:
                try:
                    local_wapt._waptdb.rollback()
                    local_wapt._waptdb = None
                except:
                    local_wapt._waptdb = None

    except Exception as e:
        logger.debug('Error in teardown, please consider upgrading Flask if <0.10. %s' % e)


def check_auth_groups(auth_header):
    """Authenticate a user and returns the self-service groups membership

    Args:
        auth_header (flask.request): request with Authorization header
        check_for_groups (list): restrict the search to this list of self service groups

    Returns:
        list: (username,list of user's self service groups memberships ex: ['compta','tech'])
    """

    # Domain admins and local admins are allowed
    # make a copy to not change global settings
    try:
        # try to get auth and groups from Bearer token
        w = scoped_wapt()
        auths = [b.strip() for b in auth_header.split(',')]
        for auth in auths:
            (bearer,token) = auth.split(' ',1)
            if bearer == 'Bearer':
                token_gen = w.get_secured_token_generator(app.config['SECRET_KEY'])
                try:
                    max_age = w.token_lifetime
                    token_content = token_gen.loads(token, max_age=max_age)
                    username = token_content.get('username', [])
                    usergroups = token_content.get('groups',[])
                    return (username,usergroups)
                except:
                    # password is not a token or token is invalid
                    pass

            # check if a valid Basic authorization header is decoded
            if not bearer == 'Basic':
                return (None,None)

            # decode Basic header
            (logon_name,password) = base64.decodebytes(token.encode('ascii')+b'==').decode('utf8').split(':',1)

            # sanity check on logon name
            domain = ''
            if logon_name.count('\\') > 1 or logon_name.count('@') > 1 or (logon_name.count('\\') == 1 and logon_name.count('@') == 1):
                tasks_logger.warning("malformed logon credential : %s " % logon_name)
                return (None,None)

            tasks_logger.debug('service_auth_type : %s' % app.waptconfig.service_auth_type)

            # we need a scoped list of groups to check to avoid unecessary requests
            # groups are those defined in rules + waptselfservice + admin groups
            rules = None
            if w.is_enterprise():
                rules = common.self_service_rules(w)
                check_for_groups = list(rules.keys())
                check_for_groups.append('waptselfservice')
            else:
                check_for_groups=['waptselfservice']

            # auth through wapt server
            if app.waptconfig.service_auth_type == 'waptserver-ldap':
                if not w.is_enterprise():
                    raise Exception("waptagent-ldap for service auth type is an enterprise feature only")
                r =  w.self_service_auth(logon_name, password, w.host_uuid, check_for_groups)
                if r['result']['error']:
                    msg = 'Self service authentication failed: %s' % r['result']['msg']
                    raise EWaptAuthException(msg)
                if not r['result']['success']:
                    msg = 'WRONG_PASSWORD_USERNAME'
                    tasks_logger.warning(msg)
                    raise EWaptAuthException(msg)
                result = (logon_name , r['result']['groups'])
            # auth directly on a ldap server
            elif app.waptconfig.service_auth_type == 'waptagent-ldap':
                if not w.is_enterprise():
                    raise EWaptAuthException("waptagent-ldap for service auth type is an enterprise feature only")
                ad_result = auth_module_ad.check_credentials_ad(
                            conf = {'ldap_auth_server': app.waptconfig.ldap_auth_server,
                                    'ldap_auth_base_dn': app.waptconfig.ldap_auth_base_dn,
                                    'ldap_auth_ssl_enabled': app.waptconfig.ldap_auth_ssl_enabled,
                                    'verify_cert_ldap':app.waptconfig.verify_cert_ldap},
                            username = logon_name,
                            password = password,
                            list_group = check_for_groups)
                if ad_result['error']:
                    raise EWaptAuthException(ad_result['msg'])
                if not ad_result['success']:
                    raise EWaptAuthException('WRONG_PASSWORD_USERNAME')
                result = (logon_name,ad_result['groups'])

            # auth with local system authentication (win32 or pam)
            elif app.waptconfig.service_auth_type == 'system' or (app.waptconfig.service_auth_type == 'nopassword' and password != ''):  # basic auth fallback
                # extract domain from logon
                if '\\' in logon_name:
                    domain = logon_name.split('\\')[0]
                    username = logon_name.split('\\')[1]
                elif '@' in logon_name:
                    username = logon_name.split('@')[0]
                    domain = logon_name.split('@')[1]
                else:
                    username = logon_name

                if sys.platform == 'win32':
                    if domain == '' :
                        if username.lower() in [i.lower() for i in local_users()]:
                            domain = '.'
                        else:
                            domain = os.environ['USERDOMAIN']

                if app.waptconfig.waptservice_password and app.waptconfig.waptservice_user == username and app.waptconfig.waptservice_password:
                    tasks_logger.debug('Auth using wapt local account')
                    # for waptstarter
                    if app.waptconfig.waptservice_password == hashlib.sha256(password.encode('utf8')).hexdigest():
                        result = (username,'waptselfservice')
                    else:
                        result = (None,None)

                elif sys.platform == 'win32':
                    try:
                        huser = win32security.LogonUser(username, domain, password, win32security.LOGON32_LOGON_NETWORK_CLEARTEXT,getattr(win32security, app.waptconfig.logon32_provider))
                    except pywintypes.error as e:
                        if e.args[0] == 1326:
                            msg = 'WRONG_PASSWORD_USERNAME\n%s' % e.args[2]
                        else:
                            msg = e.args[2]
                        tasks_logger.warning(msg)
                        raise EWaptAuthException(msg)
                    usergroups = []
                    if username in check_for_groups:
                        usergroups.append(username)
                    for group in check_for_groups:
                        if group in usergroups:
                            continue
                        if common.check_is_member_of(huser, group):
                            usergroups.append(group)

                    # check admins
                    # check if user is domain admins or member of waptselfservice
                    if not 'waptselfservice' in usergroups:
                        try:
                            domain_admins_group = common.get_domain_admins_group_name()
                            if common.check_is_member_of(huser, domain_admins_group):
                                usergroups.append('waptselfservice')
                            else:
                                if not (app.waptconfig.waptservice_admin_filter):
                                    local_admin_group = common.get_local_admins_group_name()
                                    if common.check_is_member_of(huser, local_admin_group):
                                        usergroups.append('waptselfservice')
                        except Exception as e:
                            tasks_logger.warning('Authentication issue %s' % (e,))

                    result = (username,usergroups)

                else:
                    usergroups = get_user_self_service_groups_unix(username, password, check_for_groups)
                    # add waptselfservice for admin
                    if not 'waptselfservice' in usergroups:
                        for group in ['root','sudo','wheel']:
                            if group in usergroups:
                                usergroups.append('waptselfservice')
                                break
                    result = (username,usergroups)
            else:
                return (None,None)

        tasks_logger.debug('check_auth_groups: %s' % (result,))
        return result
    except Exception as e:
        tasks_logger.warning('check_auth_groups: %s' % (e,))
        raise


def get_user_self_service_groups_unix(logon_name, password, check_for_groups = ['waptselfservice']):
    """Authenticate the user with pam, then returns all groups from check_groups to which
    the user belongs

    Returns:
        list of groups
    """
    p = pam.pam()
    success = p.authenticate(logon_name, password)
    if not success:
        raise EWaptAuthException('WRONG_PASSWORD_USERNAME')
    else:
        all_groups = []
        if logon_name in check_for_groups:
            all_groups.append(logon_name)
        user_groups = get_groups(logon_name)
        for group in check_for_groups:
            if group in user_groups:
                all_groups.append(group)
        return all_groups


def allow_local_auth(f):
    """Restrict access to localhost authenticated"""
    @wraps(f)
    def decorated(*args, **kwargs):
        if request.remote_addr in ['127.0.0.1']:
            auth_header=request.headers.get('Authorization')
            if not auth_header:
                return authenticate()

            try:
                (username,groups) = check_auth_groups(auth_header)
                if not username:
                    logger.info("Authentication failed")
                    return authenticate()
            except:
                logger.info("Authentication failed")
                return authenticate()

            session.username = username
            session.usergroups = groups
        else:
            return forbidden()
        return f(*args, **kwargs)
    return decorated


@app.route('/lang/<language>')
def lang(language=None):
    session['lang'] = language
    return redirect('/')


@app.route('/ping')
@allow_local
def ping():
    if 'uuid' in request.args:
        w = scoped_wapt()
        data = dict(
            hostname=setuphelpers.get_hostname(),
            version=Version(__version__,3),
            uuid=w.host_uuid,
            waptserver=w.waptserver,
        )
    else:
        data = dict(
            version=Version(__version__,3),
        )
    return Response(common.jsondump(data), mimetype='application/json')

def write_token_file(username,data):
    user_config_dir = user_config_directory(username)
    if not user_config_dir:
        raise Exception('invalid user')
    if not os.path.isdir(user_config_dir):
        os.makedirs(user_config_dir)
    user_wapt_appdir= makepath(user_config_dir,'wapt')
    if not os.path.isdir(user_wapt_appdir):
        os.makedirs(user_wapt_appdir)
    token_filepath = makepath(user_wapt_appdir,'waptservicetoken')
    if sys.platform == 'linux':
        with LinuxImpersonate(username):
            with open(token_filepath,'w') as f:
                f.write(data)
    else:
        # on windows, acls are inherited in the home dir ... ?
        with open(token_filepath,'w') as f:
            f.write(data)
    return token_filepath

@app.route('/login')
@allow_local
def login():
    authenticated_username = None
    groups = []

    # we assume we need a valid Basic auth header here (token based auth is not valid)
    # this is like logout / login
    w = scoped_wapt()
    if request.authorization:
        try:
            auth_header = request.headers.get('Authorization') # either a Basic or a Bearer
            (authenticated_username,groups) = check_auth_groups(auth_header)
        except EWaptAuthException as e:
            return authenticate(msg='WRONG_PASSWORD_USERNAME %s' % e)
        except Exception as e:
            if (sys.platform == 'win32') and isinstance(e,pywintypes.error):
                return authenticate(msg='WRONG_PASSWORD_USERNAME %s' % e.strerror)
            return authenticate(msg='UNHANDLED_ERROR %s' % e.__class__)
    else:
        return authenticate(msg='MISSING_AUTH')

    # generate a token for next requests
    token_gen = w.get_secured_token_generator(app.config['SECRET_KEY'])

    if not authenticated_username:
        # no user/password auth, but we can try to provide a token in a file
        if app.waptconfig.service_auth_type.lower() == 'nopassword' and is_local_user(request.authorization.username):
            # check if user is a local admin of member of waptselfservice
            groups = []
            rules = common.self_service_rules(w)
            check_for_groups = list(rules.keys())
            if not 'waptselfservice' in check_for_groups:
                check_for_groups.append('waptselfservice')
            for group in check_for_groups:
                if user_is_member_of(request.authorization.username,group):
                    groups.append(group)
            if not 'waptselfservice' in groups and (
                    (user_is_local_admin(request.authorization.username) and not app.waptconfig.waptservice_admin_filter) or app.waptconfig.waptservice_allow_all_packages):
                groups.append('waptselfservice')
            if not groups:
                return authenticate(msg='NOT_ADMIN_AND_NOT_SELFSERVICE_MEMBER')
            auth_data = {'username': request.authorization.username, 'groups': groups}
            token = token_gen.dumps(auth_data)
            token_filepath = write_token_file(request.authorization.username,token)
            auth_data['token_filepath'] = token_filepath
            return Response(common.jsondump(auth_data), mimetype='application/json')
        else:
            return authenticate(msg='WRONG_PASSWORD_USERNAME')
    else:
        token = token_gen.dumps({'username': authenticated_username, 'groups': groups})
        return Response(common.jsondump({'token': token, 'username': authenticated_username, 'groups': groups}), mimetype='application/json')


@app.route('/status')
@app.route('/status.json')
@allow_local
def status():
    rows = []
    with sqlite3.connect(app.waptconfig.dbpath) as con:
        try:
            con.row_factory = sqlite3.Row
            query = '''select s.package_uuid,s.package,s.version,s.install_date,
                                 s.install_status,s.install_output,r.icon_sha256sum,r.description,
                                 (select GROUP_CONCAT(p.version,"|") from wapt_package p where p.package=s.package) as repo_versions,
                                 explicit_by as install_par
                                 from wapt_localstatus s
                                 left join wapt_package r on r.package_uuid=s.package_uuid
                                 order by s.package'''
            cur = con.cursor()
            cur.execute(query)
            rows = []
            search = request.args.get('q', '')

            for row in cur.fetchall():
                pe = PackageEntry()
                rec_dict = dict((cur.description[idx][0], value) for idx, value in enumerate(row))
                for k in rec_dict:
                    setattr(pe, k, rec_dict[k])

                # hack to enable proper version comparison in templates
                pe.version = Version(pe.version)
                # calc most up to date repo version
                if pe.get('repo_versions', None) is not None:
                    pe.repo_version = max(Version(v) for v in pe.get('repo_versions', '').split('|'))
                else:
                    pe.repo_version = None

                if not search or pe.match_search(search):
                    rows.append(pe)

        except sqlite3.Error as e:
            logger.critical("*********** Error %s:" % e.args[0])
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(rows), mimetype='application/json')
    else:
        return render_wapt_template('status.html', packages=rows, format_isodate=format_isodate, Version=setuphelpers.Version)


def latest_only(packages):
    index = {}
    for p in sorted(packages, reverse=True):
        if not p.package in index:
            p.previous = []
            index[p.package] = p
        else:
            index[p.package].previous.append(p)

    return list(index.values())


@app.route('/keywords.json')
@allow_local
def keywords():
    with sqlite3.connect(app.waptconfig.dbpath) as con:
        try:
            con.row_factory = sqlite3.Row
            cur = con.cursor()
            rows = []
            query = "select distinct trim(keywords) from wapt_package where keywords is not null and trim(keywords)<>''"
            rows = cur.execute(query)
            result = {}
            for k in rows:
                kws = k[0].lower().split(',')
                for kw in kws:
                    kwt = kw.strip().capitalize()
                    if not kwt in result:
                        result[kwt] = 1
                    else:
                        result[kwt] += 1
            return Response(common.jsondump(sorted(result.keys())), mimetype='application/json')
        except Exception as e:
            logger.critical('Error: %s' % e)
            return Response(common.jsondump([]), mimetype='application/json')


@app.route('/check_install')
@app.route('/check_install.json')
@allow_local
def check_install():
    try:
        package = request.args.get('package')
        data = scoped_wapt().check_install(package)
        return Response(common.jsondump(data), mimetype='application/json')
    except Exception as e:
        logger.critical("*********** Error %s:" % e.args[0])
        return Response(common.jsondump(e), mimetype='application/json')


@app.route('/list/pg<int:page>')
@app.route('/packages.json')
@app.route('/packages')
@app.route('/list')
@allow_local
def all_packages(page=1):
    #############
    auth_header = request.headers.get('Authorization')
    if not auth_header:
        return authenticate('Missing auth')

    try:
        (username,usergroups) = check_auth_groups(auth_header)
        if not username:
            return authenticate()
        else:
            logger.info('User %s authenticated. Groups: %s' % (username,usergroups))
    except EWaptAuthException as e:
        return authenticate("Auth failed on local service: %s" % e)


    w = scoped_wapt()
    # on enterprise, rules are needed to know if packages removal is allowed
    rules = None
    if w.is_enterprise():
        rules = common.self_service_rules(w)
    ##############

    with sqlite3.connect(app.waptconfig.dbpath) as con:
        try:
            con.row_factory = sqlite3.Row
            query = '''\
                select
                    r.*,
                    s.version as install_version,s.install_status,s.install_date,s.explicit_by
                from wapt_package r
                left join wapt_localstatus s on s.package=r.package
                where not r.section in ("host","unit","profile","restricted","selfservice")
                order by r.package,r.version'''
            cur = con.cursor()
            cur.execute(query)
            rows = []

            search = request.args.get('q', '').replace('\\', '')
            keywords = ensure_list(request.args.get('keywords', '').lower())

            for row in cur.fetchall():
                pe = PackageEntry().load_control_from_dict(
                    dict((cur.description[idx][0], value) for idx, value in enumerate(row)))

                if len(keywords) > 0:
                    match_kw = False
                    package_keywords = ensure_list(pe.keywords.lower())
                    for kw in package_keywords:
                        if kw in keywords:
                            match_kw = True
                            break
                    if not match_kw:
                        continue

                if not search or pe.match_search(search):
                    if w.is_authorized_package_action('list', pe.package, usergroups, rules):
                        rows.append(pe)

            if request.args.get('latest', '0') == '1':
                filtered = []
                last_package_name = None
                for package in sorted(rows, reverse=True):
                    if package.package != last_package_name:
                        filtered.append(package)
                    last_package_name = package.package
                rows = list(reversed(filtered))

            if not request.args.get('all_versions', ''):
                rows = sorted(latest_only(rows))

        except sqlite3.Error as e:
            logger.critical("*********** Error %s:" % e.args[0])

    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        for pe in rows:
            # some basic search scoring
            score = 0
            if search in pe.package:
                score += 3
            if search in pe.description:
                score += 2
            pe.score = score
        rows = sorted(rows, key=lambda r: (r.score, r.signature_date, r.filename), reverse=True)
        return Response(common.jsondump(rows), mimetype='application/json')
    else:
        for pe in rows:
            # hack to enable proper version comparison in templates
            pe.install_version = Version(pe.install_version)
            pe.version = Version(pe.version)

        try:
            search = search
        except NameError:
            search = False

        pagination = None
        return render_wapt_template(
            'list.html',
            packages=rows,  # [_min:_max],
            format_isodate=format_isodate,
            Version=setuphelpers.Version,
            pagination=pagination,
        )


@app.route('/package_icon')
@allow_local
def package_icon():
    icon_sha256sum = request.args.get('icon_sha256sum')
    icons_dir = os.path.join(wapt_root_dir,'cache','icons')
    if icon_sha256sum:
        icon_fn = '%s.png' % icon_sha256sum
        if os.path.isfile(os.path.join(icons_dir,icon_fn)):
            return send_from_directory(icons_dir,icon_fn)
        else:
            return send_from_directory(icons_dir,'unknown.png')
    else:
        return send_from_directory(icons_dir,'unknown.png')

@app.route('/package_download')
@allow_local
def package_download():
    package_uuid = request.args.get('package_uuid')
    if package_uuid:
        with sqlite3.connect(app.waptconfig.dbpath) as con:
            try:
                con.row_factory = sqlite3.Row
                query = '''\
                    select r.repo,r.package_uuid,r.filename
                    from wapt_package r
                    where r.package_uuid=?
                '''
                cur = con.cursor()
                cur.execute(query,[package_uuid])
                rows = cur.fetchall()
                return Response(jsondump([dict(r) for r in  rows]), mimetype='application/json')
            except sqlite3.Error as e:
                logger.critical("*********** Error %s:" % e.args[0])

@app.route('/download_icons')
@allow_local
def download_icons():
    username = None
    grpuser = []
    auth_header = request.headers.get('Authorization')
    if not auth_header:
        return authenticate()

    try:
        (username,grpuser) = check_auth_groups(auth_header)
        if username:
            logger.debug('User %s authenticated with groups %s' % (username,grpuser))
        else:
            return authenticate()
    except EWaptAuthException as e:
        return authenticate("Auth failed on local service: %s" % e)

    with sqlite3.connect(app.waptconfig.dbpath) as con:
        try:
            con.row_factory = sqlite3.Row
            query = '''\
                select
                    r.*,
                    s.version as install_version,s.install_status,s.install_date,s.explicit_by
                from wapt_package r
                left join wapt_localstatus s on s.package=r.package
                where not r.section in ("host","unit","profile","selfservice")
                order by r.package,r.version'''
            cur = con.cursor()
            cur.execute(query)
            rows = []

            search = request.args.get('q', '').replace('\\', '')
            keywords = ensure_list(request.args.get('keywords', '').lower())

            rules = None
            w = scoped_wapt()
            if w.is_enterprise():
                rules = common.self_service_rules(w)

            for row in cur.fetchall():
                pe = PackageEntry().load_control_from_dict(
                    dict((cur.description[idx][0], value) for idx, value in enumerate(row)))

                if len(keywords) > 0:
                    match_kw = False
                    package_keywords = ensure_list(pe.keywords.lower())
                    for kw in package_keywords:
                        if kw in keywords:
                            match_kw = True
                            break
                    if not match_kw:
                        continue

                if not search or pe.match_search(search):
                    if scoped_wapt().is_authorized_package_action('list', pe.package, grpuser, rules):
                        rows.append(pe)

            if request.args.get('latest', '0') == '1':
                filtered = []
                last_package_name = None
                for package in sorted(rows, reverse=True):
                    if package.package != last_package_name:
                        filtered.append(package)
                    last_package_name = package.package
                rows = list(reversed(filtered))

            if not request.args.get('all_versions', ''):
                rows = sorted(latest_only(rows))

        except sqlite3.Error as e:
            logger.critical("*********** Error %s:" % e.args[0])

    for pe in rows:
        # some basic search scoring
        score = 0
        if search in pe.package:
            score += 3
        if search in pe.description:
            score += 2
        pe.score = score

    rows = sorted(rows, key=lambda r: (r.score, r.signature_date, r.filename), reverse=True)
    if rows:
        app.task_manager.add_task(WaptDownloadIcon(copy.deepcopy(rows), created_by=session.get('username')))
    else:
        logger.debug("No package icon to download")

    return Response('downloading %s icons' % len(rows), mimetype='text')


@app.route('/package_details')
@app.route('/package_details.json')
@allow_local
def package_details():
    # wapt=Wapt(config_filename=app.waptconfig.config_filename)
    package = request.args.get('package')
    try:
        w = scoped_wapt()
        data = w.waptdb.installed_matching(package)
        if not data:
            data = w.is_available(package)
            # take the newest...
            data = data and data[-1].as_dict()
    except Exception as e:
        data = {'errors': [ensure_unicode(e)]}

    if request.args.get('format', 'html') == 'json':
        return Response(common.jsondump(dict(result=data, errors=[])), mimetype='application/json')
    else:
        return render_wapt_template('package_details.html', data=data)


@app.route('/runstatus')
@allow_local
def get_runstatus():
    data = []
    with sqlite3.connect(app.waptconfig.dbpath) as con:
        con.row_factory = sqlite3.Row
        try:
            query = """select value,create_date from wapt_params where name='runstatus' limit 1"""
            cur = con.cursor()
            cur.execute(query)
            rows = cur.fetchall()
            data = [dict(ix) for ix in rows]
        except Exception as e:
            logger.critical("*********** error " + ensure_unicode(e))
    return Response(common.jsondump(data), mimetype='application/json')


@app.route('/checkupgrades')
@app.route('/checkupgrades.json')
@allow_local
def get_checkupgrades():
    with sqlite3.connect(app.waptconfig.dbpath) as con:
        con.row_factory = sqlite3.Row
        data = ""
        try:
            query = """select * from wapt_params where name="last_update_status" limit 1"""
            cur = con.cursor()
            cur.execute(query)
            row = cur.fetchone()
            if row:
                data = json.loads(row['value'])
                # update runing_tasks.
                if app.task_manager:
                    with app.task_manager.status_lock:
                        if app.task_manager.running_task:
                            data['running_tasks'] = [app.task_manager.running_task.as_dict()]
                        else:
                            data['running_tasks'] = []
                        data['pending_tasks'] = [task.as_dict() for task in sorted(app.task_manager.tasks_queue.queue)]

                # if enterprise, add waptwua status
                query = """select * from wapt_params where name="waptwua.status" limit 1"""
                cur = con.cursor()
                cur.execute(query)
                row = cur.fetchone()
                if row:
                    data['wua_status'] = row['value']
                    # check count of updates
                    try:
                        query = """select * from wapt_params where name="waptwua.updates_localstatus" limit 1"""
                        cur = con.cursor()
                        cur.execute(query)
                        row = cur.fetchone()
                        if row:
                            wua_localstatus = json.loads(row['value'])
                            wua_pending_count = len([u['update_id'] for u in wua_localstatus if u['status'] == 'PENDING'])
                            data['wua_pending_count'] = wua_pending_count
                    except Exception as e:
                        logger.critical('Unable to read waptwua updates_localstatus from DB: %s' % e)
            else:
                data = None

        except Exception as e:
            logger.critical("*********** error %s" % (ensure_unicode(e)))
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=_('Update status'))


@app.route('/waptupgrade')
@app.route('/waptupgrade.json')
@allow_local_auth
def waptclientupgrade():
    """Launch an external 'wapt-get waptupgrade' process to upgrade local copy of wapt client"""
    data = app.task_manager.add_task(WaptClientUpgrade(created_by=session.get('username'))).as_dict()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title='Upgrade')


@app.route('/waptservicerestart')
@app.route('/waptservicerestart.json')
@allow_local_auth
def waptservicerestart():
    """Restart local waptservice using a spawned batch file"""
    data = app.task_manager.add_task(WaptServiceRestart(created_by=session.get('username'))).as_dict()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title='Restart waptservice with task scheduler')


@app.route('/reload_config')
@app.route('/reload_config.json')
@allow_local
def reload_config():
    """trigger reload of wapt-get.ini file for the service"""
    notify_user = int(request.args.get('notify_user', '0')) == 1
    data = app.task_manager.add_task(WaptNetworkReconfig(notify_user=notify_user, created_by=session.get('username'))).as_dict()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=_('Reload configuration'))


@app.route('/upgrade')
@app.route('/upgrade.json')
@allow_local
def upgrade():
    force = int(request.args.get('force', '0')) != 0
    notify_user = int(request.args.get('notify_user', '1')) != 0
    update_packages = int(request.args.get('update', '1')) != 0

    only_priorities = None
    if 'only_priorities' in request.args:
        only_priorities = ensure_list(request.args.get('only_priorities', None), allow_none=True)
    only_if_not_process_running = int(request.args.get('only_if_not_process_running', '0')) != 0

    if scoped_wapt().is_enterprise():
        install_wua_updates = int(request.args.get('install_wua_updates', '0')) != 0
    else:
        install_wua_updates = False

    all_tasks = []
    if update_packages:
        all_tasks.append(app.task_manager.add_task(WaptUpdate(force=force, notify_user=notify_user, created_by=session.get('username'))).as_dict())

    all_tasks.append(app.task_manager.add_task(WaptUpgrade(notify_user=notify_user, only_priorities=only_priorities,
                                                           only_if_not_process_running=only_if_not_process_running, force=force,created_by=session.get('username'))).as_dict())
    all_tasks.append(app.task_manager.add_task(WaptCleanup(notify_user=False, created_by=session.get('username'))))
    if scoped_wapt().is_enterprise() and setuphelpers.get_loggedinusers():
        all_tasks.append(app.task_manager.add_task(WaptRunSessionSetup(created_by=session.get('username'))))

    # append install wua tasks only if last scan reported to something to install
    if sys.platform == 'win32' and scoped_wapt().is_enterprise() and install_wua_updates and scoped_wapt().waptwua_enabled and scoped_wapt().read_param('waptwua.status', 'UNKNONW') != 'OK':
        all_tasks.append(app.task_manager.add_task(WaptWUAInstallTask(notify_user=False, created_by=session.get('username'))))

    data = {'result': 'OK', 'content': all_tasks}
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title='Upgrade')


@app.route('/download_upgrade')
@app.route('/download_upgrade.json')
@allow_local_auth
def download_upgrade():
    force = int(request.args.get('force', '0')) != 0
    notify_user = int(request.args.get('notify_user', '1')) != 0
    update_packages = int(request.args.get('update', '0')) != 0

    all_tasks = []
    if update_packages:
        all_tasks.append(app.task_manager.add_task(WaptUpdate(force=force, notify_user=notify_user, created_by=session.get('username'))).as_dict())

    all_tasks.append(app.task_manager.add_task(WaptDownloadUpgrade(notify_user=True, created_by=session.get('username'))))

    data = {'result': 'OK', 'content': all_tasks}
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title='Donwload Upgrades')


@app.route('/update')
@app.route('/update.json')
@allow_local
def update():
    task = WaptUpdate(created_by=session.get('username'))
    task.force = int(request.args.get('force', '0')) != 0
    task.notify_user = int(request.args.get('notify_user', '0' if not waptconfig.notify_user else '1')) != 0
    task.notify_server_on_finish = int(request.args.get('notify_server', '0')) != 0
    data = app.task_manager.add_task(task).as_dict()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=_('Installed software update'))


@app.route('/audit')
@app.route('/audit.json')
@allow_local
def audit():
    if not scoped_wapt().is_enterprise():
        return
    tasks = []
    notify_user = int(request.args.get('notify_user', '0' if not waptconfig.notify_user else '1')) != 0
    notify_server_on_finish = int(request.args.get('notify_server', '1')) != 0
    force = int(request.args.get('force', '0')) != 0

    now = datetime2isodate(datetime.datetime.utcnow())

    packagenames = ensure_list(request.args.get('package', None), allow_none=True)
    if packagenames is None:
        packagenames = []
        for package_status in scoped_wapt().installed():
            if force or not package_status.next_audit_on or (now >= package_status.next_audit_on):
                packagenames.append(package_status.package)

    if packagenames:
        task = WaptAuditPackage(packagenames, force=force, created_by=session.get('username'))
        task.notify_user = notify_user
        task.notify_server_on_finish = notify_server_on_finish
        tasks.append(app.task_manager.add_task(task).as_dict())

    data = {'result': 'OK', 'content': tasks, 'message': '%s tasks queued' % len(tasks)}
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=_('Triggered packages audits'))


@app.route('/update_status')
@app.route('/update_status.json')
@allow_local
def update_status():
    task = WaptUpdateServerStatus(created_by=session.get('username'))
    data = app.task_manager.add_task(task).as_dict()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=task)


@app.route('/longtask')
@app.route('/longtask.json')
@allow_local_auth
def longtask():
    notify_user = request.args.get('notify_user', None)
    if notify_user is not None:
        notify_user = int(notify_user)
    data = app.task_manager.add_task(
        WaptLongTask(
            duration=int(request.args.get('duration', '60')),
            raise_error=int(request.args.get('raise_error', 0)),
            notify_user=notify_user,
            created_by=session.get('username'))).as_dict()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=_('LongTask'))


@app.route('/cleanup')
@app.route('/cleanup.json')
@app.route('/clean')
@allow_local
def cleanup():
    task = WaptCleanup(created_by=session.get('username'))
    task.force = int(request.args.get('force', '0')) == 1
    task.notify_user = int(request.args.get('notify_user', '0')) == 1
    data = app.task_manager.add_task(task)
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data.as_dict(), title=_('Cleanup'))


@app.route('/install_log')
@app.route('/install_log.json')
@allow_local_auth
def install_log():
    logger.info("show install log")
    try:
        packagename = request.args.get('package')
        data = scoped_wapt().last_install_log(packagename)
    except Exception as e:
        data = {'result': 'ERROR', 'message': '{}'.format(ensure_unicode(e))}
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=_('Trace of the installation of {}').format(packagename))


@app.route('/enable')
@allow_local_auth
def enable():
    logger.info("enable tasks scheduling")
    data = scoped_wapt().enable_tasks()
    return Response(common.jsondump(data), mimetype='application/json')


@app.route('/disable')
@allow_local_auth
def disable():
    logger.info("disable tasks scheduling")
    data = scoped_wapt().disable_tasks()
    return Response(common.jsondump(data), mimetype='application/json')


@app.route('/register')
@app.route('/register.json')
@allow_local_auth
def register():
    logger.info("register computer")
    notify_user = int(request.args.get('notify_user', '0')) == 1
    data = app.task_manager.add_task(WaptRegisterComputer(created_by=session.get('username'), notify_user=notify_user)).as_dict()

    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=_('Saving host to the WAPT server'))


@app.route('/inventory')
@app.route('/inventory.json')
@allow_local_auth
def inventory():
    logger.info("Inventory")
    # wapt=Wapt(config_filename=app.waptconfig.config_filename)
    data = scoped_wapt().inventory()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data, title=_('Inventory of the host'))


@app.route('/install', methods=['GET'])
@app.route('/install.json', methods=['GET'])
@app.route('/install.html', methods=['GET'])
@allow_local
def install():
    package_requests = request.args.get('package')
    if not isinstance(package_requests, list):
        package_requests = [package_requests]

    force = int(request.args.get('force', '0')) == 1
    notify_user = int(request.args.get('notify_user', '0')) == 1
    only_priorities = None
    if 'only_priorities' in request.args:
        only_priorities = ensure_list(request.args.get('only_priorities', None), allow_none=True)
    only_if_not_process_running = int(request.args.get('only_if_not_process_running', '0')) != 0

    username = None
    usergroups = []

    auth_header = request.headers.get('Authorization')
    if not auth_header:
        return authenticate('Missing auth')

    try:
        (username,usergroups) = check_auth_groups(auth_header)
        if not username:
            return authenticate()
        else:
            logger.debug('User %s authenticated. Groups: %s' % (username,usergroups))
    except EWaptAuthException as e:
        return authenticate("Auth failed on local service: %s" % e)

    w = scoped_wapt()
    # on enterprise, rules are needed to know if packages removal is allowed
    rules = None
    if w.is_enterprise():
        rules = common.self_service_rules(w)

    authorized_packages = []
    for apackage in package_requests:
        if w.is_authorized_package_action('install', apackage, usergroups, rules):
            authorized_packages.append(apackage)
        else:
            return authenticate()

    logger.info("user %s authenticated" % username)

    if authorized_packages:
        data = app.task_manager.add_task(WaptPackageInstall(authorized_packages, force=force,
                                                            only_priorities=only_priorities, only_if_not_process_running=only_if_not_process_running,
                                                            notify_user=notify_user, created_by=username)).as_dict()
        if w.is_enterprise():
            app.task_manager.add_task(WaptAuditPackage(packagenames=authorized_packages, force=force, notify_user=notify_user, priority=100, created_by=username)).as_dict()
            if setuphelpers.get_loggedinusers():
                app.task_manager.add_task(WaptRunSessionSetup(created_by=username))
    else:
        data = []

    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('install.html', data=data)


@app.route('/remove', methods=['GET'])
@app.route('/remove.json', methods=['GET'])
@app.route('/remove.html', methods=['GET'])
@allow_local
def remove():
    package_requests = request.args.get('package')
    if not isinstance(package_requests, list):
        package_requests = [package_requests]

    force = int(request.args.get('force', '0')) == 1
    notify_user = int(request.args.get('notify_user', '0')) == 1
    only_priorities = None
    if 'only_priorities' in request.args:
        only_priorities = ensure_list(request.args.get('only_priorities', None), allow_none=True)
    only_if_not_process_running = int(request.args.get('only_if_not_process_running', '0')) != 0

    auth_header = request.headers.get('Authorization')
    if not auth_header:
        return authenticate('Missing auth')

    try:
        (username,usergroups) = check_auth_groups(auth_header)
        if not username:
            return authenticate()
        else:
            logger.debug('User %s authenticated. Groups: %s' % (username,usergroups))
    except EWaptAuthException as e:
        return authenticate("Auth failed on local service: %s" % e)

    w = scoped_wapt()
    # on enterprise, rules are needed to know if packages removal is allowed
    rules = None
    if w.is_enterprise():
        rules = common.self_service_rules(w)

    authorized_packages = []
    for apackage in package_requests:
        if w.is_authorized_package_action('remove', apackage, usergroups, rules):
            authorized_packages.append(apackage)
        else:
            # if one of the package is not authorized, assume whole request is invalid...
            return authenticate('NOT_AUTHORIZED_BY_RULES')

    logger.info("user %s authenticated" % username)

    data = []
    if authorized_packages:
        for package in authorized_packages:
            data.append(app.task_manager.add_task(
                WaptPackageRemove(package, force=force, created_by=username, only_if_not_process_running=only_if_not_process_running, only_priorities=only_priorities, notify_user=notify_user)).as_dict())

    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('remove.html', data=data)


@app.route('/forget', methods=['GET'])
@app.route('/forget.json', methods=['GET'])
@allow_local_auth
def forget():
    packages = request.args.get('package')
    if not isinstance(packages, list):
        packages = [packages]
    logger.info("Forget package(s) %s" % packages)
    notify_user = int(request.args.get('notify_user', '0')) == 1
    data = app.task_manager.add_task(
        WaptPackageForget(packages, created_by=session.get('username'), notify_user=notify_user)).as_dict()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('install.html', data=data)


@app.route('/', methods=['GET'])
@allow_local
def index():
    host_info = setuphelpers.host_info()
    host_info.update(**setuphelpers.host_info_networking())
    data = dict(
                host_info=host_info,
                wapt=scoped_wapt(),
                wapt_info=scoped_wapt().wapt_status(),
                update_status=scoped_wapt().get_last_update_status(),
                _=_)
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('index.html', **data)


@app.route('/is_enterprise', methods=['GET'])
@allow_local
def is_enterprise():
    result = scoped_wapt().is_enterprise()
    return Response(common.jsondump(result), mimetype='application/json')


@app.route('/tasks')
@app.route('/tasks.json')
@allow_local
def tasks():
    last_received_event_id = int(request.args.get('last_event_id', '-1'))
    timeout = int(request.args.get('timeout', '-1'))

    data = None
    start_time = time.time()
    status = 200

    while True:
        # wait for events manager initialisation
        if app.task_manager.events:
            actual_last_event_id = app.task_manager.events.last_event_id()
            if actual_last_event_id is not None and actual_last_event_id <= last_received_event_id:
                if (time.time() - start_time) * 1000 > timeout:
                    status = 204
                    break
            elif actual_last_event_id is None or actual_last_event_id > last_received_event_id:
                status = 200
                data = app.task_manager.tasks_status()
                break
        if time.time() - start_time > timeout:
            break

        # avoid eating cpu
        time.sleep(0.1)

    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json', status=status)
    else:
        return render_wapt_template('tasks.html', data=data, status=status)


@app.route('/tasks_status')
@app.route('/tasks_status.json')
@allow_local
def tasks_status():
    last_received_event_id = int(request.args.get('last_event_id', '-1'))
    timeout = int(request.args.get('timeout', '-1'))

    result = {}
    start_time = time.time()
    data = None
    status = 200

    while True:
        if app.task_manager.events:
            actual_last_event_id = app.task_manager.events.last_event_id()
            result['last_event_id'] = actual_last_event_id
            if actual_last_event_id is not None and actual_last_event_id <= last_received_event_id:
                if (time.time() - start_time) * 1000 > timeout:
                    status = 204
                    break
            elif actual_last_event_id is None or actual_last_event_id > last_received_event_id:
                data = app.task_manager.tasks_status()
                status = 200
                break

        if time.time() - start_time > timeout:
            break

        # avoid eating cpu
        time.sleep(0.1)

    if data:
        tasks = []
        tasks.extend(data['pending'])
        if data['running']:
            tasks.append(data['running'])
        tasks.extend(data['done'])
        tasks.extend(data['errors'])
        tasks.extend(data['cancelled'])
        result['tasks'] = tasks
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(result), mimetype='application/json', status=status)
    else:
        return render_wapt_template('tasks.html', data=result, status=status)


@app.route('/task')
@app.route('/task.json')
@app.route('/task.html')
@allow_local
def task():
    id = int(request.args['id'])
    tasks = app.task_manager.tasks_status()
    all_tasks = tasks['done']+tasks['pending']+tasks['errors']
    if tasks['running']:
        all_tasks.append(tasks['running'])
    all_tasks = [task for task in all_tasks if task and task['id'] == id]
    if all_tasks:
        task = all_tasks[0]
    else:
        task = {}
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(task), mimetype='application/json')
    else:
        return render_wapt_template('task.html', task=task)


@app.route('/cancel_all_tasks')
@app.route('/cancel_all_tasks.html')
@app.route('/cancel_all_tasks.json')
@allow_local
def cancel_all_tasks():
    data = app.task_manager.cancel_all_tasks()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data)


@app.route('/cancel_running_task')
@app.route('/cancel_running_task.json')
@allow_local
def cancel_running_task():
    data = app.task_manager._wapt.task_is_cancelled.set()
    if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
        return Response(common.jsondump(data), mimetype='application/json')
    else:
        return render_wapt_template('default.html', data=data)


@app.route('/cancel_task')
@app.route('/cancel_task.json')
@allow_local
def cancel_task():
    try:
        id = int(request.args['id'])
        data = app.task_manager.cancel_task(id)
        if request.args.get('format', 'html') == 'json' or request.path.endswith('.json'):
            return Response(common.jsondump(data), mimetype='application/json')
        else:
            return render_wapt_template('default.html', data=data)
    except Exception as e:
        logger.critical("*********** Error %s:" % e.args[0])
        return Response(common.jsondump(e), mimetype='application/json')


@app.route('/wapt/<string:input_package_name>')
@allow_local
def get_wapt_package(input_package_name):
    package_name = secure_filename(input_package_name)
    cache_dir = scoped_wapt().packages_cache_dir
    local_fn = os.path.join(cache_dir, package_name)
    force = int(request.args.get('force', '0')) == 1

    if package_name == 'Packages' and (not os.path.isfile(local_fn) or force):
        local_repo = WaptLocalRepo(cache_dir)
        local_repo.update_packages_index(force_all=force)

    if os.path.isfile(local_fn):
        r = send_from_directory(cache_dir, package_name)
        if 'content-length' not in r.headers:
            r.headers.add_header(
                'content-length', int(os.path.getsize(local_fn)))
        return r
    else:
        return Response(status=404)


@app.route('/events')
@app.route('/events.json')
@allow_local
def events():
    """Get the last waptservice events.
    Blocking call for timeout seconds.

    Args:
        last_read (int): id of last read event.
        timeout (float): time to wait until new events come in
    """
    last_read = int(request.args.get('last_read', session.get('last_read_event_id', '0')))
    timeout = int(request.args.get('timeout', '10000'))
    max_count = int(request.args.get('max_count', '0')) or None
    if app.task_manager.events:
        data = app.task_manager.events.get_missed(last_read=last_read, max_count=max_count)
        if not data and timeout > 0.0:
            start_time = time.time()
            while not data and (time.time() - start_time) * 1000 <= timeout:
                time.sleep(1.0)
                data = app.task_manager.events.get_missed(last_read=last_read, max_count=max_count)
            if app.task_manager.events.events:
                session['last_read_event_id'] = app.task_manager.events.events[-1].id
    else:
        data = None
    return Response(jsondump(data), mimetype='application/json')


@app.route('/polldebug.json',methods=['POST'])
@allow_local
def polldebug():
    return Response(jsondump(request.json), mimetype='application/json')

class WaptTaskManager(threading.Thread):
    def __init__(self, config_filename='c:/wapt/wapt-get.ini'):
        threading.Thread.__init__(self)
        self.name = 'WaptTaskManager'
        self.status_lock = threading.RLock()
        self._wapt = None
        self.tasks = []

        self.tasks_queue = queue.PriorityQueue()
        self.tasks_counter = 0

        self.tasks_done = []
        self.tasks_error = []
        self.tasks_cancelled = []
        self.events = None

        self.running_task = None
        self.config_filename = config_filename

        self.last_update_server_date = None

        self.last_upgrade = None
        self.last_update = None
        self.last_audit = None
        self.last_sync = None
        self.last_forced_installs = None

        self.logger = logging.getLogger('wapttasks')


    @property
    def wapt(self):
        if not self._wapt:
            self._wapt = Wapt(config_filename=self.config_filename)
            self._wapt.threadid = threading.current_thread().ident
        if self._wapt.threadid != threading.current_thread().ident:
            raise Exception('Access to task_manager Wapt instance from a foreign Thread')
        return self._wapt

    def setup_event_queue(self):
        self.events = WaptEvents()
        return self.events

    def update_runstatus(self, status):
        # update database with new runstatus
        self.wapt.runstatus = status
        if self.events:
            # dispatch event to listening parties
            self.events.post_event("STATUS", self.wapt.get_last_update_status())

    def update_server_status(self):
        try:
            if not self.wapt.waptserver:  # starter edition
                return
            result = self.wapt.update_server_status()
            if result and result['success'] and result['result']['uuid']:
                self.last_update_server_date = datetime.datetime.utcnow()
            elif result and not result['success']:
                self.logger.critical('Unable to update server status: %s' % result['msg'])
            else:
                self.logger.info('Unable to update server status: No answer')
        except Exception as e:
            self.logger.info('Unable to update server status: %s' % repr(e))

    def broadcast_tasks_status(self, event_type, task):
        """event_type : TASK_ADD TASK_START TASK_STATUS TASK_FINISH TASK_CANCEL TASK_ERROR
        """
        if self.events and task:
            self.events.post_event(event_type, task.as_dict())

    def add_task(self, task):
        """Adds a new WaptTask for processing"""
        with self.status_lock:
            same = [pending for pending in self.tasks_queue.queue if pending.same_action(task)]
            if self.running_task and self.running_task.same_action(task):
                same.append(self.running_task)

            # keep track of last update/upgrade add date to avoid relaunching
            if isinstance(task, WaptUpdate):
                self.last_update = datetime.datetime.utcnow()
            if isinstance(task, WaptUpgrade):
                self.last_upgrade = datetime.datetime.utcnow()
            if isinstance(task, WaptSyncRepo):
                self.last_sync = datetime.datetime.utcnow()

            # not already in pending  actions...
            if not same:
                self.logger.info('Add task %s to queue' % (task,))

                task.task_manager = self

                self.tasks_counter += 1
                task.id = self.tasks_counter
                # default order is task id
                task.order = self.tasks_counter
                self.tasks_queue.put(task)
                self.tasks.append(task)
                self.broadcast_tasks_status('TASK_ADD', task)
                return task
            else:
                self.logger.debug('Not adding task %s. Already a same action in queue' % (task,))
                return same[0]

    def check_configuration(self):
        """Check wapt configuration, reload ini file if changed"""
        try:
            self.logger.debug("Checking if config file has changed")
            if self.wapt.reload_config_if_updated():
                self.logger.info("Core Wapt config file has changed, reloaded. Reloading waptservice config too")
                waptconfig.load()
                return True
            return False

        except Exception as e:
            self.logger.critical('Error reloading configuration: %s' % ensure_unicode(e))
            return False

    def run_scheduled_audits(self):
        """Add packages audit tasks to the queue"""
        now = datetime2isodate(datetime.datetime.utcnow())

        self.last_audit = datetime.datetime.utcnow()

        packages = [ pe['package'] for pe in self.wapt.waptdb.query("""select
                package,next_audit_on
            from wapt_localstatus
            where install_status = 'OK' and
                (next_audit_on='' or next_audit_on is null or next_audit_on <= '%s') and
                (next_audit_on > last_audit_on or last_audit_on is null)
            order by next_audit_on
            """ % now)]

        if packages:
            task = WaptAuditPackage(packages, created_by='SCHEDULER')
            task.notify_server_on_finish=True
            self.add_task(task)

    @property
    def last_waptwua_download(self):
        return self.wapt.read_param('last_waptwua_download', ptype='datetime')

    @last_waptwua_download.setter
    def last_waptwua_download(self, value):
        if value is None:
            self.wapt.delete_param('last_waptwua_download')
        else:
            self.wapt.write_param('last_waptwua_download', value)

    @property
    def last_waptwua_install(self):
        return self.wapt.read_param('last_waptwua_install', ptype='datetime')

    @last_waptwua_install.setter
    def last_waptwua_install(self, value):
        if value is None:
            self.wapt.delete_param('last_waptwua_install')
        else:
            self.wapt.write_param('last_waptwua_install', value)

    def run_immediate_installs(self,force=False,only_priorities=None, only_if_not_process_running=False):
        self.logger.debug('Check if package should be forcibly installed')
        last_update_status = self.wapt.get_last_update_status()
        immediate_install = last_update_status.get('pending',{}).get('immediate_installs',[])
        if immediate_install:
            for package in immediate_install:
                self.add_task(WaptPackageInstall(package, force=force, created_by='SCHEDULER', installed_by='SCHEDULER',
                    only_priorities=only_priorities, only_if_not_process_running=only_if_not_process_running, notify_server_on_finish=True))


    def check_scheduled_tasks(self):
        """Add update/upgrade tasks if elapsed time since last update/upgrade is over"""
        self.logger.debug('Check scheduled tasks')

        if datetime.datetime.utcnow() - self.start_time >= datetime.timedelta(days=1):
            self.start_time = datetime.datetime.utcnow()
            self.add_task(WaptServiceRestart(created_by='DAILY RESTART'))

        if waptconfig.waptupdate_task_period is not None:
            if self.last_update is None or \
                    (datetime.datetime.utcnow() - self.last_update) > get_time_delta(waptconfig.waptupdate_task_period, 'm') or \
                    (setuphelpers.datetime2isodate() > self.wapt.read_param('next_update_on', '9999-12-31')):
                try:
                    self.wapt.update()
                    if waptconfig.download_after_update_with_waptupdate_task_period:
                        reqs = self.wapt.check_downloads()
                        for req in reqs:
                            self.add_task(WaptDownloadPackage(req.asrequirement(), notify_user=True, created_by='SCHEDULER'))
                    self.add_task(WaptUpdate(notify_user=False, notify_server_on_finish=True, created_by='SCHEDULER'))
                except Exception as e:
                    self.logger.critical('Error for update in check_scheduled_tasks: %s' % e)

        if self.last_forced_installs is None or (datetime.datetime.utcnow() - self.last_forced_installs) > get_time_delta(waptconfig.forced_installs_task_period, 'm'):
            self.last_forced_installs = datetime.datetime.utcnow()
            self.run_immediate_installs()

        if waptconfig.waptupgrade_task_period is not None and setuphelpers.running_on_ac():
            if self.last_upgrade is None or (datetime.datetime.utcnow() - self.last_upgrade) > get_time_delta(waptconfig.waptupgrade_task_period, 'm'):
                try:
                    self.add_task(WaptUpgrade(notify_user=False, created_by='SCHEDULER', only_if_not_process_running=True, notify_server_on_finish=True))
                except Exception as e:
                    self.logger.warning('Error for upgrade in check_scheduled_tasks: %s' % e)
                self.add_task(WaptCleanup(notify_user=False, created_by='SCHEDULER'))
                if self.wapt.is_enterprise() and setuphelpers.get_loggedinusers():
                    self.add_task(WaptRunSessionSetup())

        next_audit = self.wapt.get_next_audit_datetime()
        if not next_audit:
            if self.last_audit is None:
                next_audit = datetime.datetime.utcnow()
            else:
                next_audit = self.last_audit + get_time_delta(self.wapt.waptaudit_task_period)
        if next_audit and datetime.datetime.utcnow() >= next_audit:
            try:
                self.run_scheduled_audits()
            except Exception as e:
                self.logger.critical('Error checking audit: %s' % e)

        if waptconfig.enable_remote_repo and self.wapt.is_enterprise() and not(waptconfig.sync_only_forced):
            if (self.last_sync is None or (datetime.datetime.utcnow() - self.last_sync > get_time_delta(waptconfig.local_repo_sync_task_period, 'm'))) and ((waptconfig.local_repo_time_for_sync_start is None) or is_between_two_times(waptconfig.local_repo_time_for_sync_start, waptconfig.local_repo_time_for_sync_end)):
                try:
                    self.add_task(WaptSyncRepo(notify_user=False, created_by='SCHEDULER', socketio_client=sio.socketio_client))
                except Exception as e:
                    self.logger.critical('Error syncing local repo with server repo : %s' % e)

        if sys.platform == 'win32' and self.wapt.is_enterprise() and self.wapt.waptwua_enabled:
            params = WaptWUAParams()
            if self.wapt.config and self.wapt.config.has_section('waptwua'):
                params.load_from_ini(config=self.wapt.config, section='waptwua')

            uptime_minutes = win32api.GetTickCount() / 60 / 1000
            if params.install_scheduling:
                # delayed 10 minutes after boot
                if uptime_minutes > get_time_delta(params.postboot_delay).total_seconds()/60:
                    if self.last_waptwua_install is None or (datetime.datetime.utcnow() - self.last_waptwua_install > get_time_delta(params.install_scheduling, 'd')):
                        #if self.wapt.read_param('waptwua.status', '') == 'PENDING_UPDATES':
                        self.add_task(WaptWUAInstallTask(notify_user=False, notify_server_on_finish=True, created_by='SCHEDULER'))
                        self.last_waptwua_install = datetime.datetime.utcnow()
                        self.last_waptwua_download = datetime.datetime.utcnow()

            if params.download_scheduling:
                # delayed 10 minutes after boot
                if uptime_minutes > get_time_delta(params.postboot_delay).total_seconds()/60:
                    if self.last_waptwua_download is None or (datetime.datetime.utcnow() - self.last_waptwua_download > get_time_delta(params.download_scheduling, 'd')):
                        self.add_task(WaptWUADowloadTask(notify_user=False, notify_server_on_finish=True, created_by='SCHEDULER'))
                        self.last_waptwua_download = datetime.datetime.utcnow()

    def run(self):
        """Queue management, event processing"""
        if sys.platform == 'win32':
            try:
                pythoncom.CoInitializeEx(pythoncom.COINIT_MULTITHREADED)
            except pythoncom.com_error:
                # already initialized.
                pass

        self.start_time = datetime.datetime.utcnow()
        self.setup_event_queue()

        self.logger.info('Wapt tasks management initialized with {} configuration'.format(self.config_filename))
        self.logger.info('Core Configuration: %s' % self.wapt.as_dict())

        if waptconfig.reconfig_on_network_change:
            self.start_network_monitoring()
            self.start_ipaddr_monitoring()

        self.logger.info("Wapt tasks queue started")

        while True:
            try:
                need_sleep = False
                # check wapt configuration, reload ini file if changed
                # reload wapt config
                self.check_configuration()
                # force update if host capabilities have changed
                new_capa = self.wapt.host_capabilities_fingerprint()
                old_capa = self.wapt.read_param('host_capabilities_fingerprint')
                last_update_config_fingerprint = self.wapt.read_param('last_update_config_fingerprint')
                if old_capa != new_capa or last_update_config_fingerprint != self.wapt.merged_config_hash:
                    if old_capa != new_capa:
                        self.logger.info('Host capabilities have changed since last update %s -> %s, forcing update' % (old_capa,new_capa))
                    if last_update_config_fingerprint != self.wapt.merged_config_hash:
                        self.logger.info('Service config have changed %s -> %s, forcing update' % (last_update_config_fingerprint,self.wapt.merged_config_hash))
                    # force update
                    self.last_update = None

                # check tasks queue
                next_task = self.tasks_queue.get(timeout=waptconfig.waptservice_poll_timeout)
                try:
                    nowutc = datetime2isodate()
                    if (next_task.start_not_before and nowutc < next_task.start_not_before):
                        # requeue...
                        self.logger.debug('Tasks %s is not yet ready to start. not before %s' % (next_task,next_task.start_not_before))
                        need_sleep = self.tasks_queue.empty()
                        self.tasks_queue.put(next_task)
                        # force a sleep
                        if need_sleep:
                            raise queue.Empty()
                        else:
                            continue

                    if (next_task.start_not_after and nowutc < next_task.start_not_after):
                        self.logger.info('Tasks %s is too old. Canceled. not after %s' % (next_task,next_task.start_not_after))
                        self.tasks_cancelled.append(next_task)
                        continue

                    self.running_task = next_task

                    self.logger.info('Running task %s created by %s' % (self.running_task,self.running_task.created_by))
                    # don't send update_run status for updatestatus itself...
                    self.broadcast_tasks_status('TASK_START', self.running_task)
                    if self.running_task.notify_server_on_start:
                        self.update_runstatus(_('Running: {description}').format(description=self.running_task))
                        self.update_server_status()
                    try:
                        def update_running_status(append_line=None, set_status=None):
                            if append_line:
                                self.wapt.runstatus = append_line

                                if self.running_task:
                                    self.running_task.logs.append(append_line)

                                if self.events:
                                    self.events.post_event('PRINT', ensure_unicode(append_line))

                                if self.events and self.running_task:
                                    self.broadcast_tasks_status('TASK_STATUS', self.running_task)

                        with LogOutput(console=sys.stderr, update_status_hook=update_running_status):
                            try:
                                self.running_task.run()
                            except Exception as e:
                                print(e)
                                raise


                        if self.running_task:
                            running = self.running_task
                            self.running_task = None
                            self.tasks_done.append(running)

                            self.broadcast_tasks_status('TASK_FINISH', running)
                            if running.notify_server_on_finish:
                                self.update_runstatus(_('Done: {description}\n{summary}').format(description=running, summary=running.summary))
                                self.update_server_status()

                    except common.EWaptCancelled as e:
                        self.logger.info('Task cancelled %s' % (self.running_task,))
                        if self.running_task:
                            self.running_task.logs.append("{}".format(ensure_unicode(e)))
                            self.running_task.summary = _("Canceled")
                            self.tasks_cancelled.append(self.running_task)
                            self.broadcast_tasks_status('TASK_CANCEL', self.running_task)
                    except Exception as e:
                        if self.running_task:
                            self.running_task.logs.append("{}".format(ensure_unicode(e)))
                            self.running_task.logs.append(ensure_unicode(traceback.format_exc()))
                            self.running_task.summary = "{}".format(ensure_unicode(e))
                            self.tasks_error.append(self.running_task)
                            self.broadcast_tasks_status('TASK_ERROR', self.running_task)
                        self.logger.critical('Task error %s: %s' % (self.running_task,ensure_unicode(e)))
                        try:
                            self.logger.debug(ensure_unicode(traceback.format_exc()))
                        except:
                            print("Traceback error")
                finally:
                    self.tasks_queue.task_done()
                    try:
                        self.update_runstatus('')
                    except Exception:
                        self.logger.warning('Error reset runstatus : %s' % traceback.format_exc())

                    self.running_task = None
                    # trim history lists
                    if len(self.tasks_cancelled) > waptconfig.MAX_HISTORY:
                        del self.tasks_cancelled[:len(self.tasks_cancelled)-waptconfig.MAX_HISTORY]
                    if len(self.tasks_done) > waptconfig.MAX_HISTORY:
                        del self.tasks_done[:len(self.tasks_done)-waptconfig.MAX_HISTORY]
                    if len(self.tasks_error) > waptconfig.MAX_HISTORY:
                        del self.tasks_error[:len(self.tasks_error)-waptconfig.MAX_HISTORY]
                    gc.collect()

            except queue.Empty:
                try:
                    self.update_runstatus('')
                except Exception:
                    self.logger.warning('Error reset runstatus : %s' % traceback.format_exc())

                try:
                    self.check_scheduled_tasks()
                except Exception:
                    self.logger.warning('Error checking scheduled tasks : %s' % traceback.format_exc())
                self.logger.debug("{} i'm still alive... but nothing to do".format(datetime.datetime.utcnow()))
                if need_sleep:
                    # avoid looping too fast in case there are only tasks with start_not_before in future in queue
                    time.sleep(2)

            except Exception as e:
                if sys.platform == 'win32':
                    self.logger.critical('Unhandled error in task manager loop: %s. Sleeping 120s before restarting the service' % ensure_unicode(e))
                    time.sleep(120)
                    # ask waptsvc to restart service
                    self.logger.critical('Forced restart waptservice by waptsvc')
                    harakiri(10)
                else:  # TODO linux/macos : restart service
                    self.logger.critical('Unhandled error in task manager loop: %s. Exiting service' % ensure_unicode(e))
                    os._exit(1)

    def current_task_counter(self):
        with self.status_lock:
            return self.tasks_counter

    def tasks_status(self):
        """Returns list of pending, error, done tasks, and current running one"""
        with self.status_lock:
            return dict(
                running=self.running_task and self.running_task.as_dict(),
                pending=[task.as_dict() for task in sorted(self.tasks_queue.queue)],
                done=[task.as_dict() for task in self.tasks_done],
                cancelled=[task.as_dict() for task in self.tasks_cancelled],
                errors=[task.as_dict() for task in self.tasks_error],
                last_task_id=self.tasks_counter,
                last_event_id=self.events.last_event_id() if self.events else None,
            )

    def cancel_running_task(self):
        """Cancel running task. Returns cancelled task"""
        with self.status_lock:
            if self.running_task:
                try:
                    cancelled = self.running_task
                    self.tasks_error.append(self.running_task)
                    try:
                        self.running_task.kill()
                    except Exception as e:
                        print(e)
                finally:
                    self.running_task = None
                if cancelled:
                    self.tasks_cancelled.append(cancelled)
                    self.broadcast_tasks_status('TASK_CANCEL', cancelled)
                return cancelled
            else:
                return None

    def cancel_task(self, id):
        """Cancel running or pending task with supplied id.
            return cancelled task"""
        with self.status_lock:
            cancelled = None
            if self.running_task and self.running_task.id == id:
                cancelled = self.running_task
                try:
                    self.running_task.kill()
                except Exception as e:
                    print(e)
                finally:
                    self.running_task = None
            else:
                for task in self.tasks_queue.queue:
                    if task.id == id:
                        cancelled = task
                        self.tasks_queue.queue.remove(task)
                        break
                if cancelled:
                    try:
                        cancelled.kill()
                    except Exception as e:
                        print(e)
            if cancelled:
                self.broadcast_tasks_status('TASK_CANCEL', cancelled)
            return cancelled

    def cancel_all_tasks(self):
        """Cancel running and pending tasks. Returns list of cancelled tasks"""
        with self.status_lock:
            cancelled = []
            while not self.tasks_queue.empty():
                cancelled.append(self.tasks_queue.get())

            for task in cancelled:
                self.tasks_cancelled.append(task)
                self.broadcast_tasks_status('TASK_CANCEL', task)

            if self.running_task:
                try:
                    cancelled.append(self.running_task)
                    self.tasks_error.append(self.running_task)
                    try:
                        self.running_task.kill()
                    except Exception as e:
                        print(e)
                    self.broadcast_tasks_status('TASK_CANCEL', self.running_task)
                finally:
                    self.running_task = None
            return cancelled

    def start_ipaddr_monitoring(self):
        nac = ctypes.windll.iphlpapi.NotifyAddrChange

        def addr_change(taskman):
            while True:
                nac(0, 0)
                delayed_start = datetime2isodate(datetime.datetime.utcnow() + datetime.timedelta(seconds=20))
                taskman.add_task(WaptNetworkReconfig(created_by='IPADDR_MONITORING',start_not_before=delayed_start))

        nm = threading.Thread(target=addr_change, args=(self,), name='ip_monitoring')
        nm.daemon = True
        nm.start()
        self.logger.debug("Wapt network address monitoring started")

    def start_network_monitoring(self):
        nrc = ctypes.windll.iphlpapi.NotifyRouteChange

        def connected_change(taskman):
            while True:
                nrc(0, 0)
                delayed_start = datetime2isodate(datetime.datetime.utcnow() + datetime.timedelta(seconds=20))
                taskman.add_task(WaptNetworkReconfig(created_by='NETWORK_MONITORING',start_not_before=delayed_start))

        nm = threading.Thread(target=connected_change, args=(self,), name='network_monitoring')
        nm.daemon = True
        nm.start()
        self.logger.debug("Wapt connection monitor started")

    def __str__(self):
        return "\n".join(self.tasks_status())


def install_service():
    """Setup waptservice as a windows Service managed by waptsvc
    >>> install_service()
    """
    if setuphelpers.service_installed('waptservice'):
        if not setuphelpers.service_is_stopped('waptservice'):
            logger.info('Stop running waptservice')
            setuphelpers.service_stop('waptservice')
            while not setuphelpers.service_is_stopped('waptservice'):
                logger.debug('Waiting for waptservice to terminate')
                time.sleep(2)
        logger.info('Unregister existing waptservice')
        setuphelpers.service_delete('waptservice')

    #service_binary = os.path.abspath(os.path.join(wapt_root_dir, 'waptsvc.exe'))
    # make a symlink for waptsvc.exe -> waptservice.exe
    service_link = os.path.abspath(os.path.join(wapt_root_dir, 'waptservice.exe'))
    #if os.path.exists(service_link):
    #    os.unlink(service_link)
    #os.symlink(service_binary,service_link)

    run('"%s" --install' % service_link)

    logger.info('Allow authenticated users to start/stop waptservice')
    if waptconfig.allow_user_service_restart:
        run('sc sdset waptservice D:(A;;CCLCSWRPWPDTLOCRRC;;;SY)(A;;CCDCLCSWRPWPDTLOCRSDRCWDWO;;;BA)(A;;CCLCSWLOCRRC;;;IU)(A;;CCLCSWLOCRRC;;;SU)(A;;CCLCSWRPWPDTLOCRRC;;;S-1-5-11)S:(AU;FA;CCDCLCSWRPWPDTLOCRSDRCWDWO;;;WD)')
    else:
        run('sc sdset waptservice D:(A;;CCLCSWRPLORC;;;AU)(A;;CCDCLCSWRPWPDTLOCRSDRCWDWO;;;BA)(A;;CCDCLCSWRPWPDTLOCRSDRCWDWO;;;SY)S:(AU;FA;CCDCLCSWRPWPDTLOSDRCWDWO;;;WD)')


def setup_logging(config=None):
    loglevel = config.loglevel
    for log in WAPTLOGGERS:
        sublogger = logging.getLogger(log)
        if sublogger:
            if hasattr(config, 'loglevel_%s' % log) and getattr(config, 'loglevel_%s' % log):
                setloglevel(sublogger, getattr(config, 'loglevel_%s' % log))
            else:
                if log == 'wapttasks':
                    setloglevel(sublogger, 'info')
                else:
                    setloglevel(sublogger, loglevel)


if __name__ == "__main__":
    usage = """\
    %prog -c configfile [action]

    WAPT Service.

    action is either :
      <nothing> : run service in foreground
      install   : install as a Windows service managed by waptsvc

    """

    if sys.platform == 'win32':
        try:
            pythoncom.CoInitializeEx(pythoncom.COINIT_MULTITHREADED)
        except:
            pass # can not be initialized 2 times

    parser = OptionParser(usage=usage, version='service ' + __version__+' common.py '+common.__version__+' setuphelpers.py '+setuphelpers.__version__)
    parser.add_option("-c", "--config", dest="config", default=os.path.join(wapt_root_dir, 'wapt-get.ini'), help="Config file full path (default: %default)")
    parser.add_option("-l", "--loglevel", dest="loglevel", default=None, type='choice',  choices=['debug', 'warning', 'info', 'error', 'critical'], metavar='LOGLEVEL', help="Loglevel (default: warning)")
    parser.add_option("-d", "--devel", dest="devel", default=False, action='store_true', help="Enable debug mode (for development only)")

    for log in WAPTLOGGERS:
        parser.add_option('--loglevel_%s' % log, dest='loglevel_%s' % log, default=None, type='choice',
                          choices=['debug', 'warning', 'info', 'error', 'critical'],
                          metavar='LOGLEVEL', help='Loglevel %s (default: warning)' % log)

    (options, args) = parser.parse_args()

    if args and 'install' in args[:2]:
        print('Installing waptservice')
        install_service()
        print('Done')
        sys.exit(0)

    waptconfig.config_filename = options.config
    waptconfig.load(True)

    for att in options.__dict__:
        if hasattr(waptconfig, att) and getattr(options, att) is not None:
            setattr(waptconfig, att, getattr(options, att))

    setup_logging(waptconfig)

    if waptconfig.log_to_windows_events:
        try:
            from logging.handlers import NTEventLogHandler
            hdlr = NTEventLogHandler('waptservice')
            logger.addHandler(hdlr)
        except Exception as e:
            logger.warning('Unable to initialize windows log Event handler: %s' % e)

    #for log in WAPTLOGGERS:
    for log in ['waptservice']:
        for f in glob.glob(os.path.join(wapt_root_dir,'log','%s-*.log' % log)):
            delta = datetime.date.today() - datetime.date.fromtimestamp(os.path.getmtime(f))
            if delta.days > int(waptconfig.log_retention_days):
                os.unlink(f)

    hdlr = logging.StreamHandler()
    hdlr.setFormatter(
        logging.Formatter('%(asctime)s [%(name)s %(threadName)s %(thread)s] %(levelname)s %(message)s'))
    rootlogger = logging.getLogger()
    rootlogger.addHandler(hdlr)
    setloglevel(rootlogger, options.loglevel)

    # app.logger.removeHandler(default_handler)
    # app.logger.addHandler(hdlr)

    # setup basic settings
    if sys.platform == 'win32':
        apply_host_settings(waptconfig)
        # waptwua
        apply_waptwua_settings(waptconfig)


    tasks_logger.info('Service Configuration: %s' % waptconfig)

    # starts one WaptTasksManager
    tasks_logger.info('Starting task queue')
    task_manager = WaptTaskManager(config_filename=waptconfig.config_filename)

    task_manager.daemon = True
    task_manager.start()
    app.task_manager = task_manager
    if sys.platform == 'win32':
        waptwua_api.task_manager = task_manager

    tasks_logger.info('Tasks queue running')
    try:
        sio = WaptSocketIOClient(waptconfig.config_filename, task_manager=task_manager)
        sio_logger = logging.getLogger('socketIO-client-2')
        hdlr = logging.StreamHandler()
        hdlr.setFormatter(
            logging.Formatter('%(asctime)s [%(name)-15s] %(levelname)s %(message)s'))
        sio_logger.addHandler(hdlr)

        sio.start()
        if options.loglevel:
            setloglevel(sio_logger, options.loglevel)
        else:
            setloglevel(sio_logger, waptconfig.loglevel)

        if options.devel:
            tasks_logger.info('Starting local dev waptservice...')
            app.run(host='127.0.0.1', port=8088, debug=False)
        else:
            port_config = []
            if waptconfig.waptservice_port:
                tasks_logger.info('Starting waitress waptservice on port %s' % waptconfig.waptservice_port)
                waitress_logger = logging.getLogger('waitress')
                if options.loglevel:
                    setloglevel(waitress_logger, options.loglevel)
                else:
                    setloglevel(waitress_logger, waptconfig.loglevel)
                serve(app, host='127.0.0.1', port=waptconfig.waptservice_port, threads=10)
    finally:
        tasks_logger.info('Waptservice stopped')
